/*
 *        spBase.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include <sp/spDefs.h>

#if defined(_WIN32)
#include <windows.h>
#if defined(_MSC_VER)
#pragma warning(push)
#pragma warning(disable : 4091)
#endif
#include <shlobj.h>
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
#elif defined(ANDROID)
#include <android/log.h>
#endif

#if !defined(MACOS)
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#if defined(_WIN32) && !defined(__CYGWIN32__)
#include <io.h>
#include <direct.h>
#include <process.h>
#else
#include <unistd.h>
#include <pwd.h>
#endif
#else
#if defined(MACOSX)
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#if defined(IPHONE)
#include <CoreFoundation/CoreFoundation.h>
#else
#include <CoreServices/CoreServices.h>
#if !defined(MACOSX64)
#include <Carbon/Carbon.h>
#endif
#endif
#if defined(__MACH__)
#include <sys/sysctl.h>
#endif
#else
#include <Gestalt.h>
#include <MacTypes.h>
#include <Files.h>
#include <Processes.h>
#include <Sound.h>
#include <fcntl.h>
#include <Script.h>
#include <TextEncodingConverter.h>
#include <InternetConfig.h>
#endif
#endif

#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spString.h>
#include <sp/spMemory.h>
#include <sp/spOption.h>

#if defined(_WIN32) && !defined(__CYGWIN32__)
#include <windows.h>
#include <mbctype.h>
#include <mbstring.h>

#define access _access        
#define mkdir _mkdir
#define unlink _unlink
#ifndef getcwd
#define getcwd _getcwd
#endif
#define getpid _getpid
#endif

#ifdef NO_WARNING
static int sp_warn_flag = 0;
#else
static int sp_warn_flag = 1;
#endif

static int sp_message_flag = 1;
static int sp_debug_level = -1;
static int sp_debug_stdout = 0;

static spPrintFunc sp_print_func = NULL;


#if (defined(_WIN32) && !defined(__CYGWIN32__)) || defined(ANDROID)
static FILE *SP_STDIN = NULL;
static FILE *SP_STDOUT = NULL;
static FILE *SP_STDERR = NULL;
#else
#define SP_STDIN stdin
#define SP_STDOUT stdout
#if defined(MACOS9)
#define SP_STDERR stdout
#else
#define SP_STDERR stderr
#endif

#endif

void spsetstdio(FILE *in, FILE *out, FILE *err)
{
#if (defined(_WIN32) && !defined(__CYGWIN32__)) || defined(ANDROID)
    SP_STDIN = in;
    SP_STDOUT = out;
    SP_STDERR = err;
#endif
    
    return;
}

FILE *spgetstdin(void)
{
#if defined(_MSC_VER) && (_MSC_VER >= 1900/*VS2015*/)
    if (SP_STDIN == NULL) {
        return stdin;
    }
#endif
    return SP_STDIN;
}

FILE *spgetstdout(void)
{
#if defined(_MSC_VER) && (_MSC_VER >= 1900/*VS2015*/)
    if (SP_STDOUT == NULL) {
        return stdout;
    }
#endif
    return SP_STDOUT;
}

FILE *spgetstderr(void)
{
#if defined(_MSC_VER) && (_MSC_VER >= 1900/*VS2015*/)
    if (SP_STDERR == NULL) {
        return stderr;
    }
#endif
    return SP_STDERR;
}

int spvsnprintf(char *buf, size_t n, const char *format, va_list args)
{
    int ret;
    
#if defined(HAVE_VSPRINTF_S)
    ret = vsprintf_s(buf, n, format, args);
#elif defined(HAVE_VSNPRINTF)
    ret = vsnprintf(buf, n, format, args);
#else
    ret = vsprintf(buf, format, args);
#endif

    return ret;
}

void spSetPrintFunc(spPrintFunc func)
{
    sp_print_func = func;
    return;
}

void spPrint(const char *message, FILE *fp)
{
    if (message == NULL) return;
    
    if (sp_print_func != NULL) {
        sp_print_func(message);
    } else {
#if defined(ANDROID)
        __android_log_write(ANDROID_LOG_INFO, "spPrint", message);
#else
        if (fp != NULL) {
            fputs(message, fp);
        } else {
            if (SP_STDERR != NULL) {
                fputs(message, SP_STDERR);
            } else {
                printf("%s", message);
            }
        }
#endif
    }

    return;
}

void spSetWarningFlag(int flag)
{
    if (flag == 0) sp_warn_flag = 0;
    else sp_warn_flag = 1;
    return;
}

void spWarning(const char *format, ...)
{
    va_list args;
    char buf[SP_MAX_MESSAGE];

    if (!sp_warn_flag) return;
    
    if (SP_STDERR != NULL) {
        va_start(args, format);
        spvsnprintf(buf, sizeof(buf), format, args);
        va_end(args);

        spPrint(buf, SP_STDERR);
    }

    return;
}

void spSetMessageFlag(int flag)
{
    if (flag == 0) sp_message_flag = 0;
    else sp_message_flag = 1;
    return;
}

void spMessage(const char *format, ...)
{
    va_list args;
    char buf[SP_MAX_MESSAGE];

    if (!sp_message_flag) return;
    
    if (SP_STDERR != NULL) {
        va_start(args, format);
        spvsnprintf(buf, sizeof(buf), format, args);
        va_end(args);

        spPrint(buf, SP_STDERR);
    }

    return;
}

void spError(int status, const char *format, ...)
{
    va_list args;
    char buf[SP_MAX_MESSAGE];

    if (SP_STDERR != NULL) {
        va_start(args, format);
        spvsnprintf(buf, sizeof(buf), format, args);
        va_end(args);

        spPrint(buf, SP_STDERR);
    }

#if defined(MACOS)
#if !defined(MACOSX64) && !defined(IPHONE)
    SysBeep(30);
#endif
    spMSleep(500);
#endif
    spExit(status);
    
    return;
}

void spProgError(const char *func_name, const char *format, ...)
{
    va_list args;
    char buf[SP_MAX_MESSAGE];
    char buf2[SP_MAX_MESSAGE];

    if (SP_STDERR != NULL) {
        va_start(args, format);
        spvsnprintf(buf, sizeof(buf), format, args);
        va_end(args);

        if (!strnone(func_name)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf2, sizeof(buf2), "%s: %s", func_name, buf);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf2, sizeof(buf2), "%s: %s", func_name, buf);
#else
            sprintf(buf2, "%s: %s", func_name, buf);
#endif
            spPrint(buf2, SP_STDERR);
        } else {
            spPrint(buf, SP_STDERR);
        }
    }

#if defined(MACOS)
#if !defined(MACOSX64) && !defined(IPHONE)
    SysBeep(30);
#endif
    spMSleep(500);
#endif
    spExit(1);
    
    return;
}

void spSetDebugStdout(int flag)
{
    if (flag == 0) sp_debug_stdout = 0;
    else sp_debug_stdout = 1;
    return;
}

void spSetDebugLevel(int level)
{
    sp_debug_level = level;
    return;
}

int spGetDebugLevel(void)
{
    return sp_debug_level;
}

void spDebug(int level, const char *func_name, const char *format, ...)
{
    va_list args;
    char buf[SP_MAX_MESSAGE];
    char buf2[SP_MAX_MESSAGE];
    char *ptr;

    if (sp_debug_level < level) return;
    
    va_start(args, format);
    spvsnprintf(buf, sizeof(buf), format, args);
    va_end(args);

#if defined(ANDROID)
    if (!strnone(func_name)) {
        __android_log_write(ANDROID_LOG_INFO, func_name, buf);
    } else {
        __android_log_write(ANDROID_LOG_INFO, "-", buf);
    }
#else /* defined(ANDROID) */
    if (!strnone(func_name)) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf2, sizeof(buf2), "%s: %s", func_name, buf);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf2, sizeof(buf2), "%s: %s", func_name, buf);
#else
        sprintf(buf2, "%s: %s", func_name, buf);
#endif
        ptr = buf2;
    } else {
        ptr = buf;
    }

    if (sp_debug_stdout) {
        if (SP_STDOUT != NULL) {
            spPrint(ptr, SP_STDOUT);
        }
    } else {
        if (SP_STDERR != NULL) {
            spPrint(ptr, SP_STDERR);
        }
    }
#endif /* !defined(ANDROID) */
    
    return;
}

void spSleep(int sec)
{
#if defined(_WIN32) && !defined(__CYGWIN32__)
    SleepEx((DWORD)sec * (DWORD)1000, FALSE);
#elif defined(MACOS9)
    {
        unsigned long ticks;
        Delay(60 * (unsigned long)sec, &ticks);
    }
#else
    sleep(sec);
#endif
    return;
}

void spMSleep(int msec)
{
#if defined(_WIN32) && !defined(__CYGWIN32__)
    Sleep((DWORD)msec);
#elif defined(MACOS9)
    {
        unsigned long ticks;
        unsigned long dtime;
        dtime = (unsigned long)spRound(60.0 * (double)msec / 1000.0);
        Delay(dtime, &ticks);
    }
#else
    usleep(msec * 1000);
#endif
    return;
}

long spGetProcessId(void)
{
#if defined(MACOS9)
    Str32 pname;
    FSSpec fsspec;
    ProcessSerialNumber psn;
    ProcessInfoRec pir;

    psn.highLongOfPSN = 0;
    psn.lowLongOfPSN = kCurrentProcess;
    pir.processInfoLength = sizeof(pir);
    pir.processName = pname;
    pir.processAppSpec = &fsspec;
    
    if (GetProcessInformation(&psn, &pir) != noErr) {
        return 0;
    }

    return (long)pir.processNumber.lowLongOfPSN;
#else
    return (long)getpid();
#endif
}

#if defined(_WIN32)
spBool spExecBrowser(const char *base_url, const char *sub_url, spBool local_path)
{
    HWND hwnd;
    char buf[SP_MAX_IE_URL];
    
    if (strnone(base_url)) return SP_FALSE;

    if ((hwnd = GetActiveWindow()) == NULL) {
        hwnd = GetDesktopWindow();
    }

    spStrCopy(buf, sizeof(buf), base_url);
    
    if (!strnone(sub_url)) {
        spRemoveDirSeparator(buf);
        spStrCat(buf, sizeof(buf), "/");
        spStrCat(buf, sizeof(buf), sub_url);
    }
    spDebug(10, "spExecBrowser", "buf = %s\n", buf);
    
    if (ShellExecute(hwnd, NULL, buf, NULL, NULL, SW_SHOWNORMAL) <= (HINSTANCE)32) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}
#elif defined(IPHONE)
spBool spExecBrowser(const char *base_url, const char *sub_url, spBool local_path)
{
    return SP_FALSE;
}
#elif defined(MACOS)
spBool spExecBrowser(const char *base_url, const char *sub_url, spBool local_path)
{
    OSStatus err = noErr;
    
    if (strnone(base_url)) return SP_FALSE;
    
    err = noErr + 1;
        
#if defined(MACOSX)
    if (
#if !defined(MACOSX64)
        LSInit(kLSInitializeDefaults) == noErr
#else
        1
#endif
        ) {
        char *path;
        CFStringRef urlstrref; 
        CFURLRef urlref;
        CFURLRef urlbaseref;
        char buf[SP_MAX_MESSAGE];
        
        spDebug(10, "spExecBrowser", "base_url = %s\n", base_url);

        path = NULL;
        urlbaseref = NULL;
        
        if (local_path == SP_FALSE) {
            spStrCopy(buf, sizeof(buf), base_url);
            if (!strnone(sub_url)) {
                spRemoveDirSeparator(buf);
                spStrCat(buf, sizeof(buf), "/");
                spStrCat(buf, sizeof(buf), sub_url);
            }
            path = buf;
            urlbaseref = NULL;
            err = noErr;
        } else {
            CFStringRef urlbasestrref;

            if ((urlbasestrref = CFStringCreateWithCString(kCFAllocatorDefault,
                                                           base_url, kCFStringEncodingUTF8)) != NULL) {
                if ((urlbaseref = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
                                                                urlbasestrref, kCFURLPOSIXPathStyle, true)) != NULL) {
                    path = (char *)sub_url;
                    err = noErr;
                }
                CFRelease(urlbasestrref);
            }
        }

        if (err == noErr) {
            err = noErr + 1;
            urlref = NULL;
            
            if (!strnone(path)) {
                if ((urlstrref = CFStringCreateWithCString(kCFAllocatorDefault,
                                                           path, kCFStringEncodingUTF8)) != NULL) {
                    urlref = CFURLCreateWithString(kCFAllocatorDefault, urlstrref, urlbaseref);
                    if (urlbaseref != NULL) {
                        CFRelease(urlbaseref);
                    }
                    CFRelease(urlstrref);
                }
            } else {
                urlref = urlbaseref;
            }

            if (urlref != NULL) {
                err = LSOpenCFURLRef(urlref, NULL);
                CFRelease(urlref);
            }
        }
#if !defined(MACOSX64)
        LSTerm();
#endif
    }
#else
    {
        ICInstance inst;
        
        if ((err = ICStart(&inst, '????')) == noErr) {
            spDebug(10, "spExecBrowser", "ICStart finished\n");
#if defined(MACOS9)/*!TARGET_API_MAC_CARBON*/
            err = ICFindConfigFile(inst, 0, nil);
#else
            err = noErr;
#endif
            spDebug(10, "spExecBrowser", "ICFindConfigFile: err = %d\n", err);
        
            if (err == noErr) {
                char url[SP_MAX_MESSAGE];
                SInt32 start, end;

                if (local_path == SP_TRUE) {
                    spStrCopy(url, sizeof(url), "file://");
                    spStrCat(url, sizeof(url), base_url);
                } else {
                    spStrCopy(url, sizeof(url), base_url);
                }
                
                if (!strnone(sub_url)) {
                    spRemoveDirSeparator(url);
                    spStrCat(url, sizeof(url), "/");
                    spStrCat(url, sizeof(url), sub_url);
                }
                
                spDebug(10, "spExecBrowser", "url = %s\n", url);
                start = 0;
                end = (SInt32)strlen(url);
                err = ICLaunchURL(inst, "\p", url, end, &start, &end);
            }
            ICStop(inst);
        }
    }
#endif

    if (err == noErr) {
        return SP_TRUE;
    } else {
        spDebug(10, "spExecBrowser", "error: %d\n", err);
        return SP_FALSE;
    }
}
#else
typedef struct _spBrowserCommand {
    char *exefile;
    char *command;
    char *remote_command;
    char *file_prefix;
} spBrowserCommand;

spBool spExecBrowser(const char *base_url, const char *sub_url, spBool local_path)
{
    int i, j;
    char *string;
    char *path;
    char *exactname;
    char exefile[SP_MAX_PATHNAME], url_buf[SP_MAX_IE_URL];
    char buf[2304], buf2[2560];
    static spBrowserCommand command_list[] = {
        {"kfmclient", "openURL \"%s\"", "openURL \"%s\"", "file://"},
        /*{"firefox", "\"%s\"", "-remote \"openURL(%s)\"", "file://"},*/
        {"firefox", "-new-tab \"%s\"", NULL, "file://"},
        {"mozilla", "\"%s\"", "-remote \"openURL(%s)\"", "file://"},
#if 0
        {"netscape", "\"%s\"", "-remote \"openURL(%s)\"", "file://"},
#endif
        {NULL, NULL, NULL, NULL},
    };

    if (strnone(base_url)) return SP_FALSE;
    
    if ((string = getenv(SP_BROWSER_COMMAND_ENV)) != NULL) {
        if (local_path == SP_TRUE) {
            exactname = xspGetExactName(base_url);
            spStrCopy(url_buf, sizeof(url_buf), exactname);
            xfree(exactname);
        } else {
            spStrCopy(url_buf, sizeof(url_buf), base_url);
        }
        if (!strnone(sub_url)) {
            spRemoveDirSeparator(url_buf);
            spStrCat(url_buf, sizeof(url_buf), "/");
            spStrCat(url_buf, sizeof(url_buf), sub_url);
        }
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, sizeof(buf), string, url_buf);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, sizeof(buf), string, url_buf);
#else
        sprintf(buf, string, url_buf);
#endif
        spStrCat(buf, sizeof(buf), " &");
        if (system(buf) == 0) {
            return SP_TRUE;
        }
    }

    for (i = 0;; i++) {
        if (strnone(command_list[i].exefile)) {
            break;
        }
        for (j = 0;; j++) {
            if ((path = xspCutPathList(SP_UNIX_DEFAULT_PATH_LIST, j)) == NULL) {
                break;
            }
#if defined(HAVE_SPRINTF_S)
            sprintf_s(exefile, sizeof(exefile), "%s%c%s", path, SP_DIR_SEPARATOR, command_list[i].exefile);
#elif defined(HAVE_SNPRINTF)
            snprintf(exefile, sizeof(exefile), "%s%c%s", path, SP_DIR_SEPARATOR, command_list[i].exefile);
#else
            sprintf(exefile, "%s%c%s", path, SP_DIR_SEPARATOR, command_list[i].exefile);
#endif
            xfree(path);

            spDebug(50, "spExecBrowser", "exefile = %s\n", exefile);
            
            if (spIsFile(exefile) == SP_TRUE) {
                if (local_path == SP_TRUE) {
                    if (!strnone(command_list[i].file_prefix)) {
                        spStrCopy(url_buf, sizeof(url_buf), command_list[i].file_prefix);
                    } else {
                        spStrCopy(url_buf, sizeof(url_buf), "");
                    }
                    exactname = xspGetExactName(base_url);
                    spStrCat(url_buf, sizeof(url_buf), exactname);
                    xfree(exactname);
                } else {
                    spStrCopy(url_buf, sizeof(url_buf), base_url);
                }
                if (!strnone(sub_url)) {
                    spRemoveDirSeparator(url_buf);
                    spStrCat(url_buf, sizeof(url_buf), "/");
                    spStrCat(url_buf, sizeof(url_buf), sub_url);
                }

                if (!strnone(command_list[i].remote_command)) {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(buf, sizeof(buf), "%s %s", exefile, command_list[i].remote_command);
                    sprintf_s(buf2, sizeof(buf2), buf, url_buf);
#elif defined(HAVE_SNPRINTF)
                    snprintf(buf, sizeof(buf), "%s %s", exefile, command_list[i].remote_command);
                    snprintf(buf2, sizeof(buf2), buf, url_buf);
#else
                    sprintf(buf, "%s %s", exefile, command_list[i].remote_command);
                    sprintf(buf2, buf, url_buf);
#endif
                } else {
                    spStrCopy(buf2, sizeof(buf2), "");
                }
                spDebug(50, "spExecBrowser", "buf2 = %s\n", buf2);
                
                if (!strnone(buf2) && system(buf2) == 0) {
                    return SP_TRUE;
                } else {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(buf, sizeof(buf), "%s %s &", exefile, command_list[i].command);
                    sprintf_s(buf2, sizeof(buf), buf, url_buf);
#elif defined(HAVE_SNPRINTF)
                    snprintf(buf, sizeof(buf), "%s %s &", exefile, command_list[i].command);
                    snprintf(buf2, sizeof(buf), buf, url_buf);
#else
                    sprintf(buf, "%s %s &", exefile, command_list[i].command);
                    sprintf(buf2, buf, url_buf);
#endif
                    spDebug(50, "spExecBrowser", "non-remote command: buf2 = %s\n", buf2);
                    if (system(buf2) == 0) {
                        return SP_TRUE;
                    }
                        
                }
            }
        }
    }

    return SP_FALSE;
}
#endif

int spStrCaseCmp(const char *s1, const char *s2)
{
    int c1, c2;
    
    while (*s1 != NUL && *s2 != NUL) {
        c1 = isupper((unsigned char)*s1) ? tolower((unsigned char)*s1) : *s1;
        c2 = isupper((unsigned char)*s2) ? tolower((unsigned char)*s2) : *s2;
        if (c1 != c2) {
            return (c1 - c2);
        }
        s1++;
        s2++;
    }
    
    return (((int)(unsigned char)*s1) - ((int)(unsigned char)*s2));
}

int spStrNCaseCmp(const char *s1, const char *s2, int n)
{
    int i;
    int c1, c2;
    
    for (i = 1; *s1 != NUL && *s2 != NUL; i++) {
        c1 = isupper((unsigned char)*s1) ? tolower((unsigned char)*s1) : *s1;
        c2 = isupper((unsigned char)*s2) ? tolower((unsigned char)*s2) : *s2;
        if (c1 != c2 || i >= n) {
            return (c1 - c2);
        }
        s1++;
        s2++;
    }
    
    return (((int)(unsigned char)*s1) - ((int)(unsigned char)*s2));
}

int spStrWhite(const char *string)
{
    int i;

    if (strnone(string)) return 0;
    
    for (i = (int)strlen(string) - 1; i >= 0; i--) {
        if (!(string[i] == ' ' || string[i] == '\t')) {
            return 0;
        }
    }

    return 1;
}

char *spSkipWhite(const char *string)
{
    int i;

    for (i = 0; string[i] != NUL; i++) {
        if (!(string[i] == ' ' || string[i] == '\t')) {
            break;
        }
    }

    return (char *)&string[i];
}

char *xspStrCat(const char *orig, const char *src)
{
    int len;
    char *result;

    if (strnone(orig)) {
        return strclone(src);
    }

    len = (int)(strlen(orig) + strlen(src)) + 1;
    result = xalloc(len, char);
#if defined(HAVE_STRCPY_S) && defined(HAVE_STRCAT_S)
    strcpy_s(result, len, orig);
    strcat_s(result, len, src);
#else
    strcpy(result, orig);
    strcat(result, src);
#endif
    
    return result;
}

char *spStrCat(char *dest, int destsize, const char *src)
{
    int n;
    char *result;
    
    if (dest == NULL) return NULL;

    if (strnone(src)) {
        result = dest;
    } else {
        n = destsize - (int)strlen(dest) - 1;
        if ((int)strlen(src) < n) {
#if defined(HAVE_STRCAT_S)
            if (strcat_s(dest, destsize, src) != 0) {
                result = NULL;
            } else {
                result = dest;
            }
#else
            result = strcat(dest, src);
#endif
        } else {
#if defined(HAVE_STRNCAT_S)
            if (strncat_s(dest, destsize, src, n) != 0) {
                result = NULL;
            } else {
                result = dest;
            }
#else
            result = strncat(dest, src, n);
#endif
            dest[destsize - 1] = NUL;
        }
    }
    
    return result;
}

char *spStrCopy(char *dest, int destsize, const char *src)
{
    if (dest == NULL || destsize <= 0) return NULL;

    if (strnone(src) || destsize <= 1) {
        dest[0] = NUL;
    } else {
        if ((int)strlen(src) < destsize) {
#if defined(HAVE_STRCPY_S)
            strcpy_s(dest, destsize, src);
#else
            strcpy(dest, src);
#endif
        } else {
            destsize--;
#if defined(HAVE_STRNCPY_S)
            strncpy_s(dest, destsize + 1, src, destsize);
#else
            strncpy(dest, src, destsize);
#endif
            dest[destsize] = NUL;
        }
    }
    
    return dest;
}

char *spGetSuffix(const char *name)
{
    char *p1, *p2;
    
    if ((p1 = spGetBaseName(name)) == NULL || (p2 = spStrRChr(p1, '.')) == NULL) {
        return NULL;
    }

    return p2;
}

int spEqSuffix(const char *file1, const char *file2)
{
    char *name;
    char *p1, *p2;

    if (file1 == NULL || file2 == NULL) return 0;
    
    if ((name = spGetBaseName(file1)) == NULL || (p1 = spStrRChr(name, '.')) == NULL) {
        return 0;
    } else {
        p1++;
    }
    if ((name = spGetBaseName(file2)) == NULL || (p2 = spStrRChr(name, '.')) == NULL) {
        return 0;
    } else {
        p2++;
    }

    return strcaseeq(p1, p2);
}

spBool spRemoveNSuffix(char *path, char *removed_suffix, int suffixsize)
{
    char *name;
    char *p;
    
    if (path == NULL) return SP_FALSE;

    if ((name = spGetBaseName(path)) != NULL
        && (p = spStrRChr(name, '.')) != NULL) {
        if (removed_suffix != NULL) {
            spStrCopy(removed_suffix, suffixsize, p);
        }
        *p = NUL;
    } else {
        if (removed_suffix != NULL) {
            spStrCopy(removed_suffix, suffixsize, "");
        }
    }

    return SP_TRUE;
}

spBool spRemoveSuffix(char *path, char *removed_suffix)
{
    return spRemoveNSuffix(path, removed_suffix, SP_MAX_PATHNAME);
}

spBool spReplaceNSuffix(char *path, int pathsize, const char *suffix)
{
    int len = 0;

    if (strnone(path)) return SP_FALSE;

    if (!strnone(suffix)) {
        len = (int)strlen(suffix);
        
        if (len > 0 && suffix[len-1] != '*') {
            const char *p;
        
            spRemoveSuffix(path, NULL);

            p = suffix;
            if (p[0] == '*') {
                p++;
            }
            spStrCat(path, pathsize, p);
        }
    } else {
        spRemoveSuffix(path, NULL);
    }

    return SP_TRUE;
}

spBool spReplaceSuffix(char *path, const char *suffix)
{
    return spReplaceNSuffix(path, SP_MAX_PATHNAME, suffix);
}

/* return: number of insertion, -1 is error. */
int spConvertToEscapedString(char *escaped, int bufsize, const char *string, const char *charlist)
{
    int i, j, n;
    int prev_c, c;
    int count = 0;

    if (escaped == NULL || bufsize < 1 || strnone(string)
        || strnone(charlist)) return -1;

    prev_c = 0;
    for (i = 0, j = 0;; i++) {
        c = string[i];
        if (c == NUL || j >= bufsize - 1) {
            escaped[j] = NUL;
            break;
        }

        if (spIsMBTailCandidate(prev_c, c) == SP_FALSE) {
            for (n = 0; charlist[n] != NUL; n++) {
                if (c == charlist[n]) {
                    escaped[j] = '\\';
                    j++;

                    count++;
                    break;
                }
            }
            escaped[j] = c;
        } else {
            /* MB tail */
            escaped[j] = c;
            c = 0;
        }
        j++;
        
        prev_c = c;
    }

    return count;
}

/* return: number of deletion, -1 is error.  charlist can be NULL. */
int spConvertFromEscapedString(char *string, int bufsize, const char *escaped, const char *charlist)
{
    int i, j, n;
    int prev_c, c, next_c;
    int count = 0;

    if (string == NULL || bufsize < 1 || strnone(escaped)) return -1;

    prev_c = 0;
    for (i = 0, j = 0;; i++) {
        c = escaped[i];
        if (c == NUL || j >= bufsize - 1) {
            string[j] = NUL;
            break;
        }

        if (c == '\\' && spIsMBTailCandidate(prev_c, c) == SP_FALSE) {
            if (strnone(charlist)) {
                i++;
                c = escaped[i];
                count++;
            } else {
                next_c = escaped[i + 1];
                for (n = 0; charlist[n] != NUL; n++) {
                    if (next_c == charlist[n]) {
                        i++;
                        c = next_c;
                        count++;
                        break;
                    }
                }
            }
        }

        string[j] = c;
        j++;

        if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
            prev_c = 0;
        } else {
            prev_c = c;
        }
    }

    return count;
}

/* for Mac string conversion */
void spStrPToC(const unsigned char *in_p, char *out_c)
{
    unsigned char len;

    if (in_p == NULL || out_c == NULL) return;
    
    len = *in_p;
    
    if (len > 0) {
        memmove(out_c, in_p + 1, len);
    }
    *(out_c + len) = '\0';

    return;
}

void spStrCToP(const char *in_c, unsigned char *out_p)
{
    int len;
    
    if (in_c == NULL || out_p == NULL) return;
    
    len = (int)strlen(in_c);
    len = MIN(len, 255);
    
    *out_p = (unsigned char)len;
    
    if (len > 0) {
        memmove(out_p + 1, in_c, len);
    }
    
    return;
}

void spCopyPStr(const unsigned char *source, unsigned char *dest)
{
    unsigned char len;

    if (source == NULL || dest == NULL) return;
        
    len = source[0];
    if (len > 0) {
        memmove(dest + 1, source + 1, len);
    }
    dest[0] = len;

    return;
}

void spCatPStr(unsigned char *target, const unsigned char *source)
{
    unsigned char len;
    
    if (target == NULL || source == NULL) return;
    
    len = MIN(source[0], 255 - target[0]);
    
    if (len > 0) {
        memmove(target + target[0] + 1, source + 1, len);
        target[0] += len;
    }

    return;
}

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
static spBool spGetApplicationFInfoMac(FInfo *finfo)
{
    spBool flag = SP_FALSE;
    
    if (finfo == NULL) return SP_FALSE;

#if defined(MACOSX)
    {
        CFBundleRef mainBundle;
        
        if ((mainBundle = CFBundleGetMainBundle()) != NULL) {
            finfo->fdFlags = 0;
            finfo->fdFldr = 0;
            CFBundleGetPackageInfo(mainBundle, (UInt32 *)&finfo->fdType, (UInt32 *)&finfo->fdCreator);

            flag = SP_TRUE;
        }
    }
#endif

    if (flag == SP_FALSE) {
        Str255 ppath;
        FSSpec fsspec_process;
        ProcessSerialNumber psn;
        ProcessInfoRec pir;
        
        psn.highLongOfPSN = 0;
        psn.lowLongOfPSN = kCurrentProcess;
        pir.processInfoLength = sizeof(pir);
        pir.processName = ppath;
        pir.processAppSpec = &fsspec_process;
        
        if (GetProcessInformation(&psn, &pir) == noErr) {
            if (FSpGetFInfo(&fsspec_process, finfo) == noErr) {
                flag = SP_TRUE;
            }
        }
    }

    return flag;
}

spBool spPathToFSSpecMac(const char *filename, FSSpec *fsspec)
{
    if (strnone(filename)) return SP_FALSE;

#if defined(MACOSX)
    {
        FSRef fsref;
        Boolean is_dir;

        if (FSPathMakeRef(filename, &fsref, &is_dir) != noErr
            || FSGetCatalogInfo(&fsref, kFSCatInfoNone, NULL, NULL, fsspec, NULL) != noErr) {
            return SP_FALSE;
        }
    }
#else
    {
        Str255 ppath;

        spStrCToP(filename, ppath);
        if (FSMakeFSSpec(0, 0L, ppath, fsspec) != noErr) {
            return SP_FALSE;
        }
    }
#endif
    return SP_TRUE;
}
#endif

spBool spGetMacFileInfo(const char *filename, unsigned long *creator, unsigned long *file_type)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    FSSpec fsspec;
    FInfo finfo;

    if (filename == NULL) {
        if (spGetApplicationFInfoMac(&finfo) == SP_FALSE) {
            return SP_FALSE;
        }
    } else {
        if (spPathToFSSpecMac(filename, &fsspec) == SP_FALSE) {
            return SP_FALSE;
        }
        if (FSpGetFInfo(&fsspec, &finfo) != noErr) {
            return SP_FALSE;
        }
    }
    if (creator != NULL) {
        *creator = (unsigned long)finfo.fdCreator;
    }
    if (file_type != NULL) {
        *file_type = (unsigned long)finfo.fdType;
    }
    
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

spBool spSetMacFileInfo(const char *filename, unsigned long creator, unsigned long file_type)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    FSSpec fsspec;
    FInfo finfo;
    static unsigned long application_creator = 0;

    if (spPathToFSSpecMac(filename, &fsspec) == SP_FALSE) {
        return SP_FALSE;
    }
    if (FSpGetFInfo(&fsspec, &finfo) != noErr) {
        return SP_FALSE;
    }

    if (creator == 0) {
        if (application_creator == 0) {
            FInfo finfo_process;
            
            if (spGetApplicationFInfoMac(&finfo_process) == SP_TRUE) {
                application_creator = finfo_process.fdCreator;
            }
        }
        creator = application_creator;
    }
    
    finfo.fdCreator = (OSType)creator;
#if 0
    spDebug(10, "spSetMacFileInfo", "creator: %c%c%c%c\n",
            (char)(finfo.fdCreator>>24),
            (char)(finfo.fdCreator>>16),
            (char)(finfo.fdCreator>>8),
            (char)(finfo.fdCreator));
#endif
    
    if (file_type != 0) {
        finfo.fdType = (OSType)file_type;
    }
#if 0
    spDebug(10, "spSetMacFileInfo", "file type: %c%c%c%c\n",
            (char)(finfo.fdType>>24),
            (char)(finfo.fdType>>16),
            (char)(finfo.fdType>>8),
            (char)(finfo.fdType));
#endif
    
    if (FSpSetFInfo(&fsspec, &finfo) != noErr) {
        return SP_FALSE;
    }

    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

FILE *spOpenFile(const char *path, const char *mode)
{
    FILE *fp;
    char *basename;

    if (strnone(path) || strnone(mode)) return NULL;

    basename = spGetBaseName(path);

    if ((mode[0] == 'r' && streq(basename, "-"))
        || streq(basename, "stdin")) {
        fp = SP_STDIN;
    } else if ((mode[0] == 'w' && streq(basename, "-"))
               || streq(basename, "stdout")) {
        fp = SP_STDOUT;
    } else {
        /* open file */
#if defined(HAVE_FOPEN_S)
        errno_t err;
        fp = NULL;
        if ((err = fopen_s(&fp, path, mode)) != 0) {
            spWarning("Can't open file: %s (error code = %d)\n", path, (int)err);
            return NULL;
        }
#else
        if (NULL == (fp = fopen(path, mode))) {
            spWarning("Can't open file: %s\n", path);
            return NULL;
        }
#endif
    }

    return fp;
}

int spCloseFile(FILE *stream)
{
    if (stream == SP_STDIN || stream == SP_STDOUT) return 0;
    
    return fclose(stream);
}

int spSeekFile(FILE *stream, spFileOffset offset, int origin)
{
    int status;
    
#if (defined(_WIN32) && !defined(__CYGWIN32__)) && ((defined(_MSC_VER) && _MSC_VER >= 1400/*VS2005*/) || defined(__MINGW32__))
    status = _fseeki64(stream, offset, origin);
#elif defined(HAVE_FSEEKO)
    status = fseeko(stream, offset, origin);
#else
    status = fseek(stream, offset, origin);
#endif
    
    return status;
}

spFileOffset spTellFile(FILE *stream)
{
    spFileOffset offset;
    
#if (defined(_WIN32) && !defined(__CYGWIN32__)) && ((defined(_MSC_VER) && _MSC_VER >= 1400/*VS2005*/) || defined(__MINGW32__))
    offset = _ftelli64(stream);
#elif defined(HAVE_FSEEKO)
    offset = ftello(stream);
#else
    offset = ftell(stream);
#endif

    return offset;
}

#ifdef SP_SUPPORT_64BIT_INT
int spSeekFile64(FILE *stream, spLong64 offset, int origin)
{
    int status;
#if (defined(_WIN32) && !defined(__CYGWIN32__)) && ((defined(_MSC_VER) && _MSC_VER >= 1400/*VS2005*/) || defined(__MINGW32__))
    status = _fseeki64(stream, offset, origin);
#elif defined(HAVE_FSEEKO)
    status = fseeko(stream, offset, origin);
#else
    status = fseek(stream, (long)offset, origin);
#endif

    return status;
}
#endif

int spSeekFileC64(FILE *stream, char offset_c64[8], int swap_endian, int origin)
{
    int status;
    char offset2_c64[8];
    
    memcpy(offset2_c64, offset_c64, 8);
    
    if (swap_endian) {
        spSwapLongC64(offset2_c64, 1);
    }

#if (defined(_WIN32) && !defined(__CYGWIN32__)) && ((defined(_MSC_VER) && _MSC_VER >= 1400/*VS2005*/) || defined(__MINGW32__))
    {
        spLong64 offset;
        
        memcpy(&offset, offset2_c64, sizeof(spLong64));
        
        status = _fseeki64(stream, offset, origin);
    }
#elif defined(HAVE_FSEEKO)
    {
        off_t offset;
        if (sizeof(off_t) == 8) {
            memcpy(&offset, offset2_c64, sizeof(off_t));
        } else {
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(&offset, offset2_c64, sizeof(off_t));
#else
            memcpy(&offset, &offset2_c64[4], sizeof(off_t));
#endif
        }
        
        status = fseeko(stream, offset, origin);
    }
#else
    {
        long offset;
        
        if (sizeof(long) == 8) {
            memcpy(&offset, offset2_c64, sizeof(long));
        } else {
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(&offset, offset2_c64, sizeof(long));
#else
            memcpy(&offset, &offset2_c64[8 - sizeof(long)], sizeof(long));
#endif
        }
        
        status = fseek(stream, offset, origin);
    }
#endif

    return status;
}

void spAddC64(char offset_c64[8], long value, int swap_endian)
{
    char offset2_c64[8];
        
    memcpy(offset2_c64, offset_c64, 8);
    
    if (swap_endian) {
        spSwapLongC64(offset2_c64, 1);
    }
    
#if (defined(_WIN32) && !defined(__CYGWIN32__)) && ((defined(_MSC_VER) && _MSC_VER >= 1400/*VS2005*/) || defined(__MINGW32__))
    {
        spLong64 offset;
        memcpy(&offset, offset2_c64, sizeof(spLong64));
        offset += (spLong64)value;
        memcpy(offset2_c64, &offset, sizeof(spLong64));
    }
#elif defined(HAVE_FSEEKO)
    {
        off_t offset;
        if (sizeof(off_t) == 8) {
            memcpy(&offset, offset2_c64, sizeof(off_t));
            if (value < 0) {
                offset -= (off_t)(-value);
            } else {
                offset += (off_t)value;
            }
            memcpy(offset2_c64, &offset, sizeof(off_t));
        } else {
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(&offset, offset2_c64, sizeof(off_t));
#else
            memcpy(&offset, &offset2_c64[4], sizeof(off_t));
#endif
            
            if (value < 0) {
                offset -= (off_t)(-value);
            } else {
                offset += (off_t)value;
            }
            
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(offset2_c64, &offset, sizeof(off_t));
#else
            memcpy(&offset2_c64[4], &offset, sizeof(off_t));
#endif
        }
    }
#else
    {
        long offset;
        
        if (sizeof(long) == 8) {
            memcpy(&offset, offset2_c64, sizeof(long));
            offset += value;
            memcpy(offset2_c64, &offset, sizeof(long));
        } else {
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(&offset, offset2_c64, sizeof(long));
#else
            memcpy(&offset, &offset2_c64[8 - sizeof(long)], sizeof(long));
#endif
            
            offset += value;
            
#if BYTE_ORDER == LITTLE_ENDIAN
            memcpy(offset2_c64, &offset, sizeof(long));
#else
            memcpy(&offset2_c64[8 - sizeof(long)], &offset, sizeof(long));
#endif
        }
    }
#endif
    
    if (swap_endian) {
        spSwapLongC64(offset2_c64, 1);
    }
    
    memcpy(offset_c64, offset2_c64, 8);
    
    return;
}

spULong spC64ToULong(char value_c64[8], int swap_endian)
{
    char value2_c64[8];
    spULong value;
        
    memcpy(value2_c64, value_c64, 8);
    
    if (swap_endian) {
        spSwapLongC64(value2_c64, 1);
    }
    
    if (sizeof(spULong) == 8) {
        memcpy(&value, value2_c64, sizeof(spULong));
    } else {
#if BYTE_ORDER == LITTLE_ENDIAN
        memcpy(&value, value2_c64, sizeof(spULong));
#else
        memcpy(&value, &value2_c64[4], sizeof(spULong));
#endif
    }

    return value;
}

void spULongToC64(spULong value, char value_c64[8], int swap_endian)
{
    if (sizeof(spULong) == 8) {
        memcpy(value_c64, &value, sizeof(spULong));
    } else {
#if BYTE_ORDER == LITTLE_ENDIAN
        memset(&value_c64[4], 0, 4);
        memcpy(value_c64, &value, sizeof(spULong));
#else
        memset(value_c64, 0, 4);
        memcpy(&value_c64[4], &value, sizeof(spULong));
#endif
    }
    
    if (swap_endian) {
        spSwapLongC64(value_c64, 1);
    }
    
    return;
}

#ifdef SP_SUPPORT_64BIT_INT
spLong64 spC64ToLong64(char value_c64[8], int swap_endian)
{
    char value2_c64[8];
    spLong64 value;
        
    memcpy(value2_c64, value_c64, 8);
    
    if (swap_endian) {
        spSwapLongC64(value2_c64, 1);
    }
    
    memcpy(&value, value2_c64, sizeof(spLong64));

    return value;
}

void spLong64ToC64(spLong64 value, char value_c64[8], int swap_endian)
{
    memcpy(value_c64, &value, sizeof(spLong64));
    
    if (swap_endian) {
        spSwapLongC64(value_c64, 1);
    }

    return;
}
#endif

#if !defined(MACOS9)
#ifndef S_ISDIR
#define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
#endif

#define SP_AVOID_WIN32_STAT

static spBool getFileInfo(const char *path, spBool *dir_flag, spFileOffset *size)
{
#if defined(_WIN32) && !defined(__CYGWIN32__)
#if defined(SP_AVOID_WIN32_STAT)
    WIN32_FILE_ATTRIBUTE_DATA fad;
#else
    struct _stat status;
#define stat(path,buf) _stat(path,buf)
#endif
#else
    struct stat status;
#endif
    
    if (strnone(path)) return SP_FALSE;

#if defined(_WIN32) && !defined(__CYGWIN32__) && defined(SP_AVOID_WIN32_STAT)
    if (GetFileAttributesEx(path, GetFileExInfoStandard, &fad)) {
        spDebug(50, "spGetFileInfo", "GetFileAttributesEx OK\n");
        
        if (size != NULL) {
            LARGE_INTEGER lisize;
            lisize.HighPart = fad.nFileSizeHigh;
            lisize.LowPart = fad.nFileSizeLow;
            *size = lisize.QuadPart;                
        }

        if (dir_flag != NULL) {
            if (fad.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                *dir_flag = SP_TRUE;
            } else {
                *dir_flag = SP_FALSE;
            }
        }
        
        return SP_TRUE;
    }
    spDebug(50, "spGetFileInfo", "GetFileAttributesEx failed\n");
#else
    if (stat(path, &status) == 0) {
        spDebug(50, "spGetFileInfo", "stat done\n");
        if (dir_flag != NULL) {
            if (S_ISDIR(status.st_mode)) {
#if defined(MACOSX)
                if (spEqSuffix(path, "*.app")
                    || spEqSuffix(path, "*.bundle")
                    || spEqSuffix(path, "*.pbproj")
                    || spEqSuffix(path, "*.xcodeproj")
                    || spEqSuffix(path, "*.plugin")) {
                    /* a bundle of Mac OS X */
                    *dir_flag = SP_FALSE;
                    spDebug(50, "spGetFileInfo", "bundle: %s\n", path);
                } else {
                    *dir_flag = SP_TRUE;
                }
#else
                *dir_flag = SP_TRUE;
#endif
            } else {
                *dir_flag = SP_FALSE;
            }
        }
        if (size != NULL) {
            *size = status.st_size;
        }
        
        return SP_TRUE;
    }
    spDebug(50, "spGetFileInfo", "stat failed\n");
#endif

    return SP_FALSE;
}
#endif

#if defined(MACOS9)
spBool spGetFileInfo(const char *path, spBool *dir_flag, spFileOffset *size)
{
    Str255 ppath;
    Str255 pname;
    HFileInfo file_info;
    HVolumeParam vol_param;

    spStrCToP(path, ppath);

    file_info.ioNamePtr = ppath;
    file_info.ioFDirIndex = 0;
    file_info.ioVRefNum = 0;
    file_info.ioDirID = 0L;

    if (PBGetCatInfoSync((CInfoPBPtr)&file_info) == noErr) {
        spDebug(50, "spGetFileInfo", "PBGetCatInfoSync done\n");
        vol_param.ioVolIndex = 0;
        vol_param.ioNamePtr = pname;
        vol_param.ioVRefNum = 0;
        
        if (PBHGetVInfoSync((HParmBlkPtr)&vol_param) == noErr) {
            spDebug(50, "spGetFileInfo", "PBHGetVInfoSync done\n");
            if (dir_flag != NULL) {
                if (file_info.ioFlAttrib & kioFlAttribDirMask) {
                    *dir_flag = SP_TRUE;
                } else {
                    *dir_flag = SP_FALSE;
                }
            }
            if (size != NULL) {
                *size = file_info.ioFlLgLen;
            }
            
            return SP_TRUE;
        }
    }
    spDebug(50, "spGetFileInfo", "failed\n");
#if !defined(MACOS9)
    return getFileInfo(path, dir_flag, size);
#else
    return SP_FALSE;
#endif
}
#else
spBool spGetFileInfo(const char *path, spBool *dir_flag, spFileOffset *size)
{
    return getFileInfo(path, dir_flag, size);
}
#endif

spBool spExists(const char *path)
{
    return spGetFileInfo(path, NULL, NULL);
}

spBool spIsExist(const char *path)
{
    return spExists(path);
}

spBool spIsFile(const char *path)
{
    spBool dir_flag = SP_FALSE;
    
    if (spGetFileInfo(path, &dir_flag, NULL) == SP_TRUE
        && dir_flag == SP_FALSE) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spIsDir(const char *path)
{
    spBool dir_flag = SP_FALSE;
    
    if (spGetFileInfo(path, &dir_flag, NULL) == SP_TRUE
        && dir_flag == SP_TRUE) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spCreateDir(const char *path, unsigned short mode)
{
    if (strnone(path)) return SP_FALSE;

#if defined(MACOS9)
    {
        Str255 ppath;
        HFileParam file_param;

        spStrCToP(path, ppath);

        file_param.ioNamePtr = ppath;
        file_param.ioVRefNum = 0;
        file_param.ioDirID = 0L;
        
        if (PBDirCreateSync((HParmBlkPtr)&file_param) != noErr) {
            return SP_FALSE;
        }
    }
#else
#if defined(_WIN32) && !defined(__CYGWIN32__)
    /*mkdir(path);*/
    CreateDirectory(path, NULL);
#else
    mkdir(path, mode);
#endif
#endif

    if (spIsDir(path) == SP_TRUE) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spCreateDirRecursive(const char *path, unsigned short mode)
{
    int i;
    int len;
    char buf[SP_MAX_PATHNAME];
    spBool flag = SP_FALSE;

    len = (int)strlen(path);
    
    for (i = 1; i < len; i++) {
        if (i == len - 1
            || ((path[i] == SP_DIR_SEPARATOR || path[i] == SP_ANOTHER_DIR_SEPARATOR)
                && !(i >= 1 && spIsMBTailCandidate(path[i-1], path[i])))) {
            spStrCopy(buf, i + 2, path);
            if (spIsDir(buf) == SP_FALSE) {
                if (spCreateDir(buf, mode) == SP_FALSE) {
                    return SP_FALSE;
                }
            }
            flag = SP_TRUE;
        }
    }

    if (flag == SP_FALSE) {
        /* not created */
        return spIsDir(path);
    }
    
    return flag;
}

spBool spRemoveDir(const char *path)
{
    if (strnone(path)) return SP_FALSE;
    
#if defined(_WIN32) && !defined(__CYGWIN32__)
    if (RemoveDirectory(path)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
#else
    if (rmdir(path) == 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
#endif
}

spBool spRemoveFile(const char *path)
{
    if (strnone(path)) return SP_FALSE;

    if (unlink(path) == 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spRenameFile(const char *oldpath, const char *newpath)
{
    if (strnone(oldpath) || strnone(newpath)) return SP_FALSE;
    
    if (rename(oldpath, newpath) == 0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spAddDirSeparator(char *dir)
{
    int i;
    
    if (dir == NULL) return SP_FALSE;

    i = (int)strlen(dir) - 1;
    
#if 1
    {
        char *p;

        if (i >= 0 && (((p = spStrRChr(dir, SP_DIR_SEPARATOR)) != NULL
                        && (&dir[i]) == p)
                       || ((p = spStrRChr(dir, SP_ANOTHER_DIR_SEPARATOR)) != NULL
                           && (&dir[i]) == p))) {
            /* dir[i] is a separator */
        } else {
            dir[i + 1] = SP_DIR_SEPARATOR;
            dir[i + 2] = NUL;
        }
    }
#else
    if (i < 0
        || (i >= 1 && spIsMBTailCandidate(dir[i-1], dir[i]))
        || (i >= 0 && dir[i] != SP_DIR_SEPARATOR && dir[i] != SP_ANOTHER_DIR_SEPARATOR)) {
        dir[i + 1] = SP_DIR_SEPARATOR;
        dir[i + 2] = NUL;
    }
#endif

    return SP_TRUE;
}

spBool spRemoveDirSeparator(char *dir)
{
    int i;
    
    if (dir == NULL) return SP_FALSE;

    i = (int)strlen(dir) - 1;

#if 1
    {
        char *p;

        if (i >= 0 && (((p = spStrRChr(dir, SP_DIR_SEPARATOR)) != NULL
                        && (&dir[i]) == p)
                       || ((p = spStrRChr(dir, SP_ANOTHER_DIR_SEPARATOR)) != NULL
                           && (&dir[i]) == p))) {
            /* dir[i] is a separator */
            dir[i] = NUL;
        }
    }
#else
    if (i >= 0) {
        if ((dir[i] == SP_DIR_SEPARATOR
             || dir[i] == SP_ANOTHER_DIR_SEPARATOR)
            && !(i >= 1 && spIsMBTailCandidate(dir[i-1], dir[i]))
            ) {
            dir[i] = NUL;
        }
    }
#endif

    return SP_TRUE;
}

spBool spIsExactName(const char *name)
{
    if (strnone(name)) return SP_FALSE;

#if defined(MACOS9)        /* only for Mac OS 9 in this version */
    {
        char *p;

        if ((p = spStrRChr(name, SP_DIR_SEPARATOR)) != NULL
            && name[0] != '~') {
            return SP_TRUE;
        }
#if defined(MACOSX)
        if (name[0] == '<') {
            spDebug(10, "spIsExactName", "path includes '<': %s\n", name);
            return SP_TRUE;
        }
#endif
    }
#elif defined(_WIN32)
    if (strveq(name, SP_DIR_SEPARATOR_STRING)
        || strveq(name, SP_ANOTHER_DIR_SEPARATOR_STRING)) {
        return SP_TRUE;
    } else {
        int len;

        len = (int)strlen(name);
        if (len >= 2 && name[1] == ':') {
            return SP_TRUE;
        }
    }
#else
    if (strveq(name, SP_DIR_SEPARATOR_STRING)) {
        return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

/*
 *        get exact name
 */
char *xspGetExactName(const char *name)
{
    int len;
    char buf[SP_MAX_PATHNAME];
    char *home, *username;
    char *exactname;
    
    if (name == NULL || *name == NUL) {
        exactname = xspGetCurrentDir();
    } else if (*name == '~') {
        name++;
        if (*name == SP_DIR_SEPARATOR || *name == SP_ANOTHER_DIR_SEPARATOR) {
            name++;
            home = xspGetHomeDir();
            len = (int)(strlen(home) + strlen(name)) + 2;
            exactname = xalloc(len, char);
#if defined(HAVE_SPRINTF_S)
            sprintf_s(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#elif defined(HAVE_SNPRINTF)
            snprintf(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#else
            sprintf(exactname, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#endif
            xfree(home);
        } else {
#if defined(_WIN32) || defined(MACOS)
            char *p;
            home = xspGetCurrentDir();
            if ((p = spGetBaseName(name)) != NULL) {
                name = p;
            }
            len = (int)(strlen(home) + strlen(name)) + 2;
            exactname = xalloc(len, char);
#if defined(HAVE_SPRINTF_S)
            sprintf_s(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#elif defined(HAVE_SNPRINTF)
            snprintf(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#else
            sprintf(exactname, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#endif
            xfree(home);
#else
            struct passwd *entry;

            spStrCopy(buf, sizeof(buf), name);
            if ((username = spStrChr(buf, SP_DIR_SEPARATOR)) != NULL)
                *username = NUL;
            
            if ((entry = getpwnam(buf)) == NULL) {
                char *p;
                home = xspGetCurrentDir();
                if ((p = spGetBaseName(name)) != NULL) {
                    name = p;
                }
                len = (int)(strlen(home) + strlen(name)) + 2;
                exactname = xalloc(len, char);
#if defined(HAVE_SPRINTF_S)
                sprintf_s(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#elif defined(HAVE_SNPRINTF)
                snprintf(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#else
                sprintf(exactname, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#endif
                xfree(home);
            } else {
                home = entry->pw_dir;
                while (*name != SP_DIR_SEPARATOR && *name != NUL) {
                    name++;
                }
                name++;
                len = (int)(strlen(home) + strlen(name)) + 2;
                exactname = xalloc(len, char);
#if defined(HAVE_SPRINTF_S)
                sprintf_s(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#elif defined(HAVE_SNPRINTF)
                snprintf(exactname, len, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#else
                sprintf(exactname, "%s%c%s", home, SP_DIR_SEPARATOR, name);
#endif
            }    
#endif
        }
    } else {
        if (streq(name, SP_PARENT_DIR_STRING)) {
            spStrCopy(buf, sizeof(buf), spGetCurrentDir());
            if ((username = spStrRChr(buf, SP_DIR_SEPARATOR)) != NULL)
                *username = NUL;
            exactname = strclone(buf);
        } else if (strveq(name, SP_SEPARATED_PARENT_DIR_STRING)) {
            name += strlen(SP_SEPARATED_PARENT_DIR_STRING);
            spStrCopy(buf, sizeof(buf), spGetCurrentDir());
            if ((username = spStrRChr(buf, SP_DIR_SEPARATOR)) != NULL)
                *username = NUL;
            spStrCat(buf, sizeof(buf), SP_DIR_SEPARATOR_STRING);
            spStrCat(buf, sizeof(buf), name);
            exactname = strclone(buf);
        } else if (streq(name, SP_CURRENT_DIR_STRING)) {
            spStrCopy(buf, sizeof(buf), spGetCurrentDir());
            exactname = strclone(buf);
        } else if (strveq(name, SP_SEPARATED_CURRENT_DIR_STRING)) {
            name += strlen(SP_SEPARATED_CURRENT_DIR_STRING);
            spStrCopy(buf, sizeof(buf), spGetCurrentDir());
            spStrCat(buf, sizeof(buf), SP_DIR_SEPARATOR_STRING);
            spStrCat(buf, sizeof(buf), name);
            exactname = strclone(buf);
        } else if (spIsExactName(name) == SP_TRUE) {
            exactname = strclone(name);
        } else {
            spStrCopy(buf, sizeof(buf), spGetCurrentDir());
            spStrCat(buf, sizeof(buf), SP_DIR_SEPARATOR_STRING);
            spStrCat(buf, sizeof(buf), name);
            exactname = strclone(buf);
        }
    }

    return exactname;
}

/*
 *        get base name
 */
char *spGetBaseName(const char *name)
{
    char *p, *q;
    
    if (name == NULL || *name == NUL)
        return NULL;

    p = spStrRChr(name, SP_DIR_SEPARATOR);
    q = spStrRChr(name, SP_ANOTHER_DIR_SEPARATOR);
    if (p == NULL && q == NULL) {
        return (char *)name;
    }

    p = MAX(p, q);
    if ((p - name) >= (int)strlen(name) - 1) {
        return NULL;
    }

    return (p + 1);
}

/*
 *        get allocated base name
 */
char *xspGetBaseName(const char *name)
{
    char *string;
    char *basename;

    if (name == NULL || *name == NUL)
        return NULL;

    string = spGetBaseName(name);
    basename = strclone(string);

    return basename;
}

/*
 *        get directory name
 */
char *spGetDirName(char *filename)
{
    char *p;

    if (strnone(filename)) return NULL;

    /* get directory name */
    if ((p = spStrRChr(filename, SP_DIR_SEPARATOR)) == NULL
        && (p = spStrRChr(filename, SP_ANOTHER_DIR_SEPARATOR)) == NULL) {
        return NULL;
    } else {
        *p = NUL;
    }

    return filename;
}

char *xspGetDirName(const char *filename)
{
    char *p;
    char *dirname;

    /* get directory name */
    if ((dirname = xspGetExactName(filename)) == NULL) {
        spDebug(80, "xspGetDirName", "xspGetExactName failed: %s\n", filename);
        dirname = strclone(SP_DEFAULT_DIRECTORY);
    } else {
        if ((p = spStrRChr(dirname, SP_DIR_SEPARATOR)) == NULL
            && (p = spStrRChr(dirname, SP_ANOTHER_DIR_SEPARATOR)) == NULL) {
            spDebug(80, "xspGetDirName", "dir separator not found: %s\n", dirname);
            xfree(dirname);
            dirname = strclone(SP_DEFAULT_DIRECTORY);
        } else {
            spDebug(80, "xspGetDirName", "dir separator found: %s\n", p);
            *p = NUL;
        }
    } 

    return dirname;
}

#if defined(MACOSX)
CFStringEncoding sp_mac_readable_encoding = kCFStringEncodingUTF8;
#endif

spBool spSetMacReadableEncoding(unsigned int encoding)
{
#if defined(MACOSX)
    sp_mac_readable_encoding = encoding;
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

char *xspGetReadablePath(const char *real_path)
{
    char *readable_path;

    if (real_path == NULL) return NULL;

#if defined(MACOSX)
    if (sp_mac_readable_encoding != kCFStringEncodingUTF8) {
        CFStringRef cfstr;
        char buf[SP_MAX_MESSAGE];

        cfstr = CFStringCreateWithCString(kCFAllocatorDefault, real_path, kCFStringEncodingUTF8);
        if (!CFStringGetCString(cfstr, buf, SP_MAX_PATHNAME, sp_mac_readable_encoding)) {
            CFRelease(cfstr);
            return NULL;
        }
        CFRelease(cfstr);
        
        return strclone(buf);
    }
#endif
    
    readable_path = strclone(real_path);

    return readable_path;
}

char *xspGetRealPath(const char *readable_path)
{
    char *real_path;

    if (readable_path == NULL) return NULL;
    
#if defined(MACOSX)
    if (sp_mac_readable_encoding != kCFStringEncodingUTF8) {
        CFStringRef cfstr;
        char buf[SP_MAX_MESSAGE];

        cfstr = CFStringCreateWithCString(kCFAllocatorDefault, readable_path, sp_mac_readable_encoding);
        if (!CFStringGetCString(cfstr, buf, SP_MAX_PATHNAME, kCFStringEncodingUTF8)) {
            CFRelease(cfstr);
            return NULL;
        }
        CFRelease(cfstr);

        return strclone(buf);
    }
#endif
    
    real_path = strclone(readable_path);
    
    return real_path;
}

char *xspCutPathList(const char *pathlist, int col)
{
    int i;
    int len;
    char *sep;
    char *exactname;
    char *path = NULL;
    
    if (strnone(pathlist)) {
        return NULL;
    }
    
    for (i = 0; (!strnone(pathlist)); i++) {
        sep = spStrChr(pathlist, SP_PATHLIST_SEPARATOR);
        if (i == col) {
            if (sep != NULL) {
                len = (int)(sep - pathlist);
                path = xalloc(len + 1, char);
#if defined(HAVE_STRNCPY_S)
                strncpy_s(path, len + 1, pathlist, len);
#else
                strncpy(path, pathlist, len);
#endif
                path[len] = NUL;
            } else {
                path = strclone(pathlist);
            }

            break;
        }
        
        if (sep == NULL) {
            break;
        } else {
            pathlist = sep + 1;
        }
    }

    if (path != NULL) {
        exactname = xspGetExactName(path);
        xfree(path);
    } else {
        exactname = NULL;
    }
         
    return exactname;
}

char *xspAppendPathList(const char *pathlist, const char *new_list)
{
    int orig_len;
    int len;
    char *olist;
    
    if (strnone(new_list)) {
        return strclone(pathlist);
    }

    orig_len = (int)strlen(pathlist);
    len = orig_len + (int)strlen(new_list) + 1;
    
    if (new_list[0] != SP_PATHLIST_SEPARATOR) {
        ++len;
    }

    olist = xalloc(len, char);
    spStrCopy(olist, len, pathlist);
    
    if (new_list[0] != SP_PATHLIST_SEPARATOR) {
        olist[orig_len] = SP_PATHLIST_SEPARATOR;
        olist[orig_len + 1] = NUL;
    }

    spStrCat(olist, len, new_list);

    return olist;
}

char *xspPrependPathList(const char *pathlist, const char *new_list)
{
    int orig_len;
    int new_list_len;
    int len;
    char *olist;
    
    if (strnone(new_list)) {
        return strclone(pathlist);
    }

    new_list_len = (int)strlen(new_list);
    orig_len = (int)strlen(pathlist);
    len = new_list_len + orig_len + 1;
    
    if (new_list[new_list_len] != SP_PATHLIST_SEPARATOR) {
        ++len;
    }

    olist = xalloc(len, char);
    spStrCopy(olist, len, new_list);
    
    if (new_list[new_list_len] != SP_PATHLIST_SEPARATOR) {
        olist[new_list_len + 1] = SP_PATHLIST_SEPARATOR;
        olist[new_list_len + 2] = NUL;
    }

    spStrCat(olist, len, pathlist);

    return olist;
}

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
static spBool getParentFSSpecMac(FSSpec *fsspec)
{
    CInfoPBRec        cinfo;

    cinfo.dirInfo.ioNamePtr = fsspec->name;
    cinfo.dirInfo.ioVRefNum = fsspec->vRefNum;
    cinfo.dirInfo.ioFDirIndex = -1;
    cinfo.dirInfo.ioDrDirID = fsspec->parID;
    if (PBGetCatInfo(&cinfo, false) != noErr) {
        return SP_FALSE;
    }
    fsspec->parID = cinfo.dirInfo.ioDrParID;

    return SP_TRUE;
}

spBool spFSSpecToExactNameMac(/*FSSpec *fsspec*/void *fsspec, char *filename)
{
    if (fsspec == NULL || filename == NULL) return SP_FALSE;

#if defined(MACOSX)
    {
        int flag = 1;
        FSRef fsref;
        FSSpec fss;
        CFStringRef cfstr;
        char buf[SP_MAX_PATHNAME];

        fss = *((FSSpec *)fsspec);
        spStrCopy(buf, sizeof(buf), "");
        spStrCopy(filename, SP_MAX_PATHNAME, "");
        
        while (flag) {
            if (FSpMakeFSRef(&fss, &fsref) == noErr) {
                spStrCopy(buf, sizeof(buf), filename);
                if (FSRefMakePath(&fsref, filename, SP_MAX_PATHNAME) == noErr) {
                    if ((int)strlen(buf) >= 1) {
                        spStrCat(filename, SP_MAX_PATHNAME, SP_UNIX_DIR_SEPARATOR_STRING);
                        spStrCat(filename, SP_MAX_PATHNAME, buf);
                    }
                    spDebug(50, "spFSSpecToExactNameMac", "done: filename = %s\n", filename);
                    return SP_TRUE;
                }
            }
            if (fss.parID == 2) {
                flag = 0;
            }

            spStrCopy(buf, sizeof(buf), filename);

            if ((cfstr = CFStringCreateWithPascalString(kCFAllocatorDefault, fss.name, CFStringGetSystemEncoding())) != NULL) {
                if (!CFStringGetCString(cfstr, filename, SP_MAX_PATHNAME, kCFStringEncodingUTF8)) {
                    flag = 0;
                }
                CFRelease(cfstr);
            }
            if (getParentFSSpecMac(&fss) == SP_FALSE) {
                return SP_FALSE;
            }

            if ((int)strlen(buf) >= 1) {
                spStrCat(filename, SP_MAX_PATHNAME, SP_UNIX_DIR_SEPARATOR_STRING);
                spStrCat(filename, SP_MAX_PATHNAME, buf);
            }
            spDebug(50, "spFSSpecToExactNameMac", "filename = %s\n", filename);
        }

        return SP_FALSE;
    }
#else
    {
        int flag = 1;
        char buf[SP_MAX_PATHNAME];
        FSSpec fss;

        fss = *((FSSpec *)fsspec);
        spStrPToC(fss.name, filename);
        while (flag)        {
            if (fss.parID == 2) {
                flag = 0;
            }
            if (getParentFSSpecMac(&fss) == SP_FALSE) {
                return SP_FALSE;
            }
            spStrCopy(buf, sizeof(buf), filename);
            spStrPToC(fss.name, filename);
            spStrCat(filename, SP_MAX_PATHNAME, SP_MAC_DIR_SEPARATOR_STRING);
            spStrCat(filename, SP_MAX_PATHNAME, buf);
        }
    }
#endif
    spDebug(50, "spFSSpecToExactNameMac", "filename = %s\n", filename);
    
    return SP_TRUE;
}
#endif

#if defined(ANDROID)
#include <jni.h>

static char *sp_android_package_name = NULL;
static char *sp_android_base_dir = NULL;
static char *sp_android_files_dir = NULL;
static char *sp_android_lib_dir = NULL;
static char *sp_android_sdcard_dir = NULL;
static char *sp_android_music_dir = NULL;
static char *sp_android_movies_dir = NULL;
static char *sp_android_pictures_dir = NULL;
static char *sp_android_downloads_dir = NULL;

static char *xgetFileAbsolutePathAndroid(JNIEnv *env, jobject fileobj, jmethodID getAbsolutePath)
{
    jstring str;
    char *ostr = NULL;
    
    if ((str = (*env)->CallObjectMethod(env, fileobj, getAbsolutePath)) != NULL) {
        const char *cstr;
        
        if ((cstr = (*env)->GetStringUTFChars(env, str, NULL)) != NULL) {
            ostr = strclone(cstr);
            spDebug(50, "xgetFileAbsolutePathAndroid", "ostr = %s\n", ostr);
            (*env)->ReleaseStringUTFChars(env, str, cstr);
        }
        (*env)->DeleteLocalRef(env, str);
    }

    return ostr;
}

static char *xgetExternalStoragePublicDirectory(JNIEnv *env, jclass envcls,
                                                jmethodID getAbsolutePath, jmethodID getExternalStoragePublicDirectory,
                                                const char *type)
{
    char *ostr = NULL;
    jfieldID fieldid;
    jstring fieldstr;
    jobject fileobj;

    if ((fieldid = (*env)->GetStaticFieldID(env, envcls, type, "Ljava/lang/String;")) != NULL) {
        if ((fieldstr = (*env)->GetStaticObjectField(env, envcls, fieldid)) != NULL) {
            if ((fileobj = (*env)->CallStaticObjectMethod(env, envcls, getExternalStoragePublicDirectory, fieldstr)) != NULL) {
                ostr = xgetFileAbsolutePathAndroid(env, fileobj, getAbsolutePath);
                (*env)->DeleteLocalRef(env, fileobj);
            }
            (*env)->DeleteLocalRef(env, fieldstr);
        }
    }
    
    return ostr;
}

static char *xgetAndroidLibDirectoryFromContext(JNIEnv *env, jobject clsobj, jclass cls)
{
    jmethodID getApplicationInfo;
    jobject applicationInfo;
    jclass applicationInfoCls;
    jfieldID nativeLibraryDir;
    jstring str;
    const char *cstr;
    char *lib_dir = NULL;
    
    getApplicationInfo = (*env)->GetMethodID(env, cls, "getApplicationInfo", "()Landroid/content/pm/ApplicationInfo;");
    if (getApplicationInfo != NULL) {
        if ((applicationInfo = (*env)->CallObjectMethod(env, clsobj, getApplicationInfo)) != NULL) {
            applicationInfoCls = (*env)->GetObjectClass(env, applicationInfo);
            
            nativeLibraryDir = (*env)->GetFieldID(env, applicationInfoCls, "nativeLibraryDir", "Ljava/lang/String;");
            
            if ((str = (*env)->GetObjectField(env, applicationInfo, nativeLibraryDir)) != NULL) {
                if ((cstr = (*env)->GetStringUTFChars(env, str, NULL)) != NULL) {
                    lib_dir = strclone(cstr);
                    (*env)->ReleaseStringUTFChars(env, str, cstr);
                }
                
                (*env)->DeleteLocalRef(env, str);
            }
            
            (*env)->DeleteLocalRef(env, applicationInfo);
        } else {
            spDebug(50, "xgetAndroidLibDirectoryFromContext", "getApplicationInfo failed\n");
        }
    } else {
        spDebug(50, "xgetAndroidLibDirectoryFromContext", "getApplicationInfo method is not found\n");
    }

    spDebug(50, "xgetAndroidLibDirectoryFromContext", "lib_dir = %s\n", lib_dir);
    
    return lib_dir;
}

void spUpdateDirectoryAndroid(JNIEnv *env, jobject clsobj)
{
    jclass cls;
    jclass filecls;
    jstring str;
    jmethodID getPackageName;
    char *ptr;
    char buf[SP_MAX_PATHNAME];
    const char *cstr;

    cls = (*env)->GetObjectClass(env, clsobj);
    
    getPackageName = (*env)->GetMethodID(env, cls, "getPackageName", "()Ljava/lang/String;");    
    str = (*env)->CallObjectMethod(env, clsobj, getPackageName);
    spDebug(50, "spUpdateDirectoryAndroid", "getPackageName: str = %ld\n", (long)str);
    
    if (str != NULL) {
        if ((cstr = (*env)->GetStringUTFChars(env, str, NULL)) != NULL) {
            if (sp_android_package_name != NULL) xfree(sp_android_package_name);
            sp_android_package_name = strclone(cstr);
            (*env)->ReleaseStringUTFChars(env, str, cstr);
        }
        (*env)->DeleteLocalRef(env, str);
        
        spDebug(50, "spUpdateDirectoryAndroid", "package_name = %s\n", sp_android_package_name);
   }

    if ((filecls = (*env)->FindClass(env, "java/io/File")) != NULL) {
        jmethodID getAbsolutePath;
        jmethodID getFilesDir;
        jobject fileobj;
        jclass envcls;

        getAbsolutePath = (*env)->GetMethodID(env, filecls, "getAbsolutePath", "()Ljava/lang/String;");
        getFilesDir = (*env)->GetMethodID(env, cls, "getFilesDir", "()Ljava/io/File;");
        fileobj = (*env)->CallObjectMethod(env, clsobj, getFilesDir);
        
        if (fileobj != NULL) {
            if (sp_android_files_dir != NULL) xfree(sp_android_files_dir);
            sp_android_files_dir = xgetFileAbsolutePathAndroid(env, fileobj, getAbsolutePath);
            (*env)->DeleteLocalRef(env, fileobj);

            if (sp_android_files_dir != NULL) {
                spDebug(50, "spUpdateDirectoryAndroid", "files_dir = %s\n", sp_android_files_dir);
                
                spStrCopy(buf, sizeof(buf), sp_android_files_dir);
                if ((ptr = spStrRChr(buf, SP_DIR_SEPARATOR)) != NULL) {
                    *ptr = NUL;

                    if (sp_android_base_dir != NULL) xfree(sp_android_base_dir);
                    sp_android_base_dir = strclone(buf);
                    spDebug(50, "spUpdateDirectoryAndroid", "base_dir = %s\n", sp_android_base_dir);
                
                    if (sp_android_lib_dir != NULL) xfree(sp_android_lib_dir);
                    /*sp_android_lib_dir = xspStrCat(sp_android_base_dir, "/lib");*/
                    sp_android_lib_dir = xgetAndroidLibDirectoryFromContext(env, clsobj, cls);
                    spDebug(50, "spUpdateDirectoryAndroid", "lib_dir = %s\n", sp_android_lib_dir);
                }
            }
        }
        
        if ((envcls = (*env)->FindClass(env, "android/os/Environment")) != NULL) {
            jmethodID getExternalStorageDirectory;
            jmethodID getExternalStoragePublicDirectory;
            jfieldID fieldid;
            jstring fieldstr;
            
            getExternalStorageDirectory = (*env)->GetStaticMethodID(env, envcls, "getExternalStorageDirectory", "()Ljava/io/File;");
            
            if ((fileobj = (*env)->CallStaticObjectMethod(env, envcls, getExternalStorageDirectory)) != NULL) {
                if (sp_android_sdcard_dir != NULL) xfree(sp_android_sdcard_dir);
                sp_android_sdcard_dir = xgetFileAbsolutePathAndroid(env, fileobj, getAbsolutePath);
            }

            getExternalStoragePublicDirectory = (*env)->GetStaticMethodID(env, envcls,
                                                                "getExternalStoragePublicDirectory", "(Ljava/lang/String;)Ljava/io/File;");

            if (sp_android_music_dir != NULL) xfree(sp_android_music_dir);
            sp_android_music_dir = xgetExternalStoragePublicDirectory(env, envcls,
                                                                      getAbsolutePath, getExternalStoragePublicDirectory,
                                                                      "DIRECTORY_MUSIC");
            
            if (sp_android_movies_dir != NULL) xfree(sp_android_movies_dir);
            sp_android_movies_dir = xgetExternalStoragePublicDirectory(env, envcls,
                                                                       getAbsolutePath, getExternalStoragePublicDirectory,
                                                                       "DIRECTORY_MOVIES");
            
            if (sp_android_pictures_dir != NULL) xfree(sp_android_pictures_dir);
            sp_android_pictures_dir = xgetExternalStoragePublicDirectory(env, envcls,
                                                                         getAbsolutePath, getExternalStoragePublicDirectory,
                                                                         "DIRECTORY_PICTURES");
            
            if (sp_android_downloads_dir != NULL) xfree(sp_android_downloads_dir);
            sp_android_downloads_dir = xgetExternalStoragePublicDirectory(env, envcls,
                                                                          getAbsolutePath, getExternalStoragePublicDirectory,
                                                                          "DIRECTORY_DOWNLOADS");
            
            (*env)->DeleteLocalRef(env, envcls);
        }
        
        (*env)->DeleteLocalRef(env, filecls);
    }

    return;
}
#endif

const char *spGetCurrentDir(void)
{
    static char sp_current_directory[SP_MAX_PATHNAME] = "";
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    FSSpec fsspec;
    
    if (FSMakeFSSpec(0, 0L, "\p", &fsspec) != noErr) {
        return NULL;
    }
    spFSSpecToExactNameMac(&fsspec, sp_current_directory);
#elif defined(ANDROID)
    if (sp_android_files_dir != NULL) {
        spStrCopy(sp_current_directory, sizeof(sp_current_directory), sp_android_files_dir);
    }
#else
    if (getcwd(sp_current_directory, SP_MAX_PATHNAME) == NULL) {
        spStrCopy(sp_current_directory, sizeof(sp_current_directory), SP_DEFAULT_DIRECTORY);
    } else {
        spRemoveDirSeparator(sp_current_directory);
    }
#endif

    return sp_current_directory;
}

char *xspGetCurrentDir(void)
{
    return strclone(spGetCurrentDir());
}

const char *spGetDefaultDir(void)
{
    static char sp_default_directory[SP_MAX_PATHNAME] = "";

    if (strnone(sp_default_directory)) {
#if defined(_WIN32)
        char path[SP_MAX_PATHNAME];
        
        if (GetModuleFileName(NULL, path, sizeof(path))) {
#if !defined(__CYGWIN32__)
            char drive[SP_MAX_PATHNAME];
            char dir[SP_MAX_PATHNAME];
            
#if defined(HAVE_SPLITPATH_S)
            _splitpath_s(path, drive, sizeof(drive), dir, sizeof(dir), NULL, 0, NULL, 0);
#else
            _splitpath(path, drive, dir, NULL, NULL);
#endif
            spStrCopy(path, sizeof(path), drive);
            spStrCat(path, sizeof(path), dir);
            spRemoveDirSeparator(path);
#else
            spGetDirName(path);
#endif
            spStrCopy(sp_default_directory, sizeof(sp_default_directory), path);
        } else {
            spStrCopy(sp_default_directory, sizeof(sp_default_directory), SP_DEFAULT_DIRECTORY);
        }
#elif defined(MACOS)
        char *dirname;
        
        dirname = xspGetCurrentDir();
        spStrCopy(sp_default_directory, sizeof(sp_default_directory), dirname);
        xfree(dirname);
#elif defined(ANDROID)
        if (sp_android_base_dir != NULL) {
            spStrCopy(sp_default_directory, sizeof(sp_default_directory), sp_android_base_dir);
        }
#else
        spStrCopy(sp_default_directory, sizeof(sp_default_directory), SP_DEFAULT_DIRECTORY);
#endif
    }

    spDebug(80, "spGetDefaultDir", "sp_default_directory = %s\n", sp_default_directory);
    
    return sp_default_directory;
}

#if defined(_WIN32)
static spBool getSpecialFolderWin(int folder, char *buf)
{
    spBool flag = SP_FALSE;
    LPITEMIDLIST pidl;
    IMalloc *pMalloc;

    if (SHGetMalloc(&pMalloc) == NOERROR) {
        if (SHGetSpecialFolderLocation(NULL, folder, &pidl) == NOERROR) {
            SHGetPathFromIDList(pidl, buf);
            spRemoveDirSeparator(buf);
            pMalloc->lpVtbl->Free(pMalloc, pidl);
            flag = SP_TRUE;
        }
        pMalloc->lpVtbl->Release(pMalloc);
    }

    return flag;
}
#endif

#define SP_HOME_DIR_ENV "HOME"
    
const char *spGetHomeDir(void)
{
    static char sp_home_directory[SP_MAX_PATHNAME] = "";

    if (strnone(sp_home_directory)) {
#if defined(_WIN32)
        spStrCopy(sp_home_directory, sizeof(sp_home_directory), spGetDefaultDir());
        
        if (access(sp_home_directory, 06) == -1) {
            if (getSpecialFolderWin(CSIDL_PERSONAL, sp_home_directory) == SP_FALSE) {
                size_t drivelen = 0, pathlen = 0;
                char *drive = NULL;
                char *path = NULL;

#if defined(HAVE_DUPENV_S)
                if (_dupenv_s(&drive, &drivelen, "HOMEDRIVE") != 0
                    || _dupenv_s(&path, &pathlen, "HOMEPATH") != 0
                    || drivelen <= 0 || pathlen <= 0) {
                    if (drive != NULL) free(drive);
                    if (path != NULL) free(path);
                    drive = path = NULL;
                }
#else
                if ((drive = getenv("HOMEDRIVE")) != NULL
                    && (path = getenv("HOMEPATH")) != NULL) {
                    pathlen = strlen(path);
                }
#endif
                
                if (drive != NULL && path != NULL && (int)pathlen >= 2) {
                    spStrCopy(sp_home_directory, sizeof(sp_home_directory), drive);
                    spStrCat(sp_home_directory, sizeof(sp_home_directory), path);
                    spRemoveDirSeparator(sp_home_directory);
                    spDebug(50, "spGetHomeDir", "drive = %s, path = %s, sp_home_directory = %s\n",
                            drive, path, sp_home_directory);
                }
                
#if defined(HAVE_DUPENV_S)
                if (drive != NULL) free(drive);
                if (path != NULL) free(path);
#endif
            }
        }
#elif defined(ANDROID)
        if (sp_android_files_dir != NULL) {
            spStrCopy(sp_home_directory, sizeof(sp_home_directory), sp_android_files_dir);
        }
#else
        char *string;
        
        if ((string = getenv(SP_HOME_DIR_ENV)) == NULL) {
            spStrCopy(sp_home_directory, sizeof(sp_home_directory), spGetDefaultDir());
        } else {
            spStrCopy(sp_home_directory, sizeof(sp_home_directory), string);
            spRemoveDirSeparator(sp_home_directory);
        }
#endif
    }
    
    return sp_home_directory;
}

char *xspGetHomeDir(void)
{
    return strclone(spGetHomeDir());
}

const char *spGetTempDir(void)
{
    static char sp_temp_directory[SP_MAX_PATHNAME] = "";

    if (strnone(sp_temp_directory)) {
#if defined(_WIN32) && !defined(__CYGWIN32__)
        if (GetTempPath(SP_MAX_PATHNAME, sp_temp_directory)) {
            spRemoveDirSeparator(sp_temp_directory);
        } else if (GetWindowsDirectory(sp_temp_directory, SP_MAX_PATHNAME)) {
            spRemoveDirSeparator(sp_temp_directory);
            spStrCat(sp_temp_directory, sizeof(sp_temp_directory), "\\Temp");
        } else {
            spStrCopy(sp_temp_directory, sizeof(sp_temp_directory), "C:\\Windows\\Temp");
        }
        if (access(sp_temp_directory, 06) == -1) {
            spStrCopy(sp_temp_directory, sizeof(sp_temp_directory), SP_DEFAULT_DIRECTORY);
        }
#elif defined(IPHONE)
#if defined(HAVE_SNPRINTF)
        snprintf(sp_temp_directory, sizeof(sp_temp_directory), "%s%ctmp", spGetHomeDir(), SP_DIR_SEPARATOR);
#else
        sprintf(sp_temp_directory, "%s%ctmp", spGetHomeDir(), SP_DIR_SEPARATOR);
#endif
#elif defined(MACOS9)
        spStrCopy(sp_temp_directory, sizeof(sp_temp_directory), spGetDefaultDir());
#else
        spStrCopy(sp_temp_directory, sizeof(sp_temp_directory), "/tmp");
#endif
    }

    return sp_temp_directory;
}

char *xspGetTempDir(void)
{
    return strclone(spGetTempDir());
}

static char sp_company_id[SP_MAX_LINE] = "";
static char sp_application_id[SP_MAX_LINE] = "";
static char sp_version_id[SP_MAX_LINE] = "";

static spBool sp_app_dir_id_depend = SP_FALSE;
static char sp_application_directory[SP_MAX_PATHNAME] = "";
static char sp_version_application_directory[SP_MAX_PATHNAME] = "";

static char sp_application_temp_directory[SP_MAX_PATHNAME] = "";

spBool spSetApplicationId(const char *id)
{
    int num_sep = 0;
    char *p1 = NULL, *p2 = NULL;
    char buf[SP_MAX_LINE];
    
    if (strnone(id)) return SP_FALSE;

    spDebug(50, "spSetApplicationId", "id = %s\n", id);
    spStrCopy(buf, sizeof(buf), id);

    if ((p1 = spStrChr(buf, '/')) != NULL) {
        spDebug(100, "spSetApplicationId", "p1 = %s\n", p1);
        num_sep++;
        if ((p2 = spStrChr(p1 + 1, '/')) != NULL) {
            spDebug(100, "spSetApplicationId", "p2 = %s\n", p2);
            num_sep++;
        }
    }
    spDebug(80, "spSetApplicationId", "num_sep = %d\n", num_sep);

    if (num_sep <= 0) {
        spStrCopy(sp_application_id, sizeof(sp_application_id), buf);
    } else if (num_sep <= 1) {
        spStrCopy(sp_version_id, sizeof(sp_version_id), p1 + 1);
        *p1 = NUL;
        spStrCopy(sp_application_id, sizeof(sp_application_id), buf);
    } else {
        spStrCopy(sp_version_id, sizeof(sp_version_id), p2 + 1);
        *p2 = NUL;
        spStrCopy(sp_application_id, sizeof(sp_application_id), p1 + 1);
        *p1 = NUL;
        spStrCopy(sp_company_id, sizeof(sp_company_id), buf);
    }
    
    spDebug(50, "spSetApplicationId", "company = %s, application = %s, version = %s\n",
            sp_company_id, sp_application_id, sp_version_id);

    if (strnone(sp_application_id)) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}

const char *spGetCompanyId(void)
{
    return sp_company_id;
}

const char *spGetApplicationId(void)
{
    return sp_application_id;
}

const char *spGetVersionId(void)
{
    return sp_version_id;
}

#if defined(MACOSX)
spBool spGetBundleResourcesDirMac(char *dir, int dir_size)
{
    CFBundleRef bundle;
    CFURLRef bundleurl, url;
    CFStringRef bundlepath, rsrcpath;
    spBool flag = SP_FALSE;

    if ((bundle = CFBundleGetMainBundle()) != NULL) {
        if ((bundleurl = CFBundleCopyBundleURL(bundle)) != NULL) {
            if ((bundlepath = CFURLCopyFileSystemPath(bundleurl, kCFURLPOSIXPathStyle)) != NULL) {
                flag = SP_TRUE;
                
                spStrCopy(dir, dir_size, CFStringGetCStringPtr(bundlepath, kCFStringEncodingUTF8));
                spDebug(50, "spGetBundleResourcesDirMac", "dir = %s\n", dir);

#if !defined(IPHONE)
                if (spEqSuffix(dir, ".app") == SP_TRUE
                    && (url = CFBundleCopyResourcesDirectoryURL(bundle)) != NULL) {
                    if ((rsrcpath = CFURLCopyFileSystemPath(url, kCFURLPOSIXPathStyle)) != NULL) {
                        spAddDirSeparator(dir);
                        spStrCat(dir, dir_size, CFStringGetCStringPtr(rsrcpath, kCFStringEncodingUTF8));
                        spDebug(50, "spGetBundleResourcesDirMac", "resources included: dir = %s\n", dir);
                        CFRelease(rsrcpath);
                    }
                    CFRelease(url);
                }
#endif
                CFRelease(bundlepath);
            }
            CFRelease(bundleurl);
        }
    } else {
        spDebug(50, "spGetBundleResourcesDirMac", "CFBundleGetMainBundle failed\n");
    }

    return flag;
}
#endif

const char *spGetApplicationLibDir(void)
{
    static char sp_application_lib_directory[SP_MAX_PATHNAME] = "";
    
    if (strnone(sp_application_lib_directory)) {
#if defined(_WIN32) || defined(MACOS)
#if defined(MACOSX)
        if (spGetBundleResourcesDirMac(sp_application_lib_directory,
                                       sizeof(sp_application_lib_directory)) == SP_TRUE) {
            spDebug(80, "spGetApplicationLibDir", "spGetBundleResourcesDirMac successfully done: %s\n",
                    sp_application_lib_directory);
        } else
#endif
        spStrCopy(sp_application_lib_directory, SP_MAX_PATHNAME, spGetDefaultDir());
#elif defined(ANDROID)
        spDebug(80, "spGetApplicationLibDir", "sp_android_lib_dir = %ld\n", (long)sp_android_lib_dir);
        if (sp_android_lib_dir != NULL) {
            spStrCopy(sp_application_lib_directory, SP_MAX_PATHNAME, sp_android_lib_dir);
        }
        spDebug(80, "spGetApplicationLibDir", "sp_application_lib_directory = %s\n", sp_application_lib_directory);
#else
        const char *id;
        
        id = spGetApplicationId();
        
        if (!strnone(id)) {
            int i;
            char *path;


            for (i = 0;; i++) {
                if ((path = xspCutPathList(SP_UNIX_DEFAULT_SHARE_PATH_LIST, i)) == NULL) {
                    id = NULL;
                    break;
                }
#if defined(HAVE_SPRINTF_S)
                sprintf_s(sp_application_lib_directory, sizeof(sp_application_lib_directory), "%s%c%s", path, SP_DIR_SEPARATOR, id);
#elif defined(HAVE_SNPRINTF)
                snprintf(sp_application_lib_directory, sizeof(sp_application_lib_directory), "%s%c%s", path, SP_DIR_SEPARATOR, id);
#else
                sprintf(sp_application_lib_directory, "%s%c%s", path, SP_DIR_SEPARATOR, id);
#endif
                xfree(path);
                
                if (spIsDir(sp_application_lib_directory) == SP_TRUE) {
                    break;
                }
            }
        }
        
        if (strnone(id)) {
            spStrCopy(sp_application_lib_directory, SP_MAX_PATHNAME, spGetDefaultDir());
        }
#endif
    }

    spDebug(80, "spGetApplicationLibDir", "sp_application_lib_directory = %s\n",
            sp_application_lib_directory);
    
    return sp_application_lib_directory;
}

#if defined(MACOSX)
const char *spGetApplicationDocDir(void)
{
    static char dir[SP_MAX_PATHNAME] = "";

    if (strnone(dir)) {
        spGetBundleResourcesDirMac(dir, sizeof(dir));
    }
    
    return dir;
}
#else
const char *spGetApplicationDocDir(void)
{
    return spGetApplicationLibDir();
}
#endif

spBool spGetApplicationPath(char *application_path, int path_size, spBool *id_depend,
                            const char *application_id, const char *company_id, spBool create_dir)
{
    spBool app_dir_id_depend = SP_FALSE;
    
#if defined(_WIN32)
    {
        const char *dir;
        char module_name[SP_MAX_PATHNAME];
        char buf[SP_MAX_PATHNAME];
            
        dir = spGetDefaultDir();
        if (access(dir, 06) == -1) {
            dir = spGetHomeDir();
        }

        if (strnone(application_id)
            && GetModuleFileName(NULL, module_name, sizeof(module_name))
            && spRemoveSuffix(module_name, NULL) == SP_TRUE) {
            application_id =  spGetBaseName(module_name);
        }
            
        if (!strnone(application_id)) {
            if (getSpecialFolderWin(CSIDL_APPDATA, buf) == SP_TRUE) {
                if (!strnone(company_id)) {
                    spAddDirSeparator(buf); spStrCat(buf, sizeof(buf), company_id);
                    
                    if (create_dir == SP_TRUE) {
                        if (spIsDir(buf) == SP_FALSE
                            && !CreateDirectory(buf, NULL)) {
                            return SP_FALSE;
                        }
                    }
                }
                spAddDirSeparator(buf); spStrCat(buf, sizeof(buf), application_id);
                if (create_dir == SP_FALSE
                    || spIsDir(buf) == SP_TRUE
                    || CreateDirectory(buf, NULL)) {
                    dir = buf;
                    app_dir_id_depend = SP_TRUE;
                }
            }
        }
        spDebug(80, "spCreateApplicationDir", "dir = %s\n", dir);
            
        spStrCopy(application_path, path_size, dir);
    }
#elif defined(IPHONE)
    spStrCopy(application_path, path_size, spGetHomeDir());
    app_dir_id_depend = SP_TRUE;
#else
    if (id_depend == NULL || *id_depend == SP_FALSE || strnone(application_id)) {
        spStrCopy(application_path, path_size, spGetHomeDir());
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(application_path, path_size, "%s%c.%s_dir",
                  spGetHomeDir(), SP_DIR_SEPARATOR, application_id);
#elif defined(HAVE_SNPRINTF)
        snprintf(application_path, path_size, "%s%c.%s_dir",
                 spGetHomeDir(), SP_DIR_SEPARATOR, application_id);
#else
        sprintf(application_path, "%s%c.%s_dir",
                spGetHomeDir(), SP_DIR_SEPARATOR, application_id);
#endif
        if (create_dir == SP_TRUE
            && spIsDir(application_path) == SP_FALSE
            && spCreateDir(application_path, 0700) == SP_FALSE) {
            spStrCopy(application_path, path_size, spGetHomeDir());
        } else {
            app_dir_id_depend = SP_TRUE;
        }
    }
#endif

    if (id_depend != NULL) *id_depend = app_dir_id_depend;
    
    return SP_TRUE;
}

spBool spGetVersionApplicationPath(char *version_application_path, int path_size, 
                                   const char *application_path, const char *version_id, spBool create_dir)
{
    spBool version_depend = SP_TRUE;
    
    if (!strnone(version_id)) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(version_application_path, path_size, "%s%c%s",
                  application_path, SP_DIR_SEPARATOR, version_id);
#elif defined(HAVE_SNPRINTF)
        snprintf(version_application_path, path_size, "%s%c%s",
                 application_path, SP_DIR_SEPARATOR, version_id);
#else
        sprintf(version_application_path, "%s%c%s",
                application_path, SP_DIR_SEPARATOR, version_id);
#endif

        if (create_dir == SP_TRUE) {
            if (spIsDir(version_application_path) == SP_FALSE
                && spCreateDir(version_application_path, 0700) == SP_FALSE) {
                version_depend = SP_FALSE;
            }
        }
    } else {
        version_depend = SP_FALSE;
    }

    return version_depend;
}

spBool spGetApplicationTempPath(char *application_temp_path, int path_size,
                                const char *application_path, const char *application_id, const char *version_id,
                                spBool id_depend, spBool create_dir)
{
#if defined(IPHONE)
    spStrCopy(application_temp_path, path_size, spGetTempDir());
#else /* defined(IPHONE) */
    const char *dir;
    
    dir = application_path;
        
    if (id_depend == SP_TRUE) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(application_temp_path, path_size, "%s%ctmp", dir, SP_DIR_SEPARATOR);
#elif defined(HAVE_SNPRINTF)
        snprintf(application_temp_path, path_size, "%s%ctmp", dir, SP_DIR_SEPARATOR);
#else
        sprintf(application_temp_path, "%s%ctmp", dir, SP_DIR_SEPARATOR);
#endif
#if !defined(_WIN32)
    } else if (id_depend == SP_FALSE && !strnone(application_id)) {
        if (!strnone(version_id)) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(application_temp_path, path_size, "%s%c.%s-%s_tmp",
                      dir, SP_DIR_SEPARATOR, application_id, version_id);
#elif defined(HAVE_SNPRINTF)
            snprintf(application_temp_path, path_size, "%s%c.%s-%s_tmp",
                     dir, SP_DIR_SEPARATOR, application_id, version_id);
#else
            sprintf(application_temp_path, "%s%c.%s-%s_tmp",
                    dir, SP_DIR_SEPARATOR, application_id, version_id);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(application_temp_path, path_size, "%s%c.%s_tmp",
                      dir, SP_DIR_SEPARATOR, application_id);
#elif defined(HAVE_SNPRINTF)
            snprintf(application_temp_path, path_size, "%s%c.%s_tmp",
                     dir, SP_DIR_SEPARATOR, application_id);
#else
            sprintf(application_temp_path, "%s%c.%s_tmp",
                    dir, SP_DIR_SEPARATOR, application_id);
#endif
        }
#endif /* !defined(_WIN32) */
    } else {
        spStrCopy(application_temp_path, path_size, dir);
    }
    if (create_dir == SP_TRUE) {
        if (spIsDir(application_temp_path) == SP_FALSE
            && spCreateDir(application_temp_path, 0700) == SP_FALSE) {
            return SP_FALSE;
        }
    }
#endif /* !defined(IPHONE) */

    return SP_TRUE;
}

/*
 * UNIX, Mac: *id_depend == SP_FALSE ? spGetHomeDir() : spGetHomeDir()/.<application id>_dir
 * Windows: 1. CSIDL_APPDATA/[<company id>/]<application id>, 2. Module Path, 3. spGetHomeDir()
 */
const char *spCreateApplicationDir(spBool *id_depend, spBool *version_depend)
{
    if (id_depend != NULL && *id_depend == SP_TRUE) {
        sp_app_dir_id_depend = SP_TRUE;
    } else {
        sp_app_dir_id_depend = SP_FALSE;
    }

    if (spGetApplicationPath(sp_application_directory, sizeof(sp_application_directory), &sp_app_dir_id_depend, 
                             sp_application_id, sp_company_id, SP_TRUE) == SP_FALSE) {
        return NULL;
    }

    if (version_depend != NULL && *version_depend == SP_TRUE) {
        if (sp_app_dir_id_depend == SP_TRUE) {
            *version_depend = spGetVersionApplicationPath(sp_version_application_directory,
                                                          sizeof(sp_version_application_directory),
                                                          sp_application_directory, sp_version_id, SP_TRUE);
        } else {
            spStrCopy(sp_version_application_directory, sizeof(sp_version_application_directory), "");
            *version_depend = SP_FALSE;
        }
    }

    if (id_depend != NULL) *id_depend = sp_app_dir_id_depend;
    
    return sp_application_directory;
}

const char *spGetApplicationDir(spBool *version_depend)
{
    spBool id_depend = SP_FALSE;
    
    if (version_depend != NULL && *version_depend == SP_TRUE
        && !strnone(sp_version_application_directory)) {
        return sp_version_application_directory;
    } else {
        if (strnone(sp_application_directory)) {
            spCreateApplicationDir(&id_depend, NULL);
        }
        if (version_depend != NULL) *version_depend = SP_FALSE;

        return sp_application_directory;
    }
}

char *xspGetApplicationDir(spBool *version_depend)
{
    return strclone(spGetApplicationDir(version_depend));
}

/*
 * 1. spGetApplicationDir()/tmp
 * 2. spGetApplicationDir()/.<application id>-<version id>_tmp (except for Windows)
 * 3. spGetApplicationDir()/.<application id>_tmp (except for Windows)
 * 4. spGetTempDir()
 */
const char *spGetApplicationTempDir(void)
{
    spBool id_depend = SP_TRUE;
    spBool version_depend = SP_TRUE;
    const char *dir;
    
    if (strnone(sp_application_temp_directory)) {
        dir = spGetApplicationDir(&version_depend);
        id_depend = sp_app_dir_id_depend;

        if (spGetApplicationTempPath(sp_application_temp_directory, sizeof(sp_application_temp_directory),
                                     dir, sp_application_id, sp_version_id, id_depend, SP_TRUE) == SP_FALSE) {
            spStrCopy(sp_application_temp_directory, sizeof(sp_application_temp_directory),
                      spGetTempDir());
        }
    }

    return sp_application_temp_directory;
}

char *xspGetApplicationTempDir(void)
{
    return strclone(spGetApplicationTempDir());
}

spBool spSetApplicationTempDir(const char *dir)
{
    if (strnone(dir) || access(dir, 06) == -1) {
        return SP_FALSE;
    }
    spStrCopy(sp_application_temp_directory, sizeof(sp_application_temp_directory), dir);

    return SP_TRUE;
}

#if defined(ANDROID)
char *xspGetDocumentDir(spFileKind file_kind)
{
    char *str = NULL;
    
    switch (file_kind) {
      case SP_FILE_KIND_PICTURE:
      case SP_FILE_KIND_FIGURE:
        str = strclone(sp_android_pictures_dir);
        break;
        
      case SP_FILE_KIND_MUSIC:
      case SP_FILE_KIND_SOUND:
        str = strclone(sp_android_music_dir);
        break;
        
      case SP_FILE_KIND_MOVIE:
        str = strclone(sp_android_movies_dir);
        break;
        
      case SP_FILE_KIND_LIBRARY:
        str = strclone(sp_android_lib_dir);
        break;
        
      case SP_FILE_KIND_ARCHIVE:
        str = strclone(sp_android_downloads_dir);
        break;
        
      default:
        str = strclone(sp_android_files_dir);
        break;
    }

    return str;
}
#else /* defined(ANDROID) */
static spBool findDocumentDir(char **list, char *dir, int dirsize)
{
    int i;
    char buf[SP_MAX_PATHNAME];

    for (i = 0; list[i] != NULL; i++) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, sizeof(buf), "%s%c%s", dir, SP_DIR_SEPARATOR, list[i]);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, sizeof(buf), "%s%c%s", dir, SP_DIR_SEPARATOR, list[i]);
#else
        sprintf(buf, "%s%c%s", dir, SP_DIR_SEPARATOR, list[i]);
#endif
        if (spIsDir(buf) == SP_TRUE) {
            spStrCopy(dir, dirsize, buf);
            return SP_TRUE;
        }
    }

    return SP_FALSE;
}

#if defined(IPHONE)
char *xspGetDocumentDir(spFileKind file_kind)
{
    char buf[SP_MAX_PATHNAME];
    static char *document_list[] =
    {
        "Documents",
        NULL,
    };
    
    spStrCopy(buf, sizeof(buf), spGetHomeDir());
    findDocumentDir(document_list, buf, sizeof(buf));

    return strclone(buf);
}
#else /* defined(IPHONE) */
char *xspGetDocumentDir(spFileKind file_kind)
{
    spBool flag = SP_FALSE;
    char buf[SP_MAX_PATHNAME];
    static char *document_list[] =
    {
        "Documents",
        "documents",
        "docs",
        "doc",
        NULL,
    };
    static char *source_list[] =
    {
        "Sources",
        "sources",
        "source",
        "srcs",
        "src",
        NULL,
    };
    static char *config_list[] =
    {
        "Configs",
        "configs",
        "config",
        "etc",
        NULL,
    };
    static char *include_list[] =
    {
        "Include",
        "Includes",
        "include",
        "includes",
        NULL,
    };
    static char *picture_list[] =
    {
        "My Pictures",
        "Pictures",
        "pictures",
        "picture",
        "pic",
        NULL,
    };
    static char *figure_list[] =
    {
        "My Figures",
        "Figures",
        "figures",
        "figure",
        "fig",
        NULL,
    };
    static char *music_list[] =
    {
        "My Music",
        "Music",
        "music",
        NULL,
    };
    static char *sound_list[] =
    {
        "My Sounds",
        "My Sound",
        "Sounds",
        "Sound",
        "sounds",
        "sound",
        NULL,
    };
    static char *movie_list[] =
    {
        "My Movies",
        "My Movie",
        "Movies",
        "movies",
        "movie",
        NULL,
    };
    static char *application_list[] =
    {
        "My Applications",
        "My Application",
        "Applications",
        "applications",
        "Application",
        "application",
        "bin",
        NULL,
    };
    static char *library_list[] =
    {
        "Libraries",
        "Library",
        "libraries",
        "library",
        "lib",
        NULL,
    };
    static char *archive_list[] =
    {
        "Downloads",
        "Download",
        "downloads",
        "download",
        "Archives",
        "Archive",
        "archive",
        "Tools",
        "Tool",
        "tools",
        "tool",
        NULL,
    };
    static char *uri_list[] =
    {
        "My Sites",
        "My Site",
        "Sites",
        "Site",
        "URIs",
        "URI",
        "URLs",
        "URL",
        "Sites",
        "Site",
        NULL,
    };

#if defined(_WIN32)
    switch (file_kind) {
      case SP_FILE_KIND_DOCUMENT:
        flag = getSpecialFolderWin(CSIDL_PERSONAL, buf);
        break;

#if defined(CSIDL_PICTURES)
      case SP_FILE_KIND_PICTURE:
        flag = getSpecialFolderWin(CSIDL_PICTURES, buf);
        break;
#endif

      default:
        break;
    }
#endif
    
    if (flag == SP_FALSE) {
#if defined(_WIN32)
        getSpecialFolderWin(CSIDL_PERSONAL, buf);
#else
        spStrCopy(buf, sizeof(buf), spGetHomeDir());
#endif
        flag = SP_FALSE;
        
        switch (file_kind) {
          case SP_FILE_KIND_DOCUMENT:
          case SP_FILE_KIND_SOURCE:
          case SP_FILE_KIND_CONFIG:
          case SP_FILE_KIND_INCLUDE:
            switch (file_kind) {
              case SP_FILE_KIND_SOURCE:
                flag = findDocumentDir(source_list, buf, sizeof(buf));
                break;
              case SP_FILE_KIND_CONFIG:
                flag = findDocumentDir(config_list, buf, sizeof(buf));
                break;
              case SP_FILE_KIND_INCLUDE:
                flag = findDocumentDir(include_list, buf, sizeof(buf));
                break;
              default:
                break;
            }
            if (flag == SP_FALSE) {
                findDocumentDir(document_list, buf, sizeof(buf));
            }
            break;
            
          case SP_FILE_KIND_PICTURE:
          case SP_FILE_KIND_FIGURE:
            switch (file_kind) {
              case SP_FILE_KIND_FIGURE:
                flag = findDocumentDir(figure_list, buf, sizeof(buf));
                break;
              default:
                break;
            }
            if (flag == SP_FALSE) {
                findDocumentDir(picture_list, buf, sizeof(buf));
            }
            break;
            
          case SP_FILE_KIND_MUSIC:
          case SP_FILE_KIND_SOUND:
            switch (file_kind) {
              case SP_FILE_KIND_SOUND:
                flag = findDocumentDir(sound_list, buf, sizeof(buf));
                break;
              default:
                break;
            }
            if (flag == SP_FALSE) {
                findDocumentDir(music_list, buf, sizeof(buf));
            }
            break;
            
          case SP_FILE_KIND_MOVIE:
            findDocumentDir(movie_list, buf, sizeof(buf));
            break;
            
          case SP_FILE_KIND_APPLICATION:
            findDocumentDir(application_list, buf, sizeof(buf));
            break;
            
          case SP_FILE_KIND_LIBRARY:
            findDocumentDir(library_list, buf, sizeof(buf));
            break;
            
          case SP_FILE_KIND_ARCHIVE:
            findDocumentDir(archive_list, buf, sizeof(buf));
            break;
            
          case SP_FILE_KIND_URI:
            findDocumentDir(uri_list, buf, sizeof(buf));
            break;
            
          default:
            break;
        }
    }
    
    return strclone(buf);
}
#endif /* !defined(IPHONE) */
#endif /* !defined(ANDROID) */

double spRound(double x)
{
    double y;

    y = floor(x + 0.5);

    return y;
}

double spFix(double x)
{
#if defined(MACOS) && !powerc
    long double n;
#else
    double n;
#endif

    modf(x, &n);

    return n;
}

double spFrac(double x)
{
#if defined(MACOS) && !powerc
    long double n;
#else
    double n;
#endif

    return modf(x, &n);
}

double spRem(double x, double y)
{
#if defined(MACOS) && !powerc
    long double n;
#else
    double n;
#endif
    double z;

    z = modf(x / y, &n);
    z = x - n * y;

    return z;
}

long spFactorial(int n)
{
    long k = 1;
    long m;
    
    if (n > 0) {
        m = (long)n;
        while (1) {
            k *= m;
            m--;
            if (m <= 1) {
                break;
            }
        }
    }

    return k;
}

void spNFtos(char *buf, int bufsize, double x)
{
    int j;
    int flag = 0;
    double n;
    double pn;
    double xi;
    char sxi[SP_MAX_LINE];

    if (x == 0.0) {
        spStrCopy(buf, 2, "0");
    } else {
        pn = floor(log10(fabs(x)));
        n = pow(10.0, pn);

        if (fabs(pn) >= 4) {
            xi = x / n;
            flag = 1;
        } else {
            xi = x;
        }

#if defined(HAVE_SPRINTF_S)
        sprintf_s(sxi, sizeof(sxi), "%f", xi);
#elif defined(HAVE_SNPRINTF)
        snprintf(sxi, sizeof(sxi), "%f", xi);
#else
        sprintf(sxi, "%f", xi);
#endif
        for (j = (int)strlen(sxi) - 1; j >= 0; j--) {
            if (sxi[j] == '.') {
                sxi[j] = NUL;
                break;
            } else if (sxi[j] == '-' || sxi[j] == '+') {
                sxi[j + 2] = NUL;
                break;
            } else if (sxi[j] != '0') {
                sxi[j + 1] = NUL;
                break;
            }
        }

        if (flag == 1) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, bufsize, "%se%.0f", sxi, pn);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, bufsize, "%se%.0f", sxi, pn);
#else
            sprintf(buf, "%se%.0f", sxi, pn);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, bufsize, "%s", sxi);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, bufsize, "%s", sxi);
#else
            sprintf(buf, "%s", sxi);
#endif
        }
    }

    return;
}

void spFtos(char *buf, double x)
{
    spNFtos(buf, SP_MAX_LINE, x);
}

/*
 *        next power of 2
 */
long spNextPow2(long n)
{
    long p;
    long value;

    for (p = 1;; p++) {
        value = POW2(p);
        if (value >= n) {
            break;
        }
    }

    return p;
}

int spGetBitByte(int samp_bit, spBool *float_flag)
{
    int byte;

    if (samp_bit > 33) {
        byte = sizeof(double);
    } else if (samp_bit >= 33) {
        byte = sizeof(float);
    } else if (samp_bit >= 24) {
        byte = sizeof(long);
    } else {
        byte = sizeof(short);
    }

    if (float_flag != NULL) {
        if (samp_bit >= 33) {
            *float_flag = SP_TRUE;
        } else {
            *float_flag = SP_FALSE;
        }
    }
    
    return byte;
}

double spGetBitNormalizeFactor(int samp_bit)
{
    double factor = 1.0;
    
    if (samp_bit >= 33) {
        factor = 1.0;
    } else if (samp_bit >= 32) {
        factor = 2147483648.0;
    } else if (samp_bit >= 24) {
        factor = 8388608.0;
    } else if (samp_bit >= 16){
        factor = 32768.0;
    } else if (samp_bit >= 8){
        factor = 128.0;
    }

    spDebug(50, "spGetDoubleNormalizeFactor", "factor = %f\n", factor);
    
    return factor;
}

/* 8,16:short  24,32:long  33:float  64:double */
long _spConvertDoubleToBitWeighted(double *idata, long length, char *odata, int samp_bit, double weight)
{
    long k;
    short *sbuf;
    long *lbuf;
    float *fbuf;
    double *dbuf;

    if (idata == NULL || odata == NULL || length <= 0) return 0;

    if (samp_bit >= 64 && weight == 1.0) {
        memmove(odata, idata, length * sizeof(double));
    } else {
        sbuf = (short *)odata;
        lbuf = (long *)odata;
        fbuf = (float *)odata;
        dbuf = (double *)odata;
        
        for (k = 0; k < length; k++) {
            if (samp_bit >= 64) {
                dbuf[k] = weight * idata[k];
            } else if (samp_bit >= 33) {
                fbuf[k] = (float)(weight * idata[k]);
            } else if (samp_bit >= 24) {
                lbuf[k] = (long)spRound(weight * idata[k]);
            } else {
                sbuf[k] = (short)spRound(weight * idata[k]);
            }
        }
    }
        
    return length;
}

long spConvertDoubleToBitWeighted(double *idata, long length, void *odata, int samp_bit, double weight)
{
    return _spConvertDoubleToBitWeighted(idata, length, (char *)odata, samp_bit, weight);
}

long _spConvertDoubleToBit(double *idata, long length, char *odata, int samp_bit, double max)
{
    return _spConvertDoubleToBitWeighted(idata, length, odata, samp_bit, spGetBitNormalizeFactor(samp_bit) / max);
}

long spConvertDoubleToBit(double *idata, long length, void *odata, int samp_bit, double max)
{
    return _spConvertDoubleToBit(idata, length, (char *)odata, samp_bit, max);
}

long _spConvertDoubleToClippedBitWeighted(double *idata, long length, char *odata, int samp_bit, double weight)
{
    long k;
    short *sbuf;
    long *lbuf;
    float *fbuf;
    double *dbuf;
    float fvalue;
    double dvalue;

    if (idata == NULL || odata == NULL || length <= 0) return 0;

    if (samp_bit >= 64 && weight == 1.0) {
        memmove(odata, idata, length * sizeof(double));
    } else {
        sbuf = (short *)odata;
        lbuf = (long *)odata;
        fbuf = (float *)odata;
        dbuf = (double *)odata;
        
        for (k = 0; k < length; k++) {
            if (samp_bit >= 64) {
                dvalue = weight * idata[k];
                dbuf[k] = MAX(MIN(dvalue, 1.0), -1.0);
            } else if (samp_bit >= 33) {
                fvalue = (float)(weight * idata[k]);
                fbuf[k] = MAX(MIN(fvalue, 1.0f), -1.0f);
            } else if (samp_bit >= 24) {
                dvalue = weight * idata[k];
                if (samp_bit >= 32) {
                    dvalue = MAX(MIN(dvalue, 2147483647.0), -2147483648.0);
                } else {
                    dvalue = MAX(MIN(dvalue, 8388607.0), -8388608.0);
                }
                lbuf[k] = (long)spRound(dvalue);
            } else {
                dvalue = weight * idata[k];
                dvalue = MAX(MIN(dvalue, 32767.0), -32768.0);
                sbuf[k] = (short)spRound(dvalue);
            }
        }
    }
        
    return length;
}

long spConvertDoubleToClippedBitWeighted(double *idata, long length, void *odata, int samp_bit, double weight)
{
    return _spConvertDoubleToClippedBitWeighted(idata, length, (char *)odata, samp_bit, weight);
}

long _spConvertDoubleToClippedBit(double *idata, long length, char *odata, int samp_bit, double max)
{
    return _spConvertDoubleToClippedBitWeighted(idata, length, odata, samp_bit, spGetBitNormalizeFactor(samp_bit) / max);
}

long spConvertDoubleToClippedBit(double *idata, long length, void *odata, int samp_bit, double max)
{
    return _spConvertDoubleToClippedBit(idata, length, (char *)odata, samp_bit, max);
}

long _spConvertBitToDoubleWeighted(char *idata, long length, double *odata, int samp_bit, double weight)
{
    long k;
    short *sbuf;
    long *lbuf;
    float *fbuf;
    double *dbuf;

    if (idata == NULL || odata == NULL || length <= 0) return 0;

    if (samp_bit >= 64 && weight == 1.0) {
        memmove(odata, idata, length * sizeof(double));
    } else {
        sbuf = (short *)idata;
        lbuf = (long *)idata;
        fbuf = (float *)idata;
        dbuf = (double *)idata;
        
        for (k = 0; k < length; k++) {
            if (samp_bit >= 64) {
                odata[k] = weight * dbuf[k];
            } else if (samp_bit >= 33) {
                odata[k] = weight * (double)fbuf[k];
            } else if (samp_bit >= 24) {
                odata[k] = weight * (double)lbuf[k];
            } else {
                odata[k] = weight * (double)sbuf[k];
            }
        }
    }
        
    return length;
}

long spConvertBitToDoubleWeighted(void *idata, long length, double *odata, int samp_bit, double weight)
{
    return _spConvertBitToDoubleWeighted((char *)idata, length, odata, samp_bit, weight);
}

long _spConvertBitToDouble(char *idata, long length, double *odata, int samp_bit, double max)
{
    return _spConvertBitToDoubleWeighted(idata, length, odata, samp_bit, max / spGetBitNormalizeFactor(samp_bit));
}

long spConvertBitToDouble(void *idata, long length, double *odata, int samp_bit, double max)
{
    return _spConvertBitToDouble((char *)idata, length, odata, samp_bit, max);
}

typedef struct _spPaperSizeMapping {
    spPaperSize paper_size;
    char *name;
    double width_mm;
    double height_mm;
} spPaperSizeMapping;

/* reference : http://en.wikipedia.org/wiki/Paper_size */
static spPaperSizeMapping sp_paper_size_mapping[] = {
    {SP_PAPER_SIZE_A0, "A0", 841.0, 1189.0},
    {SP_PAPER_SIZE_A1, "A1", 594.0, 841.0},
    {SP_PAPER_SIZE_A2, "A2", 420.0, 594.0},
    {SP_PAPER_SIZE_A3, "A3", 297.0, 420.0},
    {SP_PAPER_SIZE_A4, "A4", 210.0, 297.0},
    {SP_PAPER_SIZE_A5, "A5", 148.0, 210.0},
    {SP_PAPER_SIZE_A6, "A6", 105.0, 148.0},
    {SP_PAPER_SIZE_2A0, "2A0", 1189.0, 1682.0},
    {SP_PAPER_SIZE_4A0, "4A0", 1682.0, 2378.0},

    {SP_PAPER_SIZE_B0, "ISO B0", 1000.0, 1414.0},
    {SP_PAPER_SIZE_B1, "ISO B1", 707.0, 1000.0},
    {SP_PAPER_SIZE_B2, "ISO B2", 500.0, 707.0},
    {SP_PAPER_SIZE_B3, "ISO B3", 353.0, 500.0},
    {SP_PAPER_SIZE_B4, "ISO B4", 250.0, 353.0},
    {SP_PAPER_SIZE_B5, "ISO B5", 176.0, 250.0},
    {SP_PAPER_SIZE_B6, "ISO B6", 125.0, 176.0},

    {SP_PAPER_SIZE_JIS_B0, "JIS B0", 1030.0, 1456.0},
    {SP_PAPER_SIZE_JIS_B1, "JIS B1", 728.0, 1030.0},
    {SP_PAPER_SIZE_JIS_B2, "JIS B2", 515.0, 728.0},
    {SP_PAPER_SIZE_JIS_B3, "JIS B3", 364.0, 515.0},
    {SP_PAPER_SIZE_JIS_B4, "JIS B4", 257.0, 364.0},
    {SP_PAPER_SIZE_JIS_B5, "JIS B5", 182.0, 257.0},
    {SP_PAPER_SIZE_JIS_B6, "JIS B6", 128.0, 182.0},
    
    {SP_PAPER_SIZE_LETTER, "Letter", 216.0, 279.0},
    {SP_PAPER_SIZE_GOVERNMENT_LETTER, "Government Letter", 203.0, 267.0},
    {SP_PAPER_SIZE_LEGAL, "Legal", 216.0, 356.0},
    {SP_PAPER_SIZE_TABLOID, "Tabloid", 279.0, 432.0},
    {SP_PAPER_SIZE_HAGAKI, "Japanese Hagaki", 100.0, 148.0},
    {SP_PAPER_SIZE_STATEMENT, "Statement", 140.0, 216.0},
    {SP_PAPER_SIZE_QUARTO, "Quarto", 203.0, 254.0},
    {SP_PAPER_SIZE_FOOLSCAP, "Foolscap, Folio", 203.0, 330.0},
    {SP_PAPER_SIZE_EXECUTIVE, "Executive, Monarch", 184.0, 267.0},
    
    {SP_PAPER_SIZE_ANSI_A, "ANSI A", 216.0, 279.0}, /* Letter */
    {SP_PAPER_SIZE_ANSI_B, "ANSI B", 279.0, 432.0}, /* Tabloid */
    {SP_PAPER_SIZE_ANSI_C, "ANSI C", 432.0, 559.0},
    {SP_PAPER_SIZE_ANSI_D, "ANSI D", 559.0, 864.0},
    {SP_PAPER_SIZE_ANSI_E, "ANSI E", 864.0, 1118.0},

    {SP_PAPER_SIZE_ARCH_A, "Arch A", 229.0, 305.0},
    {SP_PAPER_SIZE_ARCH_B, "Arch B", 305.0, 457.0},
    {SP_PAPER_SIZE_ARCH_C, "Arch C", 457.0, 610.0},
    {SP_PAPER_SIZE_ARCH_D, "Arch D", 610.0, 914.0},
    {SP_PAPER_SIZE_ARCH_E, "Arch E", 914.0, 1219.0},
    {SP_PAPER_SIZE_ARCH_E1, "Arch E1", 762.0, 1067.0},

    {SP_PAPER_SIZE_UNKNOWN, "", 0.0, 0.0},
};

spBool spGetPaperDimensions(spPaperSize paper_size, spPaperOrientation paper_orientation,
                            double *width_mm, double *height_mm)
{
    int i;

    for (i = 0; sp_paper_size_mapping[i].paper_size != SP_PAPER_SIZE_UNKNOWN; i++) {
        if (sp_paper_size_mapping[i].paper_size == paper_size) {
            if (paper_orientation == SP_PAPER_ORIENTATION_LANDSCAPE
                || paper_orientation == SP_PAPER_ORIENTATION_SEASCAPE) {
                if (width_mm != NULL) *width_mm = sp_paper_size_mapping[i].height_mm;
                if (height_mm != NULL) *height_mm = sp_paper_size_mapping[i].width_mm;
            } else {
                if (width_mm != NULL) *width_mm = sp_paper_size_mapping[i].width_mm;
                if (height_mm != NULL) *height_mm = sp_paper_size_mapping[i].height_mm;
            }
        
            return SP_TRUE;
        }
    }

    return SP_FALSE;
}

void getPaperDimensionString(spPaperSizeMapping *mapping, unsigned long options, char *buf, int buf_size)
{
    double width, height;
    double width_r, height_r;
    char unit_str[8];
    char width_str[256];
    char height_str[256];

    width = mapping->width_mm;
    height = mapping->height_mm;

    unit_str[0] = NUL;
    
    if (options & SP_PAPER_SIZE_STRING_INCH) {
        width /= 25.4;
        height /= 25.4;
        if (!(options & SP_PAPER_SIZE_STRING_NO_UNIT)) {
            if (options & SP_PAPER_SIZE_STRING_UNIT_SYMBOL) {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(unit_str, sizeof(unit_str), "\"");
#elif defined(HAVE_SNPRINTF)
                snprintf(unit_str, sizeof(unit_str), "\"");
#else
                sprintf(unit_str, "\"");
#endif
            } else {
                if (options & SP_PAPER_SIZE_STRING_NO_SPACE) {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(unit_str, sizeof(unit_str), "in");
#elif defined(HAVE_SNPRINTF)
                    snprintf(unit_str, sizeof(unit_str), "in");
#else
                    sprintf(unit_str, "in");
#endif
                } else {
#if defined(HAVE_SPRINTF_S)
                    sprintf_s(unit_str, sizeof(unit_str), " in");
#elif defined(HAVE_SNPRINTF)
                    snprintf(unit_str, sizeof(unit_str), " in");
#else
                    sprintf(unit_str, " in");
#endif
                }
            }
        }
    } else {
        if (!(options & SP_PAPER_SIZE_STRING_NO_UNIT)) {
            if (options & SP_PAPER_SIZE_STRING_NO_SPACE) {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(unit_str, sizeof(unit_str), "mm");
#elif defined(HAVE_SNPRINTF)
                snprintf(unit_str, sizeof(unit_str), "mm");
#else
                sprintf(unit_str, "mm");
#endif
            } else {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(unit_str, sizeof(unit_str), " mm");
#elif defined(HAVE_SNPRINTF)
                snprintf(unit_str, sizeof(unit_str), " mm");
#else
                sprintf(unit_str, " mm");
#endif
            }
        }
    }
        
    
    width_r = spRound(width);
    if (FABS(width - width_r) < 0.1) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(width_str, sizeof(width_str), "%.0f", width_r);
#elif defined(HAVE_SNPRINTF)
        snprintf(width_str, sizeof(width_str), "%.0f", width_r);
#else
        sprintf(width_str, "%.0f", width_r);
#endif
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(width_str, sizeof(width_str), "%.1f", spRound(10.0 * width) / 10.0);
#elif defined(HAVE_SNPRINTF)
        snprintf(width_str, sizeof(width_str), "%.1f", spRound(10.0 * width) / 10.0);
#else
        sprintf(width_str, "%.1f", spRound(10.0 * width) / 10.0);
#endif
    }
    height_r = spRound(height);
    if (FABS(height - height_r) < 0.1) {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(height_str, sizeof(height_str), "%.0f", height_r);
#elif defined(HAVE_SNPRINTF)
        snprintf(height_str, sizeof(height_str), "%.0f", height_r);
#else
        sprintf(height_str, "%.0f", height_r);
#endif
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(height_str, sizeof(height_str), "%.1f", spRound(10.0 * height) / 10.0);
#elif defined(HAVE_SNPRINTF)
        snprintf(height_str, sizeof(height_str), "%.1f", spRound(10.0 * height) / 10.0);
#else
        sprintf(height_str, "%.1f", spRound(10.0 * height) / 10.0);
#endif
    }

    if (options & SP_PAPER_SIZE_STRING_NO_SPACE) {
        if (options & SP_PAPER_SIZE_STRING_UNIT_BOTH) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%s%sx%s%s", width_str, unit_str, height_str, unit_str);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%s%sx%s%s", width_str, unit_str, height_str, unit_str);
#else
            sprintf(buf, "%s%sx%s%s", width_str, unit_str, height_str, unit_str);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%sx%s%s", width_str, height_str, unit_str);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%sx%s%s", width_str, height_str, unit_str);
#else
            sprintf(buf, "%sx%s%s", width_str, height_str, unit_str);
#endif
        }
    } else {
        if (options & SP_PAPER_SIZE_STRING_UNIT_BOTH) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%s%s x %s%s", width_str, unit_str, height_str, unit_str);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%s%s x %s%s", width_str, unit_str, height_str, unit_str);
#else
            sprintf(buf, "%s%s x %s%s", width_str, unit_str, height_str, unit_str);
#endif
        } else {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(buf, buf_size, "%s x %s%s", width_str, height_str, unit_str);
#elif defined(HAVE_SNPRINTF)
            snprintf(buf, buf_size, "%s x %s%s", width_str, height_str, unit_str);
#else
            sprintf(buf, "%s x %s%s", width_str, height_str, unit_str);
#endif
        }
    }

    return;
}

spBool spGetPaperSizeString(spPaperSize paper_size, unsigned long options, char *buf, int buf_size)
{
    int i;
    int len;

    for (i = 0; sp_paper_size_mapping[i].paper_size != SP_PAPER_SIZE_UNKNOWN; i++) {
        if (sp_paper_size_mapping[i].paper_size == paper_size) {
            spStrCopy(buf, buf_size, sp_paper_size_mapping[i].name);

            if (options & SP_PAPER_SIZE_STRING_INCLUDE_SIZE) {
                spStrCat(buf, buf_size, " (");
                len = (int)strlen(buf);
                getPaperDimensionString(&sp_paper_size_mapping[i], options, buf + len, buf_size - len);
                spStrCat(buf, buf_size, ")");
            }
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

spBool spGetPaperDimensionString(spPaperSize paper_size, unsigned long options, char *buf, int buf_size)
{
    int i;
    
    for (i = 0; sp_paper_size_mapping[i].paper_size != SP_PAPER_SIZE_UNKNOWN; i++) {
        if (sp_paper_size_mapping[i].paper_size == paper_size) {
            getPaperDimensionString(&sp_paper_size_mapping[i], options, buf, buf_size);
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

spBool spGetPaperFromDimensions(double width_mm, double height_mm,
                                spPaperSize *paper_size, spPaperOrientation *paper_orientation)
{
    int i;

    width_mm = spRound(width_mm);
    height_mm = spRound(height_mm);
    
    for (i = 0; sp_paper_size_mapping[i].paper_size != SP_PAPER_SIZE_UNKNOWN; i++) {
        if (FABS(width_mm - sp_paper_size_mapping[i].width_mm) < 2.0
            && FABS(height_mm - sp_paper_size_mapping[i].height_mm) < 2.0) {
            if (paper_size != NULL) *paper_size = sp_paper_size_mapping[i].paper_size;
            if (paper_orientation != NULL) *paper_orientation = SP_PAPER_ORIENTATION_PORTRAIT;
            return SP_TRUE;
        }
    }

    for (i = 0; sp_paper_size_mapping[i].paper_size != SP_PAPER_SIZE_UNKNOWN; i++) {
        if (FABS(height_mm - sp_paper_size_mapping[i].width_mm) < 2.0
            && FABS(width_mm - sp_paper_size_mapping[i].height_mm) < 2.0) {
            if (paper_size != NULL) *paper_size = sp_paper_size_mapping[i].paper_size;
            if (paper_orientation != NULL) *paper_orientation = SP_PAPER_ORIENTATION_LANDSCAPE;
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

#if defined(MACOS)
spBool spIsAltiVecSupportedMac(void)
{
#if defined(__MACH__)
    int selectors[2] = {CTL_HW, HW_VECTORUNIT};
    size_t length;
    static int vectorAvailable = -1;

    if (vectorAvailable == -1) {
        length = sizeof(vectorAvailable);
        vectorAvailable = 0;
        if (sysctl(selectors, 2, &vectorAvailable, &length, NULL, 0) == 0) {
            spDebug(10, "spIsAltiVecSupportedMac", "vectorAvailable = %d\n", vectorAvailable);
        } else {
            spDebug(10, "spIsAltiVecSupportedMac", "sysctl error\n");
        }
    }
    
    if (vectorAvailable) {
        return SP_TRUE;
    }
#elif !defined(IPHONE)
    SInt32 response;
    
    if (Gestalt(gestaltPowerPCProcessorFeatures, &response) == noErr) {
        spDebug(10, "spIsAltiVecSupportedMac", "Gestalt response = %ld\n", response);
        if (response & (1 << gestaltPowerPCHasVectorInstructions)) {
            return SP_TRUE;
        }
    } else {
        spDebug(10, "spIsAltiVecSupportedMac", "Gestalt error\n");
    }
#endif

    spDebug(10, "spIsAltiVecSupportedMac", "AltiVec not supported\n");
    
    return SP_FALSE;
}

spBool spIsAltiVecDoubleSupportedMac(void)
{
#if defined(MACOSX) && !defined(IPHONE)
    if (spIsAltiVecSupportedMac() == SP_TRUE) {
        SInt32 response;
        
        if (Gestalt(gestaltSystemVersion, &response) == noErr
            && response >= 0x00001040) {
            return SP_TRUE;
        }
    }
#endif
    
    return SP_FALSE;
}

#if defined(MACOS9)
#define SP_MACOS9_DEFINE_MATHF_FUNC(name) \
float name ## f(float x) { \
    return (float) name ((double)x); \
}

SP_MACOS9_DEFINE_MATHF_FUNC(cos)
SP_MACOS9_DEFINE_MATHF_FUNC(sin)
SP_MACOS9_DEFINE_MATHF_FUNC(tan)
SP_MACOS9_DEFINE_MATHF_FUNC(acos)
SP_MACOS9_DEFINE_MATHF_FUNC(asin)
SP_MACOS9_DEFINE_MATHF_FUNC(atan)
#if 0
SP_MACOS9_DEFINE_MATHF_FUNC(acos)
SP_MACOS9_DEFINE_MATHF_FUNC(asin)
SP_MACOS9_DEFINE_MATHF_FUNC(atan)
#endif
SP_MACOS9_DEFINE_MATHF_FUNC(cosh)
SP_MACOS9_DEFINE_MATHF_FUNC(sinh)
SP_MACOS9_DEFINE_MATHF_FUNC(tanh)
SP_MACOS9_DEFINE_MATHF_FUNC(exp)
SP_MACOS9_DEFINE_MATHF_FUNC(log)
SP_MACOS9_DEFINE_MATHF_FUNC(log10)
SP_MACOS9_DEFINE_MATHF_FUNC(sqrt)
SP_MACOS9_DEFINE_MATHF_FUNC(fabs)
SP_MACOS9_DEFINE_MATHF_FUNC(floor)
SP_MACOS9_DEFINE_MATHF_FUNC(ceil)

float atan2f(float y, float x) {
    return (float)atan2((double)y, (double)x);
}
float powf(float x, float y) {
    return (float)pow((double)x, (double)y);
}
#endif
#endif
