/*
 *        spPSFile.c
 */

#include <stdio.h>
#include <math.h>

#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spPSFile.h>

#define PS_UNIT 10
#define PS_MAX_PATH_COUNT 300
#define PS_CURRENT_POSITION_NONE (-5000)

struct _spPSFile
{
    spBool color;
    spBool landscape;
    spBool eps;
    spBool half_size;
    spBool screen_coordinate;

    spBool double_sided;
    spBool tumble;
    int num_copies;
    spBool collate;

    double unit;
    
    char creator[SP_MAX_LINE];
    char lang[SP_MAX_LINE];
    char font[SP_MAX_LINE];
    int font_size;
    int char_height;

    double left_margin;
    double bottom_margin;
    double client_width;
    double client_height;
    double xfactor;
    double yfactor;
    
    int angle;
    spPSAlignment alignment;

    spPSPolygonId polygon_id;
    spBool polygon_state;

    int page;
    int path_count;
    spBool relative_state;
    spBool end_state;
    
    int current_x, current_y;
    int current_line_width;
    spBool line_type_set;
    spPSLineType current_line_type;
    double current_r, current_g, current_b;
    
    FILE *outfile;
};

static double PSGetSizeFactor(spPSFile *psfile)
{
    if (psfile->half_size == SP_TRUE) {
        return 0.5;
    } else {
        return 1.0;
    }
}

static double PSPointToUnit(spPSFile *psfile, double point)
{
    return psfile->unit * point;
}

double spPSMMToPoint(double mm)
{
    return (mm * 72.0) / 25.4;
}

double spPSPointToMM(double point)
{
    return (25.4 * point) / 72.0;
}

static spPSFile *PSAlloc(void)
{
    spPSFile *psfile;

    psfile = xalloc(1, struct _spPSFile);
    memset(psfile, 0, sizeof(struct _spPSFile));

    spStrCopy(psfile->creator, sizeof(psfile->creator), SP_PS_DEFAULT_CREATOR);
    spStrCopy(psfile->lang, sizeof(psfile->lang), SP_PS_DEFAULT_LANG);
    spStrCopy(psfile->font, sizeof(psfile->font), SP_PS_DEFAULT_FONT);
    psfile->color = SP_TRUE;
    psfile->landscape = SP_FALSE;        
    psfile->eps = SP_FALSE;
    psfile->half_size = SP_FALSE;
    psfile->screen_coordinate = SP_FALSE;

    psfile->double_sided = SP_FALSE;
    psfile->tumble = SP_FALSE;
    psfile->num_copies = 1;
    psfile->collate = SP_TRUE;
    
    psfile->unit = PS_UNIT;
    psfile->font_size = SP_PS_DEFAULT_FONT_SIZE;
    psfile->char_height = (int)spRound(PSPointToUnit(psfile, psfile->font_size));

    psfile->left_margin = spPSMMToPoint(SP_PS_DEFAULT_MARGIN_MM);
    psfile->bottom_margin = spPSMMToPoint(SP_PS_DEFAULT_MARGIN_MM);
    psfile->client_width = 0.0;
    psfile->client_height = 0.0;
    psfile->xfactor = 1.0;
    psfile->yfactor = 1.0;
    psfile->alignment = SP_PS_ALIGNMENT_BEGINNING;

    psfile->polygon_id = SP_PS_POLYGON_NONE;        
    psfile->polygon_state = SP_FALSE;
    
    psfile->page = 0;        
    psfile->path_count = 0;         
    psfile->relative_state = SP_FALSE;
    psfile->end_state = SP_FALSE;
    
    psfile->current_x = 0;
    psfile->current_y = 0;
    psfile->current_line_width = 1;
    psfile->line_type_set = SP_FALSE;
    psfile->current_line_type = SP_PS_LINE_NONE;
    psfile->current_r = psfile->current_g = psfile->current_b = -1.0;
    
    psfile->outfile = NULL;
    
    return psfile;
}

static void PSFree(spPSFile *psfile)
{
    if (psfile != NULL) {
        xfree(psfile);
    }
}

static int PSGetEndPoint(spPSFile *psfile, double size, double offset)
{
    return (int)spRound(size * PSGetSizeFactor(psfile) + offset);
}

static void PSLineWidth(spPSFile *psfile, int width)
{
    fprintf(psfile->outfile, "/sppslinewidth %.3f def\n", (0.5 * PSPointToUnit(psfile, (double)width)));
    return;
}

static double PSGetCartesianY(spPSFile *psfile, double y)
{
    if (psfile->screen_coordinate == SP_TRUE) {
        double height;

        height = psfile->client_height / psfile->yfactor;

        return height - y;
    } else {
        return y;
    }
}

static void PSInit(spPSFile *psfile)
{
    if (psfile->eps == SP_FALSE) {
        fprintf(psfile->outfile,"%%!PS-Adobe-2.0\n");
    } else {
        fprintf(psfile->outfile,"%%!PS-Adobe-2.0 EPSF-2.0\n");
    }
    fprintf(psfile->outfile, "%%%%Creator: %s\n", psfile->creator);
    fprintf(psfile->outfile, "%%%%DocumentFonts: %s\n", psfile->font);
    fprintf(psfile->outfile, "%%%%BoundingBox: %.0f %.0f ", spRound(psfile->left_margin), spRound(psfile->bottom_margin));

    if (psfile->landscape == SP_FALSE) {
        fprintf(psfile->outfile,"%d %d\n", 
                PSGetEndPoint(psfile, psfile->client_width, psfile->left_margin), PSGetEndPoint(psfile, psfile->client_height, psfile->bottom_margin));
    } else {
        fprintf(psfile->outfile,"%d %d\n", 
                PSGetEndPoint(psfile, psfile->client_height, psfile->left_margin), PSGetEndPoint(psfile, psfile->client_width, psfile->bottom_margin));
    }
    if (psfile->eps == SP_FALSE) {
        fprintf(psfile->outfile,"%%%%Pages: (atend)\n");
    }

    fprintf(psfile->outfile,"%%%%EndComments\n");

    if (psfile->double_sided == SP_TRUE || psfile->num_copies > 1) {
        fprintf(psfile->outfile,"<< ");
        if (psfile->double_sided == SP_TRUE) {
            fprintf(psfile->outfile,"/Duplex true /Tumble %s\n",
                    (psfile->tumble == SP_FALSE ? "false" : "true"));
        }
        if (psfile->num_copies > 1) {
            fprintf(psfile->outfile,"    /NumCopies %d /Collate %s",
                    psfile->num_copies, (psfile->collate == SP_FALSE ? "false" : "true"));
        }
        fprintf(psfile->outfile,">> setpagedevice\n");
    }
    
    fprintf(psfile->outfile,"/sppsdict 50 dict def\nsppsdict begin\n");
    fprintf(psfile->outfile,"/Color %s def\n", psfile->color ? "true" : "false");
    PSLineWidth(psfile, psfile->current_line_width);        /* line width */
    /* left alignment show */
    fprintf(psfile->outfile, "/Lshow { currentpoint stroke moveto\n");
    fprintf(psfile->outfile, "    0 0 rmoveto show } def\n");
    /* right alignment show */
    fprintf(psfile->outfile, "/Rshow { currentpoint stroke moveto\n");
    fprintf(psfile->outfile, "    dup stringwidth pop neg 0 rmoveto show } def\n"); 
    /* center alignment show */
    fprintf(psfile->outfile, "/Cshow { currentpoint stroke moveto\n");
    fprintf(psfile->outfile, "    dup stringwidth pop -2 div 0 rmoveto show } def\n"); 

    /* normal line */
    fprintf(psfile->outfile, "/NL { stroke sppslinewidth setlinewidth } def\n");
    /* line types */
    fprintf(psfile->outfile, "/dashlen {%d mul} def\n", (int)spRound(PSPointToUnit(psfile, 1.0)));
    fprintf(psfile->outfile, "/Lsolid { NL [] 0 setdash } def\n");
    fprintf(psfile->outfile, "/Ldash { NL [6 dashlen 3 dashlen] 0 setdash } def\n");
    fprintf(psfile->outfile, "/Ldot { NL [1 dashlen 3 dashlen] 0 setdash } def\n");
    fprintf(psfile->outfile, "/Ldashdot { NL [1 dashlen 3 dashlen 6 dashlen 3 dashlen] 0 setdash } def\n");
    fprintf(psfile->outfile, "/Ldashdotdot { NL [1 dashlen 3 dashlen 6 dashlen 3 dashlen 1 dashlen 3 dashlen] 0 setdash } def\n");
    /* border line */
    fprintf(psfile->outfile, "/BL { stroke sppslinewidth 2 mul setlinewidth } def\n");
    fprintf(psfile->outfile, "/Lborder { BL [] 0 setdash } def\n");
    /* grid line */
    fprintf(psfile->outfile, "/GL { stroke sppslinewidth 2 div setlinewidth } def\n");
    fprintf(psfile->outfile, "/Lgrid { GL [1 dashlen 3 dashlen] 0 setdash } def\n");

    fprintf(psfile->outfile,"end\n%%%%EndProlog\n");

    return;
}

static void PSNewPage(spPSFile *psfile)
{
    psfile->page++;
    if (psfile->eps == SP_FALSE) {
        fprintf(psfile->outfile,"%%%%Page: %d %d\n", psfile->page, psfile->page);
    }
    fprintf(psfile->outfile,"sppsdict begin\ngsave\n");
    fprintf(psfile->outfile,"%.0f %.0f translate\n", psfile->left_margin, psfile->bottom_margin);
    fprintf(psfile->outfile,"%.3f %.3f scale\n",
            psfile->xfactor * PSGetSizeFactor(psfile) / PSPointToUnit(psfile, 1.0),
            psfile->yfactor * PSGetSizeFactor(psfile) / PSPointToUnit(psfile, 1.0));
    if (psfile->landscape == SP_TRUE) {
        fprintf(psfile->outfile,"90 rotate\n0 %d translate\n", (int)(-PSPointToUnit(psfile, psfile->client_height)));
    }
#if 0
    if (psfile->page == 1) {
        PSLineWidth(psfile, 1);
    }
#endif
    fprintf(psfile->outfile, "0 setgray\n");
    fprintf(psfile->outfile, "/%s findfont %d scalefont setfont\n", psfile->font, psfile->char_height);
    fprintf(psfile->outfile, "newpath\n");
    
    psfile->line_type_set = SP_FALSE;
    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;
    psfile->current_x = psfile->current_y = PS_CURRENT_POSITION_NONE;

    return;
}

static void PSEndPage(spPSFile *psfile)
{
    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }
    fprintf(psfile->outfile, "grestore\nend\nshowpage\n");
    
    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;
    
    return;
}

static void PSGetPointInUnit(spPSFile *psfile, int *x, int *y, spBool relative)
{
    double yc;
    
    *x = (int)spRound(PSPointToUnit(psfile, (double)*x));

    if (relative == SP_TRUE) {
        if (psfile->screen_coordinate == SP_TRUE) {
            yc = -(double)(*y);
        } else {
            yc = (double)(*y);
        }
    } else {
        yc = PSGetCartesianY(psfile, (double)*y);
    }
    *y = (int)spRound(PSPointToUnit(psfile, yc));
    
    return;
}

static void PSGetSizeInUnit(spPSFile *psfile, int *width, int *height)
{
    *width = (int)spRound(PSPointToUnit(psfile, (double)*width));
    *height = (int)spRound(PSPointToUnit(psfile, (double)*height));
    
    return;
}

spPSFile *spPSOpenFile(const char *filename, double client_width_mm, double client_height_mm, unsigned long options)
{
    spPSFile *psfile;
    FILE *fp;
    
    if ((fp = spOpenFile(filename, "w")) == NULL) {
        return NULL;
    }
    if ((psfile = PSAlloc()) == NULL) {
        spCloseFile(fp);
        return NULL;
    }
    psfile->outfile = fp;
    
    if (options & SP_PS_MONO) psfile->color = SP_FALSE;
    if (options & SP_PS_LANDSCAPE) psfile->landscape = SP_TRUE;
    if (options & SP_PS_EPS) psfile->eps = SP_TRUE;
    if (options & SP_PS_HALF_SIZE) psfile->half_size = SP_TRUE;
    if (options & SP_PS_SCREEN_COORDINATE) psfile->screen_coordinate = SP_TRUE;
    if (options & SP_PS_DOUBLE_SIDED) {
        if (options & SP_PS_TUMBLE) {
            psfile->tumble = SP_TRUE;
        }
        psfile->double_sided = SP_TRUE;
    }

    psfile->client_width = spPSMMToPoint(client_width_mm);
    psfile->client_height = spPSMMToPoint(client_height_mm);
    
    return psfile;
}

spBool spPSCloseFile(spPSFile *psfile)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    if (psfile->end_state == SP_FALSE) {
        spPSEndDrawing(psfile);
    }
    
    spCloseFile(psfile->outfile);
    PSFree(psfile);

    return SP_TRUE;
}
     
spBool spPSSetNumCopies(spPSFile *psfile, int num_copies, spBool collate)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->page > 0) return SP_FALSE;

    psfile->num_copies = MAX(num_copies, 1);
    psfile->collate = collate;
    
    return SP_TRUE;
}

spBool spPSSetScalingFactor(spPSFile *psfile, double xfactor, double yfactor)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->page > 0) return SP_FALSE;
    
    psfile->xfactor = xfactor;
    psfile->yfactor = yfactor;

    return SP_TRUE;
}

spBool spPSSetMargin(spPSFile *psfile, double left_margin_mm, double bottom_margin_mm)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->page > 0) return SP_FALSE;
    
    psfile->left_margin = spPSMMToPoint(left_margin_mm);
    psfile->bottom_margin = spPSMMToPoint(bottom_margin_mm);

    return SP_TRUE;
}

spBool spPSSetCreator(spPSFile *psfile, const char *creator)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->page > 0
        || creator == NULL) return SP_FALSE;
    
    spStrCopy(psfile->creator, sizeof(psfile->creator), creator);

    return SP_TRUE;
}

spBool spPSGetMargin(spPSFile *psfile, double *left_margin_mm, double *bottom_margin_mm)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    if (left_margin_mm != NULL) {
        *left_margin_mm = spPSPointToMM(psfile->left_margin);
    }
    if (bottom_margin_mm != NULL) {
        *bottom_margin_mm = spPSPointToMM(psfile->bottom_margin);
    }

    return SP_TRUE;
}

spBool spPSGetClientSize(spPSFile *psfile, int *client_width, int *client_height)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    if (client_width != NULL) {
        *client_width = (int)spRound(psfile->client_width / psfile->xfactor);
    }
    if (client_height != NULL) {
        *client_height = (int)spRound(psfile->client_height / psfile->yfactor);
    }

    return SP_TRUE;
}

spBool spPSGetPhisicalSize(spPSFile *psfile, double *client_width_mm, double *client_height_mm)
{
    double factor;
    
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    factor = PSGetSizeFactor(psfile);

    if (client_width_mm != NULL) {
        *client_width_mm = spPSPointToMM(psfile->client_width * factor);
    }
    if (client_height_mm != NULL) {
        *client_height_mm = spPSPointToMM(psfile->client_height * factor);
    }
    
    return SP_TRUE;
}

spBool spPSStartDrawing(spPSFile *psfile)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->end_state == SP_TRUE) return SP_FALSE;

    /* initialize */
    PSInit(psfile);

    /* new page */
    PSNewPage(psfile);

    return SP_TRUE;
}

spBool spPSEndDrawing(spPSFile *psfile)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->end_state == SP_TRUE) return SP_FALSE;
    
    /* end page */
    PSEndPage(psfile);

    /* end */
    fprintf(psfile->outfile,"%%%%Trailer\n");
    if (psfile->eps == SP_FALSE) {
        fprintf(psfile->outfile,"%%%%Pages: %d\n", psfile->page);
    }
    psfile->end_state = SP_TRUE;

    return SP_TRUE;
}

spBool spPSForwardPage(spPSFile *psfile)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->end_state == SP_TRUE) return SP_FALSE;

    /* end page */
    PSEndPage(psfile);

    /* new page */
    PSNewPage(psfile);

    return SP_TRUE;
}

spBool spPSSetLang(spPSFile *psfile, const char *lang)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    spStrCopy(psfile->lang, sizeof(psfile->lang), lang);

    return SP_TRUE;
}

spBool spPSSetLineType(spPSFile *psfile, spPSLineType line_type)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    if (psfile->line_type_set == SP_TRUE && psfile->current_line_type == line_type) {
        return SP_TRUE;
    }

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "currentpoint\n");
    }
    
    switch (line_type) {
      case SP_PS_LINE_SOLID:
        fprintf(psfile->outfile,"Lsolid\n");
        break;
      case SP_PS_LINE_DASH:
        fprintf(psfile->outfile,"Ldash\n");
        break;
      case SP_PS_LINE_DOT:
        fprintf(psfile->outfile,"Ldot\n");
        break;
      case SP_PS_LINE_DASH_DOT:
        fprintf(psfile->outfile,"Ldashdot\n");
        break;
      case SP_PS_LINE_DASH_DOT_DOT:
        fprintf(psfile->outfile,"Ldashdotdot\n");
        break;
      case SP_PS_LINE_BORDER:
        fprintf(psfile->outfile,"Lborder\n");
        break;
      case SP_PS_LINE_GRID:
        fprintf(psfile->outfile,"Lgrid\n");
        break;
      default:
        break;
    }

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, " moveto\n");
    }
    
    psfile->relative_state = SP_FALSE;
    psfile->line_type_set = SP_TRUE;
    psfile->current_line_type = line_type;
    psfile->path_count = 0;

    return SP_TRUE;
}

static spBool PSSetupLine(spPSFile *psfile)
{
    if (psfile->current_line_type == SP_PS_LINE_NONE || psfile->line_type_set == SP_FALSE) {
        if (psfile->current_line_type == SP_PS_LINE_NONE) {
            psfile->current_line_type = SP_PS_LINE_SOLID;
        }
        PSLineWidth(psfile, psfile->current_line_width);
        spPSSetLineType(psfile, psfile->current_line_type);
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spPSSetLineWidth(spPSFile *psfile, int width)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    if (psfile->current_line_width == width) {
        return SP_TRUE;
    }
    
    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "currentpoint stroke moveto\n");
    }
    
    PSLineWidth(psfile, width);
    psfile->current_line_width = width;

    return SP_TRUE;
}

static double PSRGBToGray(double r, double g, double b)
{
    /* NTSC */
    return 0.298912 * r / 255.0 + 0.586611 * g / 255.0 + 0.114478 * b / 255.0;
}

static spBool PSEqualColor(double x, double y)
{
    double diff;
    
    diff = 256 * FABS(x - y);

    if (diff < 1.0) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPSSetRGBColor(spPSFile *psfile, double r, double g, double b)
{
    double gray;
    
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;
    
    if (PSEqualColor(psfile->current_r, r) == SP_TRUE
        && PSEqualColor(psfile->current_g, g) == SP_TRUE
        && PSEqualColor(psfile->current_b, b) == SP_TRUE) {
        return SP_TRUE;
    }

    gray = PSRGBToGray(r, g, b);
    fprintf(psfile->outfile, "Color {%.3f %.3f %.3f setrgbcolor} {%.3f setgray} ifelse\n",
            r, g, b, gray);

    psfile->current_r = r;
    psfile->current_g = g;
    psfile->current_b = b;

    return SP_TRUE;
}

spBool spPSSetFontSize(spPSFile *psfile, int size)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;
    
    if (psfile->font_size == size) {
        return SP_TRUE;
    }

    if (size <= 0) {
        psfile->font_size = SP_PS_DEFAULT_FONT_SIZE;
    } else {
        psfile->font_size = size;
    }
    psfile->char_height = (int)spRound(PSPointToUnit(psfile, psfile->font_size));
    fprintf(psfile->outfile, "/%s findfont %d scalefont setfont\n", psfile->font, psfile->char_height);

    return SP_TRUE;
}

int spPSGetFontSize(spPSFile *psfile)
{
    if (psfile == NULL) return 0;

    return psfile->font_size;
}

#if 1
spBool spPSSetFont(spPSFile *psfile, const char *font)
{
    int i;
    char *p;
    const char *string;
    int size;
    char family_name[SP_MAX_LINE];
    char style_str[SP_MAX_LINE];
    spBool bold_flag, italic_flag, narrow_flag, fixed_flag;
    spBool success;
    
    if (psfile == NULL || strnone(font) || psfile->outfile == NULL) return SP_FALSE;
    
    size = 0;
    spStrCopy(family_name, sizeof(family_name), "");
    spStrCopy(style_str, sizeof(style_str), "");
    bold_flag = italic_flag = narrow_flag = fixed_flag = SP_FALSE;

    if (font[0] != '-') {
        spStrCopy(family_name, sizeof(family_name), font);
    } else {
        string = font;
        string++;
        i = 1;
        while ((string = strchr(string, '-')) != NULL) {
            string++;
            switch (i) {
              case 1:
                spStrCopy(family_name, sizeof(family_name), string);
                        
                if ((p = strchr(family_name, '-')) != NULL) {
                    *p = NUL;
                }
                break;
              case 2:
                if (strveq(string, "black") || strveq(string, "ultrabold")
                    || strveq(string, "heavy") || strveq(string, "extrabold")
                    || strveq(string, "bold") || strveq(string, "semibold")
                    || strveq(string, "demibold")) {
                    bold_flag = SP_TRUE;
                }
                break;
              case 3:
                if (strveq(string, "i") || strveq(string, "o")) {
                    italic_flag = SP_TRUE;
                }
                break;
              case 4:
                if (strvcaseeq(string, "narrow")) {
                    narrow_flag = SP_TRUE;
                }
                break;
              case 6:
#if defined(HAVE_SSCANF_S)
                sscanf_s(string, "%d", &size);
#else
                sscanf(string, "%d", &size);
#endif
                break;
              case 10:
                if (strveq(string, "m")) {
                    fixed_flag = SP_TRUE;
                }
                break;
              default:
                break;
            }
                
            i++;
        }
    }

    success = SP_TRUE;

    if (strnone(family_name)) {
        spStrCopy(family_name, sizeof(family_name), SP_PS_DEFAULT_FONT);
    } else {
        if (fixed_flag == SP_TRUE || strvcaseeq(family_name, "Courier")) {
            spStrCopy(family_name, sizeof(family_name), "Courier");
        } else if (strvcaseeq(family_name, "Century") || strvcaseeq(family_name, "NewCentury")
            || strvcaseeq(family_name, "new century")) {
            spStrCopy(family_name, sizeof(family_name), "NewCenturySchlbk");
        } else if (strvcaseeq(family_name, "Times")) {
            spStrCopy(family_name, sizeof(family_name), "Times");
        } else if (strvcaseeq(family_name, "Palatino")) {
            spStrCopy(family_name, sizeof(family_name), "Palatino");
        } else if (strvcaseeq(family_name, "Helvetica")
                   || strvcaseeq(family_name, "gothic") || strvcaseeq(family_name, "Arial")) {
            if (narrow_flag == SP_TRUE) {
                spStrCopy(family_name, sizeof(family_name), "Helvetica-Narrow");
            } else {
                spStrCopy(family_name, sizeof(family_name), "Helvetica");
            }
        }
    }
    if (bold_flag == SP_TRUE || italic_flag == SP_TRUE) {
        if (bold_flag == SP_TRUE && italic_flag == SP_TRUE) {
            if (strveq(family_name, "Helvetica") || streq(family_name, "Courier")) {
                spStrCopy(style_str, sizeof(style_str), "-BoldOblique");
            } else {
                spStrCopy(style_str, sizeof(style_str), "-BoldItalic");
            }
        } else if (italic_flag == SP_TRUE) {
            if (strveq(family_name, "Helvetica") || streq(family_name, "Courier")) {
                spStrCopy(style_str, sizeof(style_str), "-Oblique");
            } else {
                spStrCopy(style_str, sizeof(style_str), "-Italic");
            }
        } else if (bold_flag == SP_TRUE) {
            spStrCopy(style_str, sizeof(style_str), "-Bold");
        }
    } else {
        if (streq(family_name, "Times") || streq(family_name, "Palatino")
            || streq(family_name, "NewCenturySchlbk")) {
            spStrCopy(style_str, sizeof(style_str), "-Roman");
        }
    }
    spStrCopy(psfile->font, sizeof(psfile->font), family_name);
    spStrCat(psfile->font, sizeof(psfile->font), style_str);
    
    if (size > 0) {
        if (spPSSetFontSize(psfile, size) == SP_FALSE) {
            success = SP_FALSE;
        }
    }

    return success;
}

spBool spPSSetFontFamily(spPSFile *psfile, const char *font_family)
{
    if (psfile == NULL || psfile->outfile == NULL
        || strnone(font_family)) return SP_FALSE;

    spStrCopy(psfile->font, sizeof(psfile->font), font_family);

    return SP_TRUE;
}

const char *spPSGetFontFamily(spPSFile *psfile)
{
    if (psfile == NULL) return NULL;

    return psfile->font;
}
#endif

spBool spPSSetStringAlignment(spPSFile *psfile, spPSAlignment alignment)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    psfile->alignment = alignment;

    return SP_TRUE;
}

spBool spPSSetStringRotation(spPSFile *psfile, int angle)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    psfile->angle = angle;

    return SP_TRUE;
}

spBool spPSMoveTo(spPSFile *psfile, int x, int y)
{
    int dx, dy;

    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;
    
    PSSetupLine(psfile);
    
    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    
    dx = x - psfile->current_x;
    dy = y - psfile->current_y;

    if (dx == 0 && dy == 0 && psfile->relative_state == SP_TRUE) {
        return SP_TRUE;
    }

    if (psfile->relative_state == SP_TRUE){
        fprintf(psfile->outfile, "%d %d rmoveto\n", dx, dy);
    } else {
        fprintf(psfile->outfile, "%d %d moveto\n", x, y);
    }

    psfile->relative_state = SP_TRUE;
    psfile->path_count++;
    psfile->current_x = x;
    psfile->current_y = y;

    return SP_TRUE;
}

spBool spPSDrawLineTo(spPSFile *psfile, int x, int y)
{
    int dx, dy;

    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    
    dx = x - psfile->current_x;
    dy = y - psfile->current_y;

    if (dx == 0 && dy == 0) {
        return SP_TRUE;
    }

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "%d %d rlineto\n", dx, dy);
    } else {
        fprintf(psfile->outfile, "%d %d lineto\n", x, y);
    }

    psfile->relative_state = SP_TRUE;
    psfile->path_count++;
    psfile->current_x = x;
    psfile->current_y = y;
    if (psfile->path_count >= PS_MAX_PATH_COUNT) {
        fprintf(psfile->outfile, "currentpoint stroke moveto\n");
        psfile->path_count = 0;
    }
    
    return SP_TRUE;
}

spBool spPSDrawLine(spPSFile *psfile, int x1, int y1, int x2, int y2)
{
    if (spPSMoveTo(psfile, x1, y1)) {
        return spPSDrawLineTo(psfile, x2, y2);
    }

    return SP_FALSE;
}

static void normalDrawString(spPSFile *psfile, const char *string)
{
    char c;

    putc('(', psfile->outfile);
    c = *string++;
    while (c != '\0') {
        if ((c == '(') || (c == ')') || (c == '\\')) {
            putc('\\', psfile->outfile);
        }
        putc(c, psfile->outfile);
        c = *string++;
    }

    switch (psfile->alignment) {
      case SP_PS_ALIGNMENT_CENTER:
        fprintf(psfile->outfile, ") Cshow\n");
        break;
      case SP_PS_ALIGNMENT_END:
        fprintf(psfile->outfile, ") Rshow\n");
        break;
      default:
        fprintf(psfile->outfile, ") Lshow\n");
        break;
    }
    
    return;
}

static spBool isMBTailCandidate(const char *lang, unsigned prev_c, unsigned c)
{
    if (spEqLanguage(lang, "ja_JP.SJIS")) {
        if (((prev_c >= 0x81 && prev_c <= 0x9f) || (prev_c >= 0xe0 && prev_c <= 0xfc))
            && ((c >= 0x40 && c <= 0x7e) || (c >= 0x80 && c <= 0xfc))) {
            return SP_TRUE;
        }
    } else if (spEqLanguage(lang, "ja_JP.utf8")) {
        if ((prev_c & 0x80) && ((c & 0x80) && !(c & 0x40))) {
            return SP_TRUE;
        }
    } else {
        if ((prev_c >= 0x8e && prev_c <= 0xfe) && (c >= 0xa1 && c <= 0xfe)) {
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

static void japaneseDrawString(spPSFile *psfile, const char *string)
{
    unsigned char c, pc;
    int i;
    int cu, cl;
    int njp, neng;
    int continuation;
    spBool gothic_flag;
    char jp_font[SP_MAX_LINE];
    
    if (strvcaseeq(psfile->font, "Helvetica")) {
        gothic_flag = SP_TRUE;
    } else {
        gothic_flag = SP_FALSE;
    }

    if (spEqLanguage(psfile->lang, "ja_JP.SJIS")) {
        if (gothic_flag == SP_TRUE) {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_SJIS_GOTHIC_FONT);
        } else {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_SJIS_FONT);
        }
    } else if (spEqLanguage(psfile->lang, "ja_JP.utf8")) {
        if (gothic_flag == SP_TRUE) {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_UTF8JP_GOTHIC_FONT);
        } else {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_UTF8JP_FONT);
        }
    } else {
        if (gothic_flag == SP_TRUE) {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_EUCJP_GOTHIC_FONT);
        } else {
            spStrCopy(jp_font, sizeof(jp_font), SP_PS_DEFAULT_EUCJP_FONT);
        }
    }

    fprintf(psfile->outfile, "/strfontmap [ ");
    
    neng = 0;
    njp = 0;
    pc = *string++;
    while (pc != '\0') {
        c = *string++;

        if (c != '\0' && isMBTailCandidate(psfile->lang, pc, c) == SP_TRUE) {
            if (neng > 0) {
                fprintf(psfile->outfile, ") /%s ", psfile->font);
                neng = 0;
            }
            if (njp > 0) {
                putc(' ', psfile->outfile);
            } else {
                putc('<', psfile->outfile);
            }
            cu = pc & 0xff;
            cl = c & 0xff;

            fprintf(psfile->outfile, "%x %x", cu, cl);
            
            if (spEqLanguage(psfile->lang, "ja_JP.utf8") == SP_TRUE) {
                if (spIsUTF8First(pc, &continuation) == SP_TRUE) {
                    for (i = 1; i < continuation; i++) {
                        c = *string++;
                        fprintf(psfile->outfile, " %x", (0xff & c));
                    }
                }
            }
            
            c = *string++;

            njp++;
        } else {
            if (njp > 0) {
                fprintf(psfile->outfile, "> /%s ", jp_font);
                njp = 0;
            }
            if (neng <= 0) {
                putc('(', psfile->outfile);
            }
            if ((pc == '(') || (pc == ')') || (pc == '\\')) {
                putc('\\', psfile->outfile);
            }
            putc(pc, psfile->outfile);
            
            neng++;
        }

        pc = c;
    }

    if (neng > 0) {
        fprintf(psfile->outfile, ") /%s ", psfile->font);
    } else if (njp > 0) {
        fprintf(psfile->outfile, "> /%s ", jp_font);
    }
    
    fprintf(psfile->outfile, "] def\n");
    fprintf(psfile->outfile, "strfontmap length 2 sub /n exch def\n");

    if (psfile->alignment != SP_PS_ALIGNMENT_BEGINNING) {
        fprintf(psfile->outfile, "/totalstrwidth 0 def\n");
        fprintf(psfile->outfile, "0 2 n {\n");
        fprintf(psfile->outfile, "    /i exch def\n");
        fprintf(psfile->outfile, "    i 1 add strfontmap exch get findfont %d scalefont setfont\n", psfile->char_height);
        fprintf(psfile->outfile, "    strfontmap i get\n");
        fprintf(psfile->outfile, "    stringwidth pop totalstrwidth add\n");
        fprintf(psfile->outfile, "    /totalstrwidth exch def\n");
        fprintf(psfile->outfile, "} for\n");
        
        if (psfile->alignment == SP_PS_ALIGNMENT_CENTER) {
            fprintf(psfile->outfile, "totalstrwidth -2 div 0 rmoveto\n");
        } else {
            fprintf(psfile->outfile, "totalstrwidth neg 0 rmoveto\n");
        }
    }
    fprintf(psfile->outfile, "0 2 n {\n");
    fprintf(psfile->outfile, "    /i exch def\n");
    fprintf(psfile->outfile, "    i 1 add strfontmap exch get findfont %d scalefont setfont\n", psfile->char_height);
    fprintf(psfile->outfile, "    strfontmap i get show\n");
    fprintf(psfile->outfile, "} for\n");
    
    return;
}

spBool spPSDrawString(spPSFile *psfile, int x, int y, const char *string)
{
    if (psfile == NULL || psfile->outfile == NULL || strlen(string) <= 0) return SP_FALSE;

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }
    
    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    fprintf(psfile->outfile, "%d %d moveto\n", x, y);

    if (psfile->angle != 0) {
        fprintf(psfile->outfile, "currentpoint gsave translate %d rotate 0 0 moveto\n", psfile->angle);
    }

    if (spIsJapaneseLang(psfile->lang) == SP_TRUE) {
        japaneseDrawString(psfile, string);
    } else {
        normalDrawString(psfile, string);
    }

    if (psfile->angle != 0) {
        fprintf(psfile->outfile, "grestore\n");
    }

    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;
    
    return SP_TRUE;
}

static spBool PSDrawRectangle(spPSFile *psfile, int x, int y, int width, int height, spBool fill)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    PSSetupLine(psfile);
    
    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    PSGetSizeInUnit(psfile, &width, &height);

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }
    fprintf(psfile->outfile, "%d %d moveto\n", x, y);
    fprintf(psfile->outfile, "%d 0 rlineto\n", width);
    if (psfile->screen_coordinate == SP_TRUE) {
        fprintf(psfile->outfile, "0 %d rlineto\n", -height);
    } else {
        fprintf(psfile->outfile, "0 %d rlineto\n", height);
    }
    fprintf(psfile->outfile, "%d 0 rlineto\n", -width);
    fprintf(psfile->outfile, "closepath\n");
    if (fill == SP_TRUE) {
        fprintf(psfile->outfile, "fill\n");
    } else {
        fprintf(psfile->outfile, "stroke\n");
    }

    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;
    
    return SP_TRUE;
}

spBool spPSFillRectangle(spPSFile *psfile, int x, int y, int width, int height)
{
    return PSDrawRectangle(psfile, x, y, width, height, SP_TRUE);
}

spBool spPSDrawRectangle(spPSFile *psfile, int x, int y, int width, int height)
{
    return PSDrawRectangle(psfile, x, y, width, height, SP_FALSE);
}

static double PSGetCircleAngle(double angle1, int width, int height)
{
    double tanangle1;
    double angle1d;
    
    if (spRem(angle1, 90.0) == 0.0) {
        angle1d = angle1;
    } else {
        tanangle1 = tan(PI * angle1 / 180.0);
        angle1d = 180.0 * atan2((double)width * tanangle1, (double)height) / PI;
        if (angle1d < 0.0) {
            if (angle1 > 90 && angle1 < 180) {
                angle1d += 180.0;
            } else {
                angle1d += 360.0;
            }
        } else if (angle1 > 180 && angle1 < 270) {
            angle1d += 180.0;
        }
    }

    return angle1d;
}

static spBool PSDrawArc(spPSFile *psfile, int x, int y, int width, int height, double angle1, double angle2, spBool fill)
{
    double cx, cy;
    double angle1d, angle2d;

    PSSetupLine(psfile);
    
    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    PSGetSizeInUnit(psfile, &width, &height);

    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }

    angle1d = PSGetCircleAngle(angle1, width, height);
    angle2d = PSGetCircleAngle(angle1 + angle2, width, height);

    cx = (double)x + (double)width / 2.0;
    if (psfile->screen_coordinate == SP_TRUE) {
        cy = (double)y - (double)height / 2.0;
    } else {
        cy = (double)y + (double)height / 2.0;
    }
    fprintf(psfile->outfile, "%.0f %.0f moveto\n", spRound(cx), spRound(cy));
    fprintf(psfile->outfile, "currentpoint gsave translate %.3f 1.0 scale\n", (double)width / (double)height);
    if (fill == SP_FALSE) {
        fprintf(psfile->outfile, "newpath\n");
    }
    fprintf(psfile->outfile, "0.0 0.0 %.1f %.1f %.f arc\n", (double)height / 2.0, angle1d, angle2d);
    if (fill == SP_TRUE) {
        fprintf(psfile->outfile, "closepath fill\n");
    } else {
        fprintf(psfile->outfile, "stroke\n");
    }
    fprintf(psfile->outfile, "grestore\n");
    
    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;

    return SP_TRUE;
}

spBool spPSFillArc(spPSFile *psfile, int x, int y, int width, int height, double angle1, double angle2)
{
    return PSDrawArc(psfile, x, y, width, height, angle1, angle2, SP_TRUE);
}

spBool spPSDrawArc(spPSFile *psfile, int x, int y, int width, int height, double angle1, double angle2)
{
    return PSDrawArc(psfile, x, y, width, height, angle1, angle2, SP_FALSE);
}

spBool spPSStartPolygonPoint(spPSFile *psfile, int x, int y)
{
    if (psfile == NULL || psfile->outfile == NULL) return SP_FALSE;

    PSSetupLine(psfile);
    
    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    
    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }
    fprintf(psfile->outfile, "/sppspolygon%ld {\n", ++psfile->polygon_id);
    fprintf(psfile->outfile, "    %d %d moveto\n", x, y);

    psfile->polygon_state = SP_TRUE;
    
    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;
    
    return SP_TRUE;
}

spBool spPSAddPolygonPoint(spPSFile *psfile, int x, int y)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->polygon_state == SP_FALSE) return SP_FALSE;

    PSGetPointInUnit(psfile, &x, &y, SP_FALSE);
    
    fprintf(psfile->outfile, "    %d %d lineto\n", x, y);
    
    return SP_TRUE;
}

spBool spPSAddPolygonRelativePoint(spPSFile *psfile, int dx, int dy)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->polygon_state == SP_FALSE) return SP_FALSE;

    PSGetPointInUnit(psfile, &dx, &dy, SP_TRUE);
    
    fprintf(psfile->outfile, "    %d %d rlineto\n", dx, dy);
    
    return SP_TRUE;
}

spPSPolygonId spPSEndPolygonPoint(spPSFile *psfile)
{
    if (psfile == NULL || psfile->outfile == NULL || psfile->polygon_state == SP_FALSE) return SP_PS_POLYGON_NONE;

    fprintf(psfile->outfile, "    closepath } def\n");
    
    psfile->polygon_state = SP_FALSE;
    
    return psfile->polygon_id;
}

static spBool PSDrawPolygon(spPSFile *psfile, spPSPolygonId polygon_id, int dx, int dy, spBool fill)
{
    if (psfile == NULL || psfile->outfile == NULL
        || polygon_id == SP_PS_POLYGON_NONE) return SP_FALSE;

    spDebug(50, "PSDrawPolygon", "original dx = %d, dy = %d\n", dx, dy);
    
    PSGetPointInUnit(psfile, &dx, &dy, SP_TRUE);
    
    if (psfile->relative_state == SP_TRUE) {
        fprintf(psfile->outfile, "stroke\n");
    }
    
    spDebug(50, "PSDrawPolygon", "dx = %d, dy = %d, polygon_id = %ld\n", dx, dy, polygon_id);
    fprintf(psfile->outfile, "gsave %d %d translate sppspolygon%ld\n", dx, dy, polygon_id);
    if (fill == SP_TRUE) {
        fprintf(psfile->outfile, "eofill\n");
    } else {
        fprintf(psfile->outfile, "stroke\n");
    }
    fprintf(psfile->outfile, "grestore\n");
    
    psfile->path_count = 0;
    psfile->relative_state = SP_FALSE;

    return SP_TRUE;
}

spBool spPSFillPolygon(spPSFile *psfile, spPSPolygonId polygon_id, int dx, int dy)
{
    return PSDrawPolygon(psfile, polygon_id, dx, dy, SP_TRUE);
}

spBool spPSDrawPolygon(spPSFile *psfile, spPSPolygonId polygon_id, int dx, int dy)
{
    return PSDrawPolygon(psfile, polygon_id, dx, dy, SP_FALSE);
}
