/*
 *        spPlugin.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spLibrary.h>
#include <sp/spDirectory.h>
#include <sp/spOption.h>
#include <sp/spThread.h>

#include <sp/spPluginP.h>

#if defined(IPHONE)
#include <dlfcn.h>
#endif

static char sp_plugin_lang[SP_MAX_LINE] = "";

static char sp_plugin_default_search_path[] = SP_PLUGIN_DEFAULT_SEARCH_PATH;
static char *sp_plugin_search_path = NULL;

typedef struct _spPluginRecList *spPluginRecList;
struct _spPluginRecList
{
    char *name;
    spPluginRec *plugin_rec;
    void *plugin_handle;

    spPluginRecList prev;
    spPluginRecList next;
};

struct _spPluginHostData
{
    long version_id;

    int num_plugin_file;
    char **plugin_files;
    
    char lang[SP_MAX_LINE];
    spBool exit_callback_added;
    spPluginRecList rec_list;
};

static spPluginHostData sp_plugin_host_data = NULL;

static void spFreeAllPluginRecList(spPluginRecList rec_list)
{
    spPluginRecList list;
    spPluginRecList prev_list;
    
    spDebug(50, "spFreeAllPluginRecList", "in\n");
    
    list = rec_list;
    if (list != NULL) {
        while (list->next != NULL) {
            list = list->next;
        }
    }
    while (list != NULL) {
        prev_list = list->prev;
        spRemovePluginRecList(list->plugin_rec);
        list = prev_list;
    }

    spDebug(50, "spFreeAllPluginRecList", "done\n");
    
    return;
}

static void spFreePluginHostDataFileList(spPluginHostData host_data)
{
    int i;

    if (host_data->plugin_files != NULL) {
        for (i = 0; i < host_data->num_plugin_file; i++) {
            if (host_data->plugin_files[i] != NULL) {
                xfree(host_data->plugin_files[i]);
            }
        }
        xfree(host_data->plugin_files);
    }
    host_data->num_plugin_file = 0;
    host_data->plugin_files = NULL;

    return;
}

static spPluginHostData spAllocPluginHostData(void)
{
    spPluginHostData host_data;
    
    host_data = xalloc(1, struct _spPluginHostData);
    host_data->version_id = SP_PLUGIN_VERSION_ID;
        
    host_data->num_plugin_file = 0;
    host_data->plugin_files = NULL;
        
    spStrCopy(host_data->lang, SP_MAX_LINE, sp_plugin_lang);
    host_data->exit_callback_added = SP_FALSE;
    host_data->rec_list = NULL;

    return host_data;
}

spPluginHostData spGetPluginHostData(void)
{
    if (sp_plugin_host_data == NULL) {
        sp_plugin_host_data = spAllocPluginHostData();
    }

    return sp_plugin_host_data;
}

spPluginHostData spGetPluginHostDataFromHost(spPluginHost *host)
{
    if (host == NULL) return NULL;
    
    return (spPluginHostData)host->data;
}

spBool spFreePluginHostData(spPluginHostData host_data, spBool keep_in_memory)
{
    if (host_data == NULL) return SP_FALSE;

    if (keep_in_memory == SP_FALSE) {
        if (host_data == sp_plugin_host_data) {
            sp_plugin_host_data = NULL;
        }
        if (host_data->rec_list != NULL) {
            spFreeAllPluginRecList(host_data->rec_list);
        }
            
        /* free the list of plugin files */
        spFreePluginHostDataFileList(host_data);
            
        xfree(host_data);
    }
    
    return SP_TRUE;
}

spBool spIsPluginInMemory(spPlugin *plugin)
{
    spPluginRecList list;
    
    if (plugin != NULL && plugin->rec != NULL && sp_plugin_host_data != NULL) {
        list = sp_plugin_host_data->rec_list;
        while (list != NULL) {
            spDebug(100, "spIsPluginInMemory", "address = %ld, %ld\n", plugin->rec, list->plugin_rec);
            if (list->plugin_rec == plugin->rec) {
                spDebug(80, "spIsPluginInMemory", "found in memory: %ld\n", plugin->rec);
                return SP_TRUE;
            }
            list = list->next;
        }
    }
    
    return SP_FALSE;
}

static void pluginExitCB(void *data)
{
    spFreePluginGlobals();
    
    spDebug(80, "pluginExitCB", "done\n");
    
    return;
}

static spBool freePluginRecHost(spPluginRec *rec, spBool call_free)
{
    spPluginHost *host;
    
    if (rec == NULL) return SP_FALSE;

    spDebug(50, "freePluginRecHost", "call_free = %d\n", call_free);
    
    if (call_free == SP_TRUE) {
        if (rec->free != NULL) {
            spDebug(50, "freePluginRecHost", "call free\n");
            (*(rec->free))();
            spDebug(50, "freePluginRecHost", "after free\n");
        }
    }

    host = (spPluginHost *)rec->host_data;

    if (host != NULL) {
        if (host->version_id >= 1007 && host->plugin_specific_data != NULL) {
            xfree(host->plugin_specific_data);
            host->plugin_specific_data = NULL;
        }
        
        if (host->buffer != NULL) {
            xfree(host->buffer);
        }
        host->buffer = NULL;

        if (host->mutex != NULL) {
            spDestroyMutex(host->mutex);
        }
        host->mutex = NULL;
            
        if (host->name != NULL) {
            xfree(host->name);
        }
        host->name = NULL;
        host->data = NULL;

        xfree(host);
    }

    rec->host_data = NULL;
    
    spDebug(50, "freePluginRecHost", "done\n");
    
    return SP_TRUE;
}

static spBool addPluginRecList(spPluginHostData host_data, spPluginRec *plugin_rec, const char *name, void *plugin_handle)
{
    spPluginRecList next;
    spPluginRecList list;
    
    if (host_data == NULL || plugin_rec == NULL || name == NULL) return SP_FALSE;
    
    next = host_data->rec_list;
    while (next != NULL) {
        spDebug(50, "addPluginRecList", "name = %s (%ld), next->name = %s (%ld)\n",
                name, plugin_rec, next->name, next->plugin_rec);
        if (next->plugin_rec == plugin_rec || spEqPluginNameString(next->name, name)) {
            /* already added */
            spDebug(50, "addPluginRecList", "plugin already added: %s\n", name);
            return SP_FALSE;
        }
        next = next->next;
    }
    
    spDebug(50, "addPluginRecList", "adding plugin rec: name = %s\n", name);
    
    list = xalloc(1, struct _spPluginRecList);
    list->name = strclone(name);
    list->plugin_rec = plugin_rec;
    list->plugin_handle = plugin_handle;
    
    list->prev = NULL;
    list->next = NULL;
    
    if (host_data->rec_list == NULL) {
        host_data->rec_list = list;
    } else {
        next = host_data->rec_list;
        while (next != NULL) {
            if (next->next == NULL) {
                next->next = list;
                list->prev = next;
                break;
            }
            next = next->next;
        }
    }
    
    if (host_data->exit_callback_added == SP_FALSE) {
        /* set callback function to free memory */
        spAddExitCallback(pluginExitCB, NULL);
        host_data->exit_callback_added = SP_TRUE;
        spDebug(80, "addPluginRecList", "pluginExitCB added\n");
    }
    
    return SP_TRUE;
}

spBool spAddPluginRecList(spPluginRec *plugin_rec, const char *name)
{
    return addPluginRecList(spGetPluginHostData(), plugin_rec, name, NULL);
}

static spBool removePluginRecList(spPluginHostData host_data, spPluginRec *plugin_rec)
{
    void *handle;
    spPluginRecList list;
    
    if (plugin_rec == NULL || host_data == NULL || host_data->rec_list == NULL) return SP_FALSE;

    list = host_data->rec_list;
        
    while (list != NULL) {
        if (list->plugin_rec == plugin_rec) {
            spDebug(50, "spRemovePluginRecList", "removing plugin rec: name = %s\n", list->name);
            
            handle = list->plugin_handle;

            spFreePluginRecHost(list->plugin_rec);
            
            if (list->prev != NULL) {
                list->prev->next = list->next;
            }
            if (list->next != NULL) {
                list->next->prev = list->prev;
            }
            if (host_data->rec_list->plugin_rec == plugin_rec) {
                if (host_data->rec_list->next == NULL) {
                    host_data->rec_list = NULL;
                } else {
                    host_data->rec_list = host_data->rec_list->next;
                }
            }
            xfree(list->name);
            xfree(list);

            if (handle != NULL) {
                spCloseLibrary(handle);
            }
            break;
        }

        list = list->next;
    }

    return SP_TRUE;
}

spBool spRemovePluginRecList(spPluginRec *plugin_rec)
{
    return removePluginRecList(sp_plugin_host_data, plugin_rec);
}

spBool _spRegisterPluginRec(spPluginRec *plugin_rec, const char *name, long system_version_id, long supported_version_id)
{
    if (system_version_id < SP_PLUGIN_SUPPORTED_VERSION_ID) {
        spDebug(10, "_spRegisterPluginRec", "old plugin file\n");
        return SP_FALSE;
    }
    if (supported_version_id > SP_PLUGIN_VERSION_ID) {
        spDebug(10, "checkPluginVersion", "unsupported plugin file\n");
        return SP_FALSE;
    }

    return spAddPluginRecList(plugin_rec, name);
}

static spPluginRec *searchPluginRecList(spPluginHostData host_data, const char *name, spBool static_only)
{
    spPluginRecList list;
    
    if (host_data == NULL || host_data->rec_list == NULL || name == NULL) {
        spDebug(10, "searchPluginRecList", "input arg is invalid\n");
        return NULL;
    }

    list = host_data->rec_list;

    while (list != NULL) {
        spDebug(100, "searchPluginRecList", "name = %s, list->name = %s\n", name, list->name);
        
        if (spEqPluginNameString(list->name, name)
            && (static_only == SP_FALSE || list->plugin_handle == NULL)) {
            spDebug(80, "searchPluginRecList", "found plugin: %s\n", list->name);
            return list->plugin_rec;
        }
        list = list->next;
    }
    
    spDebug(80, "searchPluginRecList", "not found\n");
    
    return NULL;
}

spPluginRec *spSearchPluginRecList(const char *name, spBool static_only)
{
    return searchPluginRecList(sp_plugin_host_data, name, static_only);
}

static const char *getPluginRecListName(spPluginHostData host_data, int index, spBool static_only)
{
    int i;
    spPluginRecList list;
    
    if (host_data->rec_list == NULL || index < 0) return NULL;

    list = host_data->rec_list;

    for (i = 0; list != NULL; i++) {
        if (i == index
            && (static_only == SP_FALSE || list->plugin_handle == NULL)) {
            return list->name;
        }
        list = list->next;
    }
    
    return NULL;
}

const char *spGetPluginRecListName(int index, spBool static_only)
{
    return getPluginRecListName(sp_plugin_host_data, index, static_only);
}

void spSetPluginLang(const char *lang)
{
    if (!strnone(lang)) {
        spStrCopy(sp_plugin_lang, sizeof(sp_plugin_lang), lang);
        if (sp_plugin_host_data != NULL) {
            spStrCopy(sp_plugin_host_data->lang, sizeof(sp_plugin_host_data->lang), lang);
        }
    }
    
    return;
}

const char *spGetPluginLang(void)
{
    if (sp_plugin_host_data != NULL) {
        return sp_plugin_host_data->lang;
    } else {
        return sp_plugin_lang;
    }
}

const char *spGetPluginSearchPath(void)
{
    char *path = NULL;
    
    if (sp_plugin_search_path == NULL) {
#if defined(HAVE_DUPENV_S)
        size_t pathlen = 0;
        spBool alloc_flag = SP_FALSE;
        
        path = NULL;
        if (_dupenv_s(&path, &pathlen, SP_PLUGIN_PATH_ENV) != 0
            || pathlen <= 0) {
            if (path != NULL) free(path);
            path = NULL;
        } else {
            alloc_flag = SP_TRUE;
        }
#else
        path = getenv(SP_PLUGIN_PATH_ENV);
#endif
        if (path == NULL) {
            path = sp_plugin_default_search_path;
        }
        sp_plugin_search_path = strclone(path);
        
#if defined(HAVE_DUPENV_S)
        if (alloc_flag == SP_TRUE) {
            free(path);
        }
#endif
    }

    return sp_plugin_search_path;
}

void spSetPluginSearchPath(const char *pathlist)
{
    if (sp_plugin_search_path != NULL) {
        xfree(sp_plugin_search_path);
    }
    
    if (pathlist == NULL) {
        sp_plugin_search_path = NULL;
    } else {
        sp_plugin_search_path = strclone(pathlist);
    }
    
    return;
}

spBool spAppendPluginSearchPath(const char *pathlist)
{
    if (!strnone(pathlist)) {
        const char *list;
        char *olist;
        
        spDebug(80, "spAppendPluginSearchPath", "pathlist = %s\n", pathlist);

        list = spGetPluginSearchPath();
        olist = xspAppendPathList(list, pathlist);
        if (olist != NULL) {
            spDebug(80, "spAppendPluginSearchPath", "olist = %s\n", olist);
            if (sp_plugin_search_path != NULL) xfree(sp_plugin_search_path);
            sp_plugin_search_path = olist;
            return SP_TRUE;
        }
    }
    
    spDebug(80, "spAppendPluginSearchPath", "list not updated\n");
    
    return SP_FALSE;
}

spBool spPrependPluginSearchPath(const char *pathlist)
{
    if (!strnone(pathlist)) {
        const char *list;
        char *olist;
        
        spDebug(80, "spPrependPluginSearchPath", "pathlist = %s\n", pathlist);

        list = spGetPluginSearchPath();
        olist = xspPrependPathList(list, pathlist);
        if (olist != NULL) {
            spDebug(80, "spPrependPluginSearchPath", "olist = %s\n", olist);
            if (sp_plugin_search_path != NULL) xfree(sp_plugin_search_path);
            sp_plugin_search_path = olist;
            return SP_TRUE;
        }
    }
    
    spDebug(80, "spPrependPluginSearchPath", "list not updated\n");
    
    return SP_FALSE;
}

static char *xspCutPluginPathList(int index)
{
#if defined(ANDROID)
    if (index <= 0) {
        return strclone(spGetApplicationLibDir());
    } else {
        return NULL;
    }
#else
    int offset = 0;

#if defined(_WIN32) || defined(MACOS)
    offset = 2;
    if (index <= 0) {
        return strclone(spGetApplicationLibDir());
    } else if (index <= 1) {
        char buf[SP_MAX_PATHNAME];
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, sizeof(buf), "%s%cplugins", spGetApplicationLibDir(), SP_DIR_SEPARATOR);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, sizeof(buf), "%s%cplugins", spGetApplicationLibDir(), SP_DIR_SEPARATOR);
#else
        sprintf(buf, "%s%cplugins", spGetApplicationLibDir(), SP_DIR_SEPARATOR);
#endif
        return strclone(buf);
    }
#endif
    
    return xspCutPathList(spGetPluginSearchPath(), index - offset);
#endif
}

static spBool isPluginCandidate(const char *filename)
{
    if (strnone(filename)) return SP_FALSE;

    spDebug(100, "isPluginCandidate", "filename: %s\n", filename);
    
#if defined(MACOSX)
    if (spExists(filename) == SP_TRUE) {
        spDebug(50, "isPluginCandidate", "this may be plugin: %s\n", filename);
        return SP_TRUE;
    }
#else
    if (spIsFile(filename) == SP_TRUE) {
        return SP_TRUE;
    }
#endif

    spDebug(80, "isPluginCandidate", "not plugin candidate: %s\n", filename);
    
    return SP_FALSE;
}

#if !defined(IPHONE)
static char *xgetExactPluginName(const char *plugin_name)
{
    int i, j;
    int suffix_count = 1;
    char *suffix;
    char *path;
    char *exactname;
    char filename[SP_MAX_PATHNAME];
#if defined(ANDROID)
    char buf[SP_MAX_PATHNAME];
#endif

    spDebug(100, "xgetExactPluginName", "plugin_name = %s\n", plugin_name);
    
#if defined(SP_PLUGIN_ANOTHER_SUFFIX)
    suffix_count = 2;
#endif
            
    if (spIsExactName(plugin_name) == SP_TRUE) {
        if (isPluginCandidate(plugin_name) == SP_TRUE) {
            spDebug(100, "xgetExactPluginName", "%s is plugin candidate\n", plugin_name);
            exactname = strclone(plugin_name);
#if defined(MACOSX)
            spRemoveDirSeparator(exactname);
#endif
            return exactname;
        } else {
            for (j = 0; j < suffix_count; j++) {
                if (j == 1) {
#if defined(SP_PLUGIN_ANOTHER_SUFFIX)
                    suffix = SP_PLUGIN_ANOTHER_SUFFIX;
#else
                    suffix = SP_PLUGIN_SUFFIX;
#endif
                } else {
                    suffix = SP_PLUGIN_SUFFIX;
                }
#if defined(HAVE_SPRINTF_S)
                sprintf_s(filename, sizeof(filename), "%s%s", plugin_name, suffix);
#elif defined(HAVE_SNPRINTF)
                snprintf(filename, sizeof(filename), "%s%s", plugin_name, suffix);
#else
                sprintf(filename, "%s%s", plugin_name, suffix);
#endif
                spDebug(80, "xgetExactPluginName", "added suffix: %s\n", filename);
                if (isPluginCandidate(filename) == SP_TRUE) {
                    spDebug(80, "xgetExactPluginName", "exact plugin name = %s\n", filename);
                    exactname = strclone(filename);
#if defined(MACOSX)
                    spRemoveDirSeparator(exactname);
#endif
                    return exactname;
                }
            }
        }
        plugin_name = spGetBaseName(plugin_name);
#if defined(ANDROID)
    } else {
#if defined(HAVE_SPRINTF_S)
        sprintf_s(buf, sizeof(buf), "lib%s", plugin_name);
#elif defined(HAVE_SNPRINTF)
        snprintf(buf, sizeof(buf), "lib%s", plugin_name);
#else
        sprintf(buf, "lib%s", plugin_name);
#endif
        plugin_name = buf;
#endif
    }

    if (strnone(plugin_name)) {
        return NULL;
    }

    spStrCopy(filename, sizeof(filename), "");
    for (i = 0;; i++) {
        if ((path = xspCutPluginPathList(i)) == NULL) {
#if defined(HAVE_SPRINTF_S)
            sprintf_s(filename, sizeof(filename), "%s%s", plugin_name, SP_PLUGIN_SUFFIX);
#elif defined(HAVE_SNPRINTF)
            snprintf(filename, sizeof(filename), "%s%s", plugin_name, SP_PLUGIN_SUFFIX);
#else
            sprintf(filename, "%s%s", plugin_name, SP_PLUGIN_SUFFIX);
#endif
            exactname = strclone(filename);
            break;
        }
        spDebug(100, "xgetExactPluginName", "i = %d, path = %s\n", i, path);

        exactname = NULL;
        for (j = 0; j < suffix_count; j++) {
            if (j == 1) {
#if defined(SP_PLUGIN_ANOTHER_SUFFIX)
                suffix = SP_PLUGIN_ANOTHER_SUFFIX;
#else
                suffix = SP_PLUGIN_SUFFIX;
#endif
            } else {
                suffix = SP_PLUGIN_SUFFIX;
            }
            
#if defined(HAVE_SPRINTF_S)
            sprintf_s(filename, sizeof(filename), "%s%c%s%s", path, SP_DIR_SEPARATOR,
                      plugin_name, suffix);
#elif defined(HAVE_SNPRINTF)
            snprintf(filename, sizeof(filename), "%s%c%s%s", path, SP_DIR_SEPARATOR,
                     plugin_name, suffix);
#else
            sprintf(filename, "%s%c%s%s", path, SP_DIR_SEPARATOR,
                    plugin_name, suffix);
#endif
            exactname = xspGetExactName(filename);
        
            spDebug(100, "xgetExactPluginName",
                    "filename = %s, exactname = %s\n", filename, exactname);
        
            if (isPluginCandidate(exactname) == SP_TRUE) {
                break;
            } else {
#if defined(HAVE_SPRINTF_S)
                sprintf_s(filename, sizeof(filename), "%s%c%s", path, SP_DIR_SEPARATOR, plugin_name);
#elif defined(HAVE_SNPRINTF)
                snprintf(filename, sizeof(filename), "%s%c%s", path, SP_DIR_SEPARATOR, plugin_name);
#else
                sprintf(filename, "%s%c%s", path, SP_DIR_SEPARATOR, plugin_name);
#endif
                xfree(exactname);
                exactname = xspGetExactName(filename);
            
                if (isPluginCandidate(exactname) == SP_TRUE) {
                    break;
                }
            }
            xfree(exactname);
            exactname = NULL;
        }

        xfree(path);
        
        if (exactname != NULL) {
#if defined(MACOSX)
            spRemoveDirSeparator(exactname);
#endif
            break;
        }
        
    }

    spDebug(100, "xgetExactPluginName", "exactname = %s\n", exactname);
    
    return exactname;
}

static spBool checkPluginVersion(void *handle)
{
    spPluginVersionFunc version_func;
    spBool supported = SP_TRUE;
    long system_version_id = -1;
    long supported_version_id = -1;
    
    if ((version_func = (spPluginVersionFunc)spGetSymbolAddress(handle,
                                                                SP_PLUGIN_SYSTEM_VERSION_FUNC)) == NULL
        || (system_version_id = version_func()) < SP_PLUGIN_SUPPORTED_VERSION_ID) {
        spDebug(10, "checkPluginVersion", "old plugin file\n");
        supported = SP_FALSE;
    }
    spDebug(80, "checkPluginVersion", "system version = %ld\n", system_version_id);
    
    if ((version_func = (spPluginVersionFunc)spGetSymbolAddress(handle,
                                                                SP_PLUGIN_SUPPORTED_VERSION_FUNC)) == NULL
        || (supported_version_id = version_func()) > SP_PLUGIN_VERSION_ID) {
        spDebug(10, "checkPluginVersion", "unsupported plugin file\n");
        supported = SP_FALSE;
    }
    spDebug(80, "checkPluginVersion", "supported version = %ld\n", supported_version_id);

    return supported;
}
#endif

static spBool freePlugin(spPlugin *plugin, spBool call_free)
{
    void *handle;
    spBool keep_in_memory = SP_FALSE;

    if (plugin == NULL) return SP_FALSE;

    spDebug(80, "freePlugin", "in\n");

    handle = plugin->handle;

    if (plugin->instance != NULL) {
        spFreePluginInstance(plugin);
        plugin->instance = NULL;
    }
    spDebug(80, "freePlugin", "spFreePluginInstance done\n");
    
    if (spIsPluginInMemory(plugin) == SP_TRUE) {
        keep_in_memory = SP_TRUE;
        handle = NULL;
    }
    spDebug(80, "freePlugin", "spIsPluginInMemory done\n");
    
    if (plugin->host != NULL) {
        spDebug(40, "freePlugin", "%s: ref_count = %ld, keep_in_memory = %d\n",
                plugin->host->name, plugin->host->ref_count, keep_in_memory);
        
        if (plugin->host->ref_count <= 1) {
            if (keep_in_memory == SP_TRUE) {
            } else {
                freePluginRecHost(plugin->rec, call_free);
                spDebug(40, "freePlugin", "after freePluginRecHost\n");
                plugin->handle = NULL;
                plugin->host = NULL;
            }
        } else {
            plugin->host->ref_count--;
            spDebug(40, "freePlugin", "updated ref_count = %ld\n", plugin->host->ref_count);
        }
    }
    if (handle != NULL) {
        spCloseLibrary(handle);
    }
    
    xfree(plugin);

    spDebug(80, "freePlugin", "done\n");
    
    return SP_TRUE;
}

spPlugin *spAllocNullPlugin(void)
{
    spPlugin *plugin;
    
    plugin = xalloc(1, struct _spPlugin);
    memset(plugin, 0, sizeof(struct _spPlugin));

    return plugin;
}

static spPlugin *allocPlugin(spPluginRec *rec, const char *name, void *handle)
{
    spPlugin *plugin;
    spPluginHost *host;
    
    plugin = xalloc(1, struct _spPlugin);
    plugin->host = NULL;
    plugin->rec = rec;
    plugin->handle = handle;
    plugin->instance = NULL;

    spDebug(30, "allocPlugin", "rec->host_data = %ld\n", (long)rec->host_data);
    
    if (rec->host_data == NULL) {
        spPluginHostData host_data;
        
        host = xalloc(1, struct _spPluginHost);
        memset(host, 0, sizeof(struct _spPluginHost));
        host->name = strclone(name);
        host->version_id = SP_PLUGIN_VERSION_ID;
        host->ref_count = 1;
        host_data = spGetPluginHostData();
        host->data = host_data;
        host->mutex = spCreateMutex(NULL);
        host->num_instance = 0;

        host->buffer_size = 0;
        host->buffer = NULL;

        rec->host_data = (void *)host;
        
        if (rec->init != NULL && rec->init(host_data->lang) == SP_FALSE) {
            plugin->host = host;
            spDebug(10, "allocPlugin", "init failed\n");
            freePlugin(plugin, SP_FALSE);
            return NULL;
        }
        spDebug(80, "allocPlugin", "init done\n");

        /* keep plugin into memory */
        if (/*plugin->rec->caps & SP_PLUGIN_CAPS_KEEP_IN_MEMORY*/1) {
            if (addPluginRecList(host_data, plugin->rec, host->name, plugin->handle) == SP_TRUE) {
            }
        }
    } else {
        host = (spPluginHost *)rec->host_data;
        host->ref_count++;
    }

    spDebug(40, "allocPlugin", "%s: ref_count = %ld\n", host->name, host->ref_count);
    
    plugin->host = host;
    
    return plugin;
}

spPluginRec *spCallUniqueGetPluginRecFunc(spUniqueGetPluginRecFunc func)
{
    long system_version_id = -1;
    long supported_version_id = -1;
    spPluginRec *rec;
    
    rec = func(&system_version_id, &supported_version_id);
    spDebug(80, "spCallUniqueGetPluginRecFunc", "system_version_id = %ld, supported_version_id = %ld\n",
            system_version_id, supported_version_id);
        
    if (system_version_id < SP_PLUGIN_SUPPORTED_VERSION_ID
        || supported_version_id > SP_PLUGIN_VERSION_ID) {
        spDebug(10, "spCallUniqueGetPluginRecFunc", "not supported version: %ld, %ld\n", system_version_id, supported_version_id);
        return NULL;
    }

    return rec;
}

spPlugin *spLoadPlugin(const char *plugin_name)
{
    char *filename;
    void *handle;
    spPlugin *plugin;
    spPluginRec *rec;
    spGetPluginRecFunc plugin_rec_func = NULL;
#if defined(_WIN32) && !defined(__CYGWIN32__)
    static char current_dir[SP_MAX_PATHNAME] = "";
    static char plugin_dir[SP_MAX_PATHNAME] = "";
#endif

    if (plugin_name == NULL) return NULL;

    filename = NULL;
    handle = NULL;
    
    if ((rec = spSearchPluginRecList(plugin_name, SP_FALSE)) == NULL) {
#if defined(IPHONE)
        char func_name[SP_MAX_LINE];
        spUniqueGetPluginRecFunc unique_plugin_rec_func = NULL;
        
#if defined(HAVE_SPRINTF_S)
        sprintf_s(func_name, sizeof(func_name), "%s_%s", SP_GET_PLUGIN_REC_FUNC, plugin_name);
#elif defined(HAVE_SNPRINTF)
        snprintf(func_name, sizeof(func_name), "%s_%s", SP_GET_PLUGIN_REC_FUNC, plugin_name);
#else
        sprintf(func_name, "%s_%s", SP_GET_PLUGIN_REC_FUNC, plugin_name);
#endif
        spDebug(80, "spLoadPlugin", "plugin_name = %s, func_name = %s\n", plugin_name, func_name);
        
        if ((unique_plugin_rec_func = dlsym(/*RTLD_SELF*/RTLD_MAIN_ONLY, func_name)) == NULL) {
            spDebug(10, "spLoadPlugin", "dlsym failed: %s\n", func_name);
            return NULL;
        }

        if ((rec = spCallUniqueGetPluginRecFunc(unique_plugin_rec_func)) == NULL) {
            return NULL;
        }
#else
        if ((filename = xgetExactPluginName(plugin_name)) == NULL) {
            return NULL;
        }
        spDebug(80, "spLoadPlugin", "filename = %s\n", filename);
        
        if ((rec = spSearchPluginRecList(filename, SP_FALSE)) == NULL) {
#if defined(_WIN32) && !defined(__CYGWIN32__)
            GetCurrentDirectory(sizeof(current_dir), current_dir);
            spDebug(80, "spLoadPlugin", "current_dir = %s\n", current_dir);
#if 0
            if (GetModuleFileName(NULL, plugin_dir, sizeof(plugin_dir))) {
                char drive[SP_MAX_PATHNAME];
                char dir[SP_MAX_PATHNAME];
            
#if defined(HAVE_SPLITPATH_S)
                _splitpath_s(plugin_dir, drive, sizeof(drive), dir, sizeof(dir), NULL, 0, NULL, 0);
#else
                _splitpath(plugin_dir, drive, dir, NULL, NULL);
#endif
                spStrCopy(plugin_dir, sizeof(plugin_dir), drive);
                spStrCat(plugin_dir, sizeof(plugin_dir), dir);
                spStrCat(plugin_dir, sizeof(plugin_dir), "plugins");
                SetCurrentDirectory(plugin_dir);
                spDebug(80, "spLoadPlugin", "add plugin path: %s\n", plugin_dir);
            }
#else
            {
                char *p;

                if ((p = xspGetDirName(filename)) != NULL) {
                    SetCurrentDirectory(p);
                    spDebug(80, "spLoadPlugin", "add plugin path: %s\n", p);
                    xfree(p);
                }
            }
#endif
#endif
        
            spDebug(100, "spLoadPlugin", "call spOpenLibrary for %s\n", filename);
            
            if ((handle = spOpenLibrary(filename)) != NULL) {
                spDebug(100, "spLoadPlugin", "get handle done\n");
            
                if ((plugin_rec_func = (spGetPluginRecFunc)spGetSymbolAddress(handle,
                                                                              SP_GET_PLUGIN_REC_FUNC)) != NULL) {
                    spDebug(100, "spLoadPlugin", "get symbol done\n");
                
#ifdef SP_PLUGIN_VERIFY_VERSION
                    if (checkPluginVersion(handle) == SP_FALSE) {
                        plugin_rec_func = NULL;
                    }
#endif
                }
            }
            
#if defined(_WIN32) && !defined(__CYGWIN32__)
            SetCurrentDirectory(current_dir);
#endif
    
            if (plugin_rec_func == NULL) {
                if (handle != NULL) {
                    spCloseLibrary(handle);
                }
                if (filename != NULL) xfree(filename);
                return NULL;
            }
            rec = plugin_rec_func();
        }
#endif /* !defined(IPHONE) */
    }

    if ((plugin = allocPlugin(rec, (filename == NULL ? plugin_name : filename), handle)) == NULL) {
        spDebug(10, "spLoadPlugin", "allocPlugin error\n");
    } else {
        /*spDebug(80, "spLoadPlugin", "handle = %ld\n", (long)plugin->handle);*/
    }
    
    if (filename != NULL) {
        xfree(filename);
    }
    
    return plugin;
}

spPlugin *spLoadPluginFromHostData(spPluginHostData host_data, const char *plugin_name)
{
    spPluginRec *rec;

    if (host_data == NULL) {
        spDebug(10, "spLoadPluginFromHostData", "host_data is null\n");
        return NULL;
    }
    
    if ((rec = searchPluginRecList(host_data, plugin_name, SP_FALSE)) == NULL) {
        spDebug(10, "spLoadPluginFromHostData", "searchPluginRecList failed\n");
        return NULL;
    }

    return allocPlugin(rec, plugin_name, NULL);
}

spPlugin *spLoadPluginFromHostMemory(spPluginHost *host, const char *plugin_name)
{
    spPluginHostData host_data;

    if (host == NULL || host->data == NULL) return NULL;
    
    host_data = (spPluginHostData)host->data;

    return spLoadPluginFromHostData(host_data, plugin_name);
}

spBool spFreePluginRecHost(spPluginRec *rec)
{
    return freePluginRecHost(rec, SP_TRUE);
}

spBool spFreePlugin(spPlugin *plugin)
{
    return freePlugin(plugin, SP_TRUE);
}

long spGetPluginRefCount(spPlugin *plugin)
{
    if (plugin == NULL || plugin->host == NULL) return -1;

    return plugin->host->ref_count;
}

spBool spLockPluginMutex(spPlugin *plugin)
{
    if (plugin == NULL
        || plugin->host->mutex == NULL) return SP_FALSE;

    return spLockMutex(plugin->host->mutex);
}

spBool spUnlockPluginMutex(spPlugin *plugin)
{
    if (plugin == NULL
        || plugin->host->mutex == NULL) return SP_FALSE;

    return spUnlockMutex(plugin->host->mutex);
}

spBool spEqPluginTypeMain(spPluginType x_type, spPluginType y_type)
{
    int n;
    int version = 0, t_version = 0;
    int revision = 0, t_revision = 0;
    char buf[SP_MAX_LINE];
    char t_buf[SP_MAX_LINE];

#if defined(HAVE_SSCANF_S)
    n = sscanf_s(x_type, "%s %d.%d", buf, (unsigned int)sizeof(buf), &version, &revision);
#else
    n = sscanf(x_type, "%s %d.%d", buf, &version, &revision);
#endif
    if (n == 1) {
        if (streq(x_type, y_type)) {
            return SP_TRUE;
        }
    } else if (n == 3) {
        spDebug(100, "spEqPluginType", "buf = %s, version = %d, revision = %d\n",
                buf, version, revision);

#if defined(HAVE_SSCANF_S)
        n = sscanf_s(y_type, "%s %d.%d", t_buf, (unsigned int)sizeof(t_buf), &t_version, &t_revision);
#else
        n = sscanf(y_type, "%s %d.%d", t_buf, &t_version, &t_revision);
#endif

        if (n == 1) {
            if (streq(buf, t_buf)) {
                return SP_TRUE;
            }
        } else if (n == 3) {
            spDebug(100, "spEqPluginType", "t_buf = %s, t_version = %d, t_revision = %d\n",
                    t_buf, t_version, t_revision);

            if (streq(buf, t_buf)
                && version == t_version && revision == t_revision) {
                return SP_TRUE;
            }
        }
    }

    return SP_FALSE;
}

spBool spEqPluginType(spPlugin *plugin, spPluginType type)
{
    if (plugin == NULL || type == NULL) return SP_FALSE;

    return spEqPluginTypeMain(plugin->rec->type, type);
}

const char *spGetPluginName(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return plugin->host->name;
}

const char *spGetPluginId(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return plugin->rec->plugin_id;
}

const char *spGetPluginDescription(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return plugin->rec->description;
}

const char *spGetPluginInformation(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return plugin->rec->information;
}


long spGetPluginVersionId(spPlugin *plugin)
{
    if (plugin == NULL) return -1;

    return plugin->rec->version_id;
}

int spGetPluginPriority(spPlugin *plugin)
{
    if (plugin == NULL) return SP_PLUGIN_PRIORITY_NULL;

    spDebug(100, "spGetPluginPriority", "priority = %d\n", plugin->rec->priority);
    
    return plugin->rec->priority;
}

spBool spGetPluginVersion(spPlugin *plugin, int *version, int *revision)
{
    long v, r;
    
    if (plugin == NULL) return SP_FALSE;

    v = spGetPluginVersionId(plugin) / 1000;
    r = spGetPluginVersionId(plugin) - 1000 * v;
    
    if (version != NULL) {
        *version = (int)v;
    }
    if (revision != NULL) {
        *revision = (int)r;
    }
    
    return SP_TRUE;
}

unsigned long spGetPluginCaps(spPlugin *plugin)
{
    if (plugin == NULL) return 0;

    return plugin->rec->caps;
}

spBool spIsPluginCapable(spPlugin *plugin, unsigned long cap)
{
    if (plugin == NULL) return SP_FALSE;

    if (plugin->rec->caps & cap) {
        return SP_TRUE;
    }
    return SP_FALSE;
}

void *spGetPluginInstance(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return plugin->instance;
}

long spGetNumPluginInstance(spPlugin *plugin)
{
    if (plugin == NULL || plugin->host == NULL) return -1;

    return plugin->host->num_instance;
}

spBool spIsPluginInstantiatable(spPlugin *plugin)
{
    if (plugin == NULL || plugin->host == NULL
        || plugin->rec->init_instance == NULL) return SP_FALSE;
    
    spDebug(80, "spIsPluginInstantiatable", "num_instance = %ld\n", plugin->host->num_instance);
        
    if (spIsPluginCapable(plugin, SP_PLUGIN_CAPS_THREAD_SAFE) == SP_FALSE
        && plugin->host->num_instance >= 1) {
        spDebug(80, "spIsPluginInstantiatable", "not thread safe\n");
        return SP_FALSE;
    }

    spDebug(80, "spIsPluginInstantiatable", "instantiatable\n");
    
    return SP_TRUE;
}

spBool spInitPluginBuffer(spPlugin *plugin, long buffer_size)
{
    if (plugin == NULL || plugin->host == NULL) return SP_FALSE;

    if (plugin->host->buffer == NULL) {
        if (buffer_size <= 0) buffer_size = SP_PLUGIN_DEFAULT_BUFFER_SIZE;

        plugin->host->buffer_size = buffer_size;
        plugin->host->buffer = xalloc(plugin->host->buffer_size, char);
    } else if (plugin->host->buffer_size < buffer_size) {
        plugin->host->buffer_size = buffer_size;
        plugin->host->buffer = xrealloc(plugin->host->buffer, plugin->host->buffer_size, char);
    }
    if (plugin->host->buffer == NULL) {
        plugin->host->buffer_size = 0;
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}

static spPluginHostInstanceList spAllocPluginHostInstanceList(spPluginHost *host, void *instance, spThreadId thread_id)
{
    spPluginHostInstanceList list;
    
    list = xalloc(1, struct _spPluginHostInstanceList);
    list->instance = instance;
    list->thread_id = thread_id;
    list->prev = NULL;
    list->next = NULL;

    if (host->instance_list == NULL) {
        host->instance_list = list;
    } else {
        spPluginHostInstanceList next;

        next = host->instance_list;
        
        while (1) {
            if (next->next == NULL) {
                next->next = list;
                list->prev = next;
                break;
            }

            next = next->next;
        }
    }

    return list;
}

static void spFreePluginHostInstanceList(spPluginHost *host, void *instance)
{
    spPluginHostInstanceList next;

    next = host->instance_list;

    while (next != NULL) {
        if (next->instance == instance) {
            if (host->instance_list == next) {
                host->instance_list = host->instance_list->next;
            }
                
            if (next->prev != NULL) next->prev->next = next->next;
            if (next->next != NULL) next->next->prev = next->prev;
            
            xfree(next);
            break;
        }

        next = next->next;
    }

    return;
}

spBool spInitPluginInstance(spPlugin *plugin)
{
    spThreadId thread_id = 0;
    
    if (spIsPluginInstantiatable(plugin) == SP_FALSE) {
        if (plugin->rec->caps & SP_PLUGIN_CAPS_THREAD_BY_THREAD_SAFE && plugin->host->version_id >= 1006) {
            spPluginHostInstanceList list;
            thread_id = spGetCurrentThreadId();

            list = plugin->host->instance_list;
            while (list != NULL) {
                if (spEqThreadId(list->thread_id, thread_id)) {
                    spDebug(10, "spInitPluginInstance", "an instance for this thread already exists in thread-by-thread basis mode\n");
                    return SP_FALSE;
                }
            }
        } else {
            spDebug(10, "spInitPluginInstance", "not instantiatable\n");
            return SP_FALSE;
        }
    }
    
    if ((plugin->instance = plugin->rec->init_instance(spGetPluginLang())) == NULL) {
        spDebug(10, "spInitPluginInstance", "init_instance failed\n");
        return SP_FALSE;
    }
    plugin->host->num_instance++;

    if (plugin->rec->caps & SP_PLUGIN_CAPS_THREAD_BY_THREAD_SAFE && plugin->host->version_id >= 1006) {
        spAllocPluginHostInstanceList(plugin->host, plugin->instance, thread_id);
    }
    
    spDebug(80, "spInitPluginInstance", "num_instance = %ld\n", plugin->host->num_instance);
    
    return SP_TRUE;
}

spBool spFreePluginInstance(spPlugin *plugin)
{
    spBool flag;
    
    if (plugin == NULL || plugin->host == NULL || plugin->instance == NULL
        || plugin->rec->free_instance == NULL) return SP_FALSE;

    if (plugin->rec->caps & SP_PLUGIN_CAPS_THREAD_BY_THREAD_SAFE && plugin->host->version_id >= 1006) {
        spFreePluginHostInstanceList(plugin->host, plugin->instance);
    }
    
    flag = plugin->rec->free_instance(plugin->instance);
    plugin->instance = NULL;
    plugin->host->num_instance--;
    
    return flag;
}

spPlugin *spInstantiatePlugin(const char *plugin_name)
{
    spPlugin *plugin;
    
    if ((plugin = spLoadPlugin(plugin_name)) == NULL) {
        return NULL;
    }
    if (spInitPluginInstance(plugin) == SP_FALSE) {
        spFreePlugin(plugin);
        return NULL;
    }

    return plugin;
}

spOptions spInitPluginOptions(spPlugin *plugin)
{
    if (plugin == NULL || plugin->instance == NULL) return NULL;
    if (plugin->rec->init_options == NULL) return NULL;

    return plugin->rec->init_options(plugin->instance, spGetPluginLang());
}

spBool spFreePluginOptions(spPlugin *plugin, spOptions options)
{
    if (plugin == NULL || plugin->instance == NULL || options == NULL) return SP_FALSE;
    if (plugin->rec->free_options == NULL) return SP_FALSE;

    return plugin->rec->free_options(plugin->instance, options);
}

spBool spDisplayPluginCustomOptionDialog(spPlugin *plugin)
{
    if (plugin == NULL
        /*|| spIsPluginCapable(plugin, SP_PLUGIN_CAPS_CUSTOM_OPTION_DIALOG) == SP_FALSE*/) return SP_FALSE;
    if (plugin->rec->init_options != NULL /* init_options must be NULL */
        || plugin->rec->free_options == NULL) return SP_FALSE;

    if (plugin->instance == NULL
        && spInitPluginInstance(plugin) == SP_FALSE) return SP_FALSE;

    return plugin->rec->free_options(plugin->instance, NULL);
}

void spPrintPluginOptions(spPlugin *plugin)
{
    spOptions options;
    
    /* get options */
    if ((options = spInitPluginOptions(plugin)) != NULL) {
        /* print options */
        spPrintOptions(options);
        
        spFreePluginOptions(plugin, options);
    }

    return;
}

spBool spPluginUsage(const char *plugin_name)
{
    spPlugin *plugin;

    /* load plugin */
    if ((plugin = spLoadPlugin(plugin_name)) == NULL) {
        return SP_FALSE;
    }
    
    if (spInitPluginInstance(plugin) == SP_TRUE) {
        /* print options */
        spPrintPluginOptions(plugin);
    
        spFreePluginInstance(plugin);
    }
    
    /* free input plugin */
    spFreePlugin(plugin);

    return SP_TRUE;
}


spBool spIsPluginFile(const char *filename)
{
    if (filename == NULL) return SP_FALSE;

    if (isPluginCandidate(filename) == SP_TRUE
        && spEqSuffix(filename, SP_PLUGIN_SUFFIX) == SP_TRUE) {
#if 1
        spPlugin *plugin;
        if ((plugin = spLoadPlugin(filename)) != NULL) {
            spFreePlugin(plugin);
            return SP_TRUE;
        } else {
            spDebug(80, "spIsPluginFile", "spLoadPlugin failed: %s\n", filename);
        }
#else
        return SP_TRUE;
#endif
    }
    
    return SP_FALSE;
}

#define SP_NUM_PLUGIN_BUFFER 8

const char *spSearchPluginFileFromHostData(spPluginHostData host_data, int index)
{
    int i;
    char *path;
    char buf[SP_MAX_PATHNAME];
    struct dirent *dp;
    DIR *dirp = NULL;
    int num_plugin_buffer;

    spDebug(80, "spSearchPluginFile", "in: index = %d\n", index);
    
    if (index < 0) {
        spFreePluginHostDataFileList(host_data);
        return NULL;
    }

    if (host_data->plugin_files == NULL) {
        num_plugin_buffer = SP_NUM_PLUGIN_BUFFER;
        host_data->plugin_files = xalloc(num_plugin_buffer, char *);
        
        for (i = 0;; i++) {
            if ((path = xspCutPluginPathList(i)) == NULL) {
                if (i == 0) {
                    path = xspGetCurrentDir();
                } else {
                    break;
                }
            }
            spDebug(80, "spSearchPluginFile", "path = %s\n", path);
            
            /* open directory */
            if ((dirp = spOpenDir(path)) != NULL) {
                while (1) {
                    /* read directory */
                    if ((dp = spReadDir(dirp)) == NULL) {
                        spDebug(80, "spSearchPluginFile", "spReadDir returns NULL: %s\n", path);
                        break;
                    }
                    spDebug(80, "spSearchPluginFile", "dp->d_name = %s\n", dp->d_name);

                    if (!streq(dp->d_name, SP_CURRENT_DIR_STRING)
                        && !streq(dp->d_name, SP_PARENT_DIR_STRING)) {
#if defined(HAVE_SPRINTF_S)
                        sprintf_s(buf, sizeof(buf), "%s%c%s", path, SP_DIR_SEPARATOR, dp->d_name);
#elif defined(HAVE_SNPRINTF)
                        snprintf(buf, sizeof(buf), "%s%c%s", path, SP_DIR_SEPARATOR, dp->d_name);
#else
                        sprintf(buf, "%s%c%s", path, SP_DIR_SEPARATOR, dp->d_name);
#endif
                        spDebug(80, "spSearchPluginFile", "found file = %s\n", buf);
                
                        if (spIsPluginFile(buf) == SP_TRUE) {
                            spDebug(50, "spSearchPluginFile", "num_plugin_file = %d, buf = %s\n",
                                    host_data->num_plugin_file, buf);
                
                            host_data->plugin_files[host_data->num_plugin_file] = strclone(buf);
                            host_data->num_plugin_file++;
            
                            if (host_data->num_plugin_file >= num_plugin_buffer) {
                                num_plugin_buffer += SP_NUM_PLUGIN_BUFFER;
                                host_data->plugin_files = xrealloc(host_data->plugin_files, num_plugin_buffer, char *);
                            }
                        }
                    }
                }
                spCloseDir(dirp);
            }
            spDebug(80, "spSearchPluginFile", "%s dir search finished\n", path);
            xfree(path);
        }
    }

    spDebug(80, "spSearchPluginFile", "index = %d, num_plugin_file = %d\n",
            index, host_data->num_plugin_file);
    
    if (index >= host_data->num_plugin_file) {
        const char *cpath;
        
        if ((cpath = getPluginRecListName(host_data, index - host_data->num_plugin_file, SP_TRUE)) != NULL) {
            return cpath;
        } else {
            return NULL;
        }
    }
    spDebug(80, "spSearchPluginFile", "plugin file = %s\n", host_data->plugin_files[index]);
    
    return host_data->plugin_files[index];
}

char *xspSearchPluginFileFromHostData(spPluginHostData host_data, int index)
{
    return strclone(spSearchPluginFileFromHostData(host_data, index));
}


const char *spSearchPluginFile(int index)
{
    spPluginHostData host_data;

    if ((host_data = spGetPluginHostData()) == NULL) {
        return NULL;
    }

    return spSearchPluginFileFromHostData(host_data, index);
}

char *xspSearchPluginFile(int index)
{
    return strclone(spSearchPluginFile(index));
}

void spFreePluginGlobals(void)
{
    if (sp_plugin_host_data != NULL) {
        spFreePluginHostData(sp_plugin_host_data, SP_FALSE);
        sp_plugin_host_data = NULL;
    }
    
    spSetPluginSearchPath(NULL);

    return;
}

spBool spIsPluginHostCallbackSet(spPlugin *plugin)
{
    if (plugin != NULL && plugin->host != NULL && plugin->host->version_id >= 1005
        && plugin->host->callback_func != NULL) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spSetPluginHostCallback(spPlugin *plugin, spPluginHostCallback callback_func, void *user_data)
{
    if (plugin != NULL && plugin->host != NULL && plugin->host->version_id >= 1005) {
        plugin->host->callback_func = callback_func;
        plugin->host->callback_user_data = user_data;

        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spEmitPluginHostCallback(spPluginRec *rec, spPluginHostCallbackReason reason, void *callback_data)
{
    spPluginHost *host;
    
    host = (spPluginHost *)rec->host_data;

    if (host == NULL || host->version_id < 1005
        || host->callback_func == NULL) return SP_FALSE;

    return host->callback_func(rec, reason, callback_data, host->callback_user_data);
}

spBool spAllocPluginSpecificDataToHost(spPluginHost *host, int size)
{
    if (host == NULL || host->version_id < 1007) return SP_FALSE;

    if (host->plugin_specific_data != NULL) {
        xfree(host->plugin_specific_data);
    }
    if (size > 0) {
        host->plugin_specific_data = xalloc(size, char);
        memset(host->plugin_specific_data, 0, size);
    } else {
        host->plugin_specific_data = NULL;
    }

    return SP_TRUE;
}

spBool spAllocPluginSpecificData(spPlugin *plugin, int size)
{
    if (plugin == NULL) return SP_FALSE;

    return spAllocPluginSpecificDataToHost(plugin->host, size);
}

void *spGetPluginSpecificDataFromHost(spPluginHost *host)
{
    if (host == NULL || host->version_id < 1007) return NULL;

    return host->plugin_specific_data;
}

void *spGetPluginSpecificData(spPlugin *plugin)
{
    if (plugin == NULL) return NULL;

    return spGetPluginSpecificDataFromHost(plugin->host);
}

spBool spEqPluginNameString(const char *name1, const char *name2)
{
    char name1_buf[SP_MAX_LINE];
    char name2_buf[SP_MAX_LINE];

    spStrCopy(name1_buf, sizeof(name1_buf), spGetBaseName(name1));
    spRemoveSuffix(name1_buf, NULL);
    
    spStrCopy(name2_buf, sizeof(name2_buf), spGetBaseName(name2));
    spRemoveSuffix(name2_buf, NULL);

    if (streq(name1_buf, name2_buf)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spEqPluginName(spPlugin *plugin, const char *name)
{
    const char *plugin_name;
    const char *plugin_id;

    if (plugin == NULL) return SP_FALSE;
    
    plugin_name = spGetPluginName(plugin);

    if (strcaseeq(plugin_name, name) || strcaseeq(spGetBaseName(plugin_name), name)) {
        return SP_TRUE;
    }

    plugin_id = spGetPluginId(plugin);
    
    if (strcaseeq(plugin_id, name)) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spPlugin *spSearchSpecificPlugin(int *index, spSearchSpecificPluginCallback call_func, void *call_data)
{
    int i;
    const char *filename;
    int tindex = 0;
    spPlugin *plugin = NULL;

    if (call_func == NULL) return NULL;

    if (index != NULL) {
        tindex = *index;
    }
    tindex = MAX(tindex, 0);
    spDebug(80, "spSearchSpecificPlugin", "before loop: tindex = %d\n", tindex);

    for (i = tindex;; i++) {
        spDebug(100, "spSearchSpecificPlugin", "i = %d\n", i);
        
        if ((filename = spSearchPluginFile(i)) == NULL) {
            spDebug(80, "spSearchSpecificPlugin", "search end: i = %d\n", i);
            /* search end */
            tindex = -1;
            break;
        }
        spDebug(80, "spSearchSpecificPlugin", "i = %d, flename = %s, call spLoadPlugin\n", i, filename);
        if ((plugin = spLoadPlugin(filename)) != NULL) {
            if (call_func(plugin, call_data) == SP_TRUE) {
                tindex = i;
                spDebug(80, "spSearchSpecificPlugin", "found: index = %d\n", tindex);
                break;
            }
            
            spFreePlugin(plugin);
            plugin = NULL;
        }
    }

    spDebug(80, "spSearchSpecificPlugin", "after loop: tindex = %d\n", tindex);
    
    if (index != NULL) *index = tindex;

    return plugin;
}

char *xspSearchSpecificPluginFile(int *index, spSearchSpecificPluginCallback call_func, void *call_data)
{
    int tindex = 0;
    spPlugin *plugin;
    const char *filename;

    if (index != NULL) tindex = *index;
    
    if ((plugin = spSearchSpecificPlugin(&tindex, call_func, call_data)) != NULL) {
        filename = spSearchPluginFile(tindex);
        spFreePlugin(plugin);

        if (index != NULL) *index = tindex;
    
        return strclone(filename);
    }

    return NULL;
}

int spGetNumSpecificPlugin(spSearchSpecificPluginCallback call_func, void *call_data)
{
    int tindex = 0;
    int num = 0;
    spPlugin *plugin;

    while (1) {
        if ((plugin = spSearchSpecificPlugin(&tindex, call_func, call_data)) == NULL) {
            break;
        }
        spFreePlugin(plugin);
        ++num;
        ++tindex;
    }

    return num;
}
